/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.commons.events;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.almendra.commons.threads.LockUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.IListEvaluator;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;


/**
 * Event registry that lets the user register for events they are interested
 * in. When an event has occurred all the registered event handlers are
 * invoked.
 * 
 * @author Oliver Plohmann
 */
public class EventRegistry implements IEventRegistry
{
	/**
	 * Registry of events with event name and event notifiers for the event as
	 * values.
	 */
	protected Map registry = new Hashtable();
	protected ReentrantLock lock = new ReentrantLock();

	/**
	 * Creates a new <code>EventRegistry</code> object.
	 */
	public EventRegistry()
	{
		super();
	}

	/**
	 * Register for an event providing a list of event handlers that should be
	 * invoked on event occurrence.
	 * 
	 * @param eventName
	 * @param notifier
	 * @throws IllegalArgumentException
	 * @see org.almendra.commons.commons.events.IEventRegistry#register(org.almendra.commons.commons.events.Event,
	 *      org.almendra.commons.commons.events.EventNotifier)
	 */
	public void register(String eventName, IListEvaluator listEvaluator) throws IllegalArgumentException
	{
		if (!StringUtility.isValid(eventName))
			throw new IllegalArgumentException("no name specified for event"); //$NON-NLS-1$
		LockUtility.acquire(lock);
		List notifiers = (List) registry.get(eventName);
		if (notifiers == null)
		{
			notifiers = new LinkedList();
			registry.put(eventName, notifiers);
		}
		notifiers.add(getEventNotifier(listEvaluator));
		lock.release();
	}

	/**
	 * @param listEvaluator
	 * @return
	 */
	protected EventNotifier getEventNotifier(IListEvaluator listEvaluator)
	{
		return new EventNotifier(listEvaluator);
	}

	/**
	 * Deregister for event. All event handlers for the event are removed from
	 * the event registry.
	 * 
	 * @param eventName
	 * @return @see org.almendra.commons.commons.events.IEventRegistry#deregister(org.almendra.commons.commons.events.Event)
	 */
	public boolean deregister(String eventName, IListEvaluator listEvaluator) throws NoSuchElementException
	{
		LockUtility.acquire(lock);
		List notifiers = (List) registry.get(eventName);
		if (notifiers == null)
		{	
			lock.release();
			throw new NoSuchElementException("no event registered named " + eventName); //$NON-NLS-1$
		}
		Iterator it = new ArrayList(notifiers).iterator();
		while (it.hasNext())
		{
			EventNotifier notifier = (EventNotifier) it.next();
			if (notifier.getEventHandlers() == listEvaluator)
			{	
				boolean result = notifiers.remove(notifier);
				lock.release();
				return result;
			}
		}
		lock.release();
		throw new NoSuchElementException("list evaluator for event named " + eventName + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	/**
	 * Notify the event registry that an event has occurred. The event registry
	 * will notify all event handlers registered for the event.
	 * 
	 * @param event
	 *            event that has occurred specified by its name.
	 */
	public void occured(Event event)
	{
		LockUtility.acquire(lock);
		List handlers = (List) registry.get(event.getName());
		if (handlers == null)
		{
			lock.release();			
			return;
		}
		Iterator it = handlers.iterator();
		while (it.hasNext())
		{
			EventNotifier notifier = (EventNotifier) it.next();
			notifier.occured(event);
		}
		lock.release();
	}

	/**
	 * Answer whether an event is registred with the event registry.
	 * 
	 * @param eventName
	 *            name of the event
	 * @return
	 */
	public boolean isRegistered(String eventName)
	{
		LockUtility.acquire(lock);
		boolean contains = registry.containsKey(eventName);
		lock.release();
		return contains;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.ILocalClusterEventRegistryRemote#getHandlerCount(java.lang.String)
	 */
	public int getHandlerCount(String eventName)
	{
		int count = 0;
		List handlers = (List) registry.get(eventName);
		if(handlers == null)
			return count;
		Iterator it = handlers.iterator();
		while (it.hasNext())
		{
			EventNotifier notifier = (EventNotifier) it.next();
			count += notifier.getEventHandlers().get().size();
		}
		return count;
	}

}
