/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.commons.events;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.logging.ErrorLogger;


import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;


public class SimpleEventRegistry implements ISimpleEventRegistry
{
	protected Map registry = new Hashtable();
	protected ReentrantLock lock = new ReentrantLock();
	
	/**
	 * Creates a new <code>SimpleEventRegistry</code> object.
	 */
	public SimpleEventRegistry()
	{
		super();
	}
	
	public void register(String eventName, IEventHandler eventHandler) throws IllegalArgumentException
	{
		registerPrim(eventName, eventHandler);
	}
	
	public void registerPrim(String eventName, Object handler) throws IllegalArgumentException
	{
		checkEventNameValid(eventName);
		
		try
		{
			lock.acquire();
			List handlers = (List) registry.get(eventName);
			if (handlers == null)
			{
				handlers = new LinkedList();
				registry.put(eventName, handlers);
			}
			handlers.add(handler);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			ErrorLogger.exception(e);
		}
		finally
		{
			lock.release();
		}
	}
	
	protected void checkEventNameValid(String eventName)
	{
		if (!StringUtility.isValid(eventName))
			throw new IllegalArgumentException("no name specified for event"); //$NON-NLS-1$
	}
	
	public void deregister(String eventName, IEventHandler handler) throws NoSuchElementException, IllegalArgumentException
	{
		deregisterPrim(eventName, handler);
	}
	
	protected void deregisterPrim(String eventName, Object handler) throws NoSuchElementException, IllegalArgumentException
	{
		checkEventNameValid(eventName);
		
		try
		{
			lock.acquire();
			List handlers = (List) registry.get(eventName);
			if(handlers == null)
				throwNoSuchEvent(eventName);
			boolean found = handlers.remove(handler);
			if(!found)
				throwNoSuchEvent(eventName);
			if(handlers.isEmpty())
				registry.remove(eventName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			// TODO - write ex to logger
		}
		finally
		{
			lock.release();
		}
	}

	/**
	 * @param eventName
	 */
	protected void throwNoSuchEvent(String eventName)
	{
		throw new NoSuchElementException("no handler registered for event " + eventName + "!"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	public void occured(Event event)
	{
		try
		{
			lock.acquire();
			List handlers = (List) registry.get(event.getName());
			if(handlers == null)
				return;
			Iterator it = handlers.iterator();
			while (it.hasNext())
			{
				IEventHandler handler = (IEventHandler) it.next();
				handler.occured(event);
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			// TODO - write ex to logger
		}
		finally
		{
			lock.release();
		}
	}
}
