/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.threads;

import java.util.ArrayList;
import java.util.List;

import org.almendra.commons.IHistorizable;
import org.almendra.commons.IHistorySummary;
import org.almendra.commons.INamedObject;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.logging.ErrorLogger;

import EDU.oswego.cs.dl.util.concurrent.Latch;

/**
 * @author Oliver
 *
 */
public abstract class Future implements INamedObject, ISignalable, IHistorizable
{
	List historyItems = new ArrayList();
	
	/**
	 * Creates a new <code>Future</code> object.
	 */
	public Future()
	{
		super();
	}
	
	/**
	 * Creates a new <code>Future</code> object.
	 */
	public Future(String name)
	{
		super();
		this.name = name;
	}

	/**
	 * User-defined name of the future. Can be used to write trace information
	 * to the console to determine which future is waiting for a result.
	 */
	protected String name = null;
	
	/**
	 * Used to block the calling thread if it requests the result that has not been
	 * produced so far. A latch is a condition that is acquirable forever more after 
	 * the first release. See http://gee.cs.oswego.edu/dl/classes/EDU/oswego/cs/dl/util/concurrent/intro.html 
	 */
	protected Latch latch = new Latch();
	
	/**
	 * Answers whether a timeout occurred waiting for the result.
	 */
	protected boolean timeoutOccurred = false;
	
	/**
	 * Variable set to true when the <code>latch</code> has been released. Used to determine
	 * whether latch has been released explicitly after result has arrived or whether a
	 * timeout has occurred.
	 */
	protected boolean signaled = false;
	
	/**
	 * Used to synchronize access to the <code>signaled</code> and <code>holds</code> state variable.
	 */
	protected Object lock = new Object();
	
	/**
	 * Stores exception that may occur waiting for result. 
	 */
	protected Exception exception = null;
	
	/**
	 * Indicates whether the <code>Future</code> is waiting or not.
	 */
	protected boolean holds = false;
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @return Returns the timeout.
	 */
	public boolean isTimeout()
	{
		return timeoutOccurred;
	}

	/**
		 * Flag that indicates whether trace information should be written to the console
		 * when the future is waiting for the result.
		 */
	protected boolean verbose = false;

	/**
	 * @param verbose The verbose to set.
	 */
	public void setVerbose(boolean verbose)
	{
		this.verbose = verbose;
	}

	/**
	 * @return Returns the verbose.
	 */
	public boolean isVerbose()
	{
		return verbose;
	}

	/**
	 * @return Returns the exception.
	 */
	public Exception getException()
	{
		return exception;
	}

	/**
	 * @param exception The exception to set.
	 */
	public void setException(Exception exception)
	{
		this.exception = exception;
	}

	/**
	 * Make the calling thread wait till the result has arrived. Can be called explicitly without
	 * calling <code>getResult(...)</code> when the calling thread only has to be blocked till some
	 * event has occurred and there is no interest in some result.
	 */
	public void acquire()
	{
		try
		{
			synchronized(latch)
			{
				if(isVerbose())
					System.out.println("future " + getName() + " waiting for result ..."); //$NON-NLS-1$ //$NON-NLS-2$
				latch.acquire();
			}
		}
		catch (InterruptedException e)
		{
			setException(e);
			System.err.println(e.getMessage());
			e.printStackTrace();
			ErrorLogger.exception(e);
		}
	}

	/**
	 * Make the calling thread wait till the result has arrived. Can be called explicitly without
	 * calling <code>getResult(...)</code> when the calling thread only has to be blocked till some
	 * event has occurred and there is no interest in some result.
	 * @param timeoutInMillisOrZero timeout period in milliseconds after which the calling thread is released
	 * @throws IllegalArgumentException thrown if <code>timeoutInMillisOrZero</code> smaller than zero
	 */
	public void acquire(long timeoutInMillisOrZero)
	throws IllegalArgumentException, TimeoutException
	{
		if(timeoutInMillisOrZero < 0)
			throw new IllegalArgumentException(" timeoutInMillisOrZero must be 0 or greater!"); //$NON-NLS-1$
		if(timeoutInMillisOrZero == 0)
		{
			acquire();
			return;
		}
		
		try
		{
			primAcquire(timeoutInMillisOrZero);
		}
		catch (TimeoutException e)
		{
			setException(e);
			throw e;
		}
	}

	/**
	 * Utility method to block the calling thread until the result has arrived or the timeout period expired. 
	 * @param milliseconds the timeout period in milliseconds
	 * @throws TimeoutException thrown when timeout period expired waiting for the result
	 */
	protected void primAcquire(long milliseconds)
	throws TimeoutException
	{
		try
		{			
			if(isVerbose())
				System.out.println("future " + getName() + " waiting for result ..."); //$NON-NLS-1$ //$NON-NLS-2$
			
			synchronized(lock)
			{	
				holds = true;				
			}
			// TODO - race condition still possible in this line
			latch.attempt(milliseconds);
			// TODO - race condition still possible in this line
			synchronized(lock)
			{			
				holds = false;
				if(!signaled)
				{
					timeoutOccurred = true;
					throw new TimeoutException(
						"timeout period expired waiting for latch of future " +	//$NON-NLS-1$ 
						StringUtility.print(getName()));
				}
			}
		}
		catch (InterruptedException e)
		{
			setException(e);
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Anser whether an exception has occurred waiting for the result.
	 * @return
	 */
	public boolean hasException()
	{
		return getException() != null;
	}

	/**
	 * Unblock the calling thread in case the result has arrived or some event has occurred.
	 */
	public void signal()
	{				
		synchronized(lock)
		{			
			signaled = true;
			holds = false;
			latch.release();
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ISignalable#holds()
	 */
	public boolean holds()
	{
		synchronized(lock)
		{
			return holds;
		}
	}

	/**
	 * Returns null if called before <code>acquire()</code> has resumed.
	 */
	/* (non-Javadoc)
	 * @see org.almendra.commons.IHistorizable#getHistoryItems()
	 */
	public List getHistoryItems()
	{
		return historyItems;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.IHistorizable#createHistorySummary(org.almendra.commons.IHistorySummary)
	 */
	public void createHistorySummary(IHistorySummary historySummary)
	{
		historySummary.create(getHistoryItems());
	}

	public void addHistoryItems(List items)
	{
		historyItems.addAll(items);
	}
	
}
