/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.commons.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Contains several helper methods not found in java.lang.String to manipulate strings.
 * As java.lang.String is a system class it cannot be subclassed or extended. 
 * Class <code>StringUtility</code> is therefore used to hold these helper methods.
 * 
 * @author Oliver Plohmann
 */

public class StringUtility
{
	public static final String TAB = "    "; //$NON-NLS-1$
	
	/**
	 * Creates a new <code>StringUtility</code> object.
	 */
	public StringUtility()
	{
		super();
	}

	/**
	 * Reverse a string without using <code>StringBuffer</code> everytime a string has to be reversed.
	 * @param str the string to be reversed
	 * @return
	 */
	public static String reverse(String str)
	{
		return new StringBuffer(str).reverse().toString();
	}

	/**
	 * Check whether a string is valid, e.g. not null and not empty.
	 * 
	 * @param str
	 * @return
	 */
	public static boolean isValid(String str)
	{
		return !(str == null || str.length() == 0);
	}

	/**
	 * Reverse a fully qualified Java package name, e.g. java.lang.String turns
	 * into String.lang.java.
	 * 
	 * @param name the package name
	 * @return the reversed package name
	 */
	public static String reversePackageName(String name)
	{
		String packageDelimiter = "."; //$NON-NLS-1$
		StringTokenizer tokenizer = new StringTokenizer(name, packageDelimiter);
		List tokens = new LinkedList();
		while (tokenizer.hasMoreTokens())
		{
			tokens.add(0, tokenizer.nextToken());
		}
		StringBuffer buffer = new StringBuffer();
		Iterator it = tokens.iterator();
		while (it.hasNext())
		{
			String token = (String) it.next();
			buffer.append(token);
			if (it.hasNext())
				buffer.append(packageDelimiter);
		}
		return buffer.toString();
	}

	/**
	 * Create a print string (string additionally enclosed with quotes) for 
	 * parameter str. Return ""null"" (with double quotes) in case str is null
	 * to make it be displayed in quotes.
	 * @param str
	 * @return
	 */
	public static String print(String str)
	{
		if (str == null)
			return "null"; //$NON-NLS-1$
		return "\"" + str + "\""; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Check whether parameter str contains parmater subStr without using the
	 * "ugly" call <code>indexOf(subStr) != -1</code> every time.
	 * @param str
	 * @param subStr
	 * @return
	 */
	public static boolean contains(String str, String subStr)
	{
		return str.indexOf(subStr) != -1;
	}

	/**
	 * Convert a String into an <code>Integer</code>. Return null if str contains non-digits. 
	 * @param str String to be converted into an <code>Integer</code>
	 * @return Integer or null if str contained non-digit chars.
	 */
	public static Integer toInteger(String str)
	{
		try
		{
			return new Integer(Integer.parseInt(str));
		}
		catch (Exception e)
		{			
		}
		return null;		
	}
	
	/**
	 * Return an array with all the substrings of a string. 
	 * @param stringWithDelimiters the string to be converted
	 * @param delimiter the delimiter that separates substrings
	 * @return an array with every substring as an element
	 */
	public static String[] toArray(String stringWithDelimiters, String delimiter)
	{
		List elements = new ArrayList();
		StringTokenizer tokenizer = new StringTokenizer(stringWithDelimiters, delimiter);
		while(tokenizer.hasMoreTokens())
			elements.add(tokenizer.nextToken());
		String[] elementArray = new String[elements.size()];
		for (int i = 0; i < elements.size(); i++)
			elementArray[i] = (String) elements.get(i);
		return elementArray;
	}
	
	public static String remove(String str, String subStr)
	{
		int index = str.indexOf(subStr);
		if(index == -1)
			return str;
		return str.substring(0, index) + str.substring(index + subStr.length(), str.length());
	}
	
	public static String getIntegerSuffix(String str)
	{
		String integerSuffix = ""; //$NON-NLS-1$
		for (int i = str.length() - 1; i > 0; i--)
		{
			char ch = str.charAt(i);
			if(Character.isDigit(ch))
			{
				integerSuffix = ch + integerSuffix;
			}
			else
			{
				return integerSuffix;
			}
		}
		return integerSuffix;
	}

	/**
	 * @param attribute
	 * @param i
	 * @return
	 */
	public static String rightPadded(String attribute, int paddingValue)
	{
		int length = attribute.length();
		if(length > paddingValue)
			return attribute.substring(0, paddingValue);
		for (int i = 0; i < paddingValue - length; i++)
			attribute += " "; //$NON-NLS-1$
		return attribute;
	}

	/**
	 * @param maxCommandsString
	 * @return
	 */
	public static boolean isAsterisk(String str)
	{
		return "*".equals(str); //$NON-NLS-1$
	}
	
	public static int linesCount(String str)
	{
		int count = count(str, "\n"); //$NON-NLS-1$
		if(count == 0 && str != null)
			count++;
		return count;
	}
	
	public static String spaces(int count)
	{
		String spaces = ""; //$NON-NLS-1$
		for (int i = 0; i < count; i++)
			spaces += " "; //$NON-NLS-1$
		return spaces;
	}
	
	public static int count(String str, String subString)
	{
		if(!StringUtility.isValid(str))
			return 0;
		if(!StringUtility.isValid(subString))
			return 0;
		StringTokenizer tok = new StringTokenizer(str, subString);
		int count = tok.countTokens();
		if(str.substring(str.length() - subString.length(), str.length()).equals(subString))
			return count;
		count--;
		return count;
	}
	
	public static String[] toArray(List list)
	{
		String[] array = new String[list.size()];
		for (int i = 0; i < list.size(); i++)
			array[i] = (String) list.get(i);
		return array;
	}
}
