/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.util.xml;

import java.util.List;

import org.almendra.commons.util.StringUtility;
import org.jdom.Element;
import org.jdom.Text;

/**
 * <code>JDOMUtility</code> provides several utility methods to deal 
 * with <code>Element</code> objects of the JDOM XML parser.
 * 
 * @author Oliver Plohmann
 */
public class JDOMUtility
{

	/**
	 * Creates a new <code>JDOMUtility</code> object. Not used since all methods are static.
	 */
	private JDOMUtility()
	{
		super();
	}

	/**
	 * Answer the child element of an JDOM element with a given tag name.
	 * By default, element must have a child, otherwise an XMLParseException is thrown.
	 * @param element JDOM element to ask for a specific sub-element  
	 * @param tagName tag name of the child element
	 * @return the child element with the given tag name
	 * @throws XMLParseException thrown if no child found with the given tag name
	 */
	public static Element getChild(Element element, String tagName)
	throws XMLParseException
	{
		return getChild(element, tagName, true);
	}
	
	/**
	 * Answer the child element of an JDOM element with a given tag name.
	 * @param element JDOM element to ask for a specific sub-element  
	 * @param tagName tag name of the child element
	 * @param mustBeNonNull throw an XMLParseException in case child must exist
	 * @return the child element with the given tag name
	 * @throws XMLParseException thrown if no child found with the given tag name
	 */
	
	public static Element getChild(Element element, String tagName, boolean mustBeNonNull)
	throws XMLParseException
	{
		Element child = element.getChild(tagName);
		if(child == null && mustBeNonNull)
			throw new XMLParseException("tag " + StringUtility.print(tagName) + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
		return child;
	}
	
	/**
	 * Answer all children of an JDOM element with a given tag name.
	 * By default, element must have children, otherwise XMLParseException is thrown.
	 * @param element JDOM element to ask for a specific child elements 
	 * @param tagName tag name of the child elements
	 * @return list of child elements
	 * @throws XMLParseException thrown if no child elements exists for the given tag name
	 */
	public static List getChildren(Element element, String tagName)
	throws XMLParseException
	{
		return getChildren(element, tagName, true);
	}
	
	/**
	 * Answer all children of an JDOM element with a given tag name.
	 * @param element JDOM element to ask for a specific child elements 
	 * @param tagName tag name of the child elements
	 * @param mustBeNonEmpty specifiy whether element without children is allowed
	 * @return list of child elements
	 * @throws XMLParseException
	 */
	public static List getChildren(Element element, String tagName, boolean mustBeNonEmpty)
	throws XMLParseException
	{
		List children = element.getChildren(tagName);
		if(mustBeNonEmpty && children.isEmpty())
			throw new XMLParseException("no elements found with tag " + StringUtility.print(tagName) + "!"); //$NON-NLS-1$ //$NON-NLS-2$
		return children;
	}
	
	/**
	 * Answer the child text of some JDOM element with a given tag name.
	 * @param element JDOM element to ask for text of specific child element
	 * @param tagName tag name of the child element
	 * @return child text of child with given tag name
	 * @throws XMLParseException thrown if no child elements exists for the given tag name
	 */
	public static String getChildText(Element element, String tagName)
	throws XMLParseException
	{
		return getChild(element, tagName).getText();
	}
	
	/**
	 * Answer the child text of the single child of some JDOM element.
	 * @param element JDOM element to ask for child text
	 * @param tagName tag name of the child element
	 * @return child text of single child
	 * @throws XMLParseException thrown if less or more than a single chil exist
	 */
	public static String getSingleChildContent(Element element, String tagName)
	throws XMLParseException
	{
		List content = element.getContent();
		if(content.size() == 0)
			throw new XMLParseException("no content found for tag " + StringUtility.print(tagName) + "!"); //$NON-NLS-1$ //$NON-NLS-2$
		if(content.size() > 1)
			throw new XMLParseException("more than 1 children for tag " + StringUtility.print(tagName) + "!"); //$NON-NLS-1$ //$NON-NLS-2$
		return ((Text) content.get(0)).getText();
	}
	
}
