/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.demo.fibonacci;

import java.math.BigInteger;

import org.almendra.commons.util.collection.StringVector;
import org.almendra.commons.util.exception.InvalidMessageException;
import org.almendra.janet.cas.scheduling.AbstractWorkLoadBalancingInterpreter;
import org.almendra.janet.cas.scheduling.CommandAccessor;
import org.almendra.janet.cas.scheduling.IWorkLoadBalancingInterpreter;

import EDU.oswego.cs.dl.util.concurrent.Latch;



public class FibonacciBalancingInterpreter extends AbstractWorkLoadBalancingInterpreter implements IWorkLoadBalancingInterpreter, Cloneable
{
	protected volatile boolean stopped = false;
	protected FibonacciBalancingCommand command = null;
	protected Latch latch = null;
	
	/**
	 * Creates a new <code>FibonacciBalancingInterpreter</code> object.
	 */
	public FibonacciBalancingInterpreter()
	{
		super();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#execute(org.almendra.commons.janet.cas.scheduling.CommandAccessor)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if (cmdAccessor.getCommand() instanceof FibonacciBalancingCommand)
		{
			command = (FibonacciBalancingCommand) cmdAccessor.getCommand();
			calculate(cmdAccessor);
		}
	}
	
	/**
	 * @param cmdAccessor
	 */
	protected void calculate(CommandAccessor cmdAccessor)
	{
		calculateTimes();
		handleResult(cmdAccessor);
	}

	/**
	 * @param i
	 * @param j
	 * @return
	 */
	protected void calculateTimes()
	{
		int completedTimes = command.getCompletedTimes();
		int totalTimes = command.getTotalTimes();
	
		for (int i = completedTimes; i < totalTimes; i++)
		{			
			checkStopped();
			if(!suspended)
			{								
				calculateFibonacci();
				if(!suspended)
				{
					command.setCompletedTimes(i + 1);
					command.setCompletedIterations(0);
				}
			}
			else
			{
				break;
			}
		}
	}

	/**
	 * 
	 */
	private void checkStopped()
	{
		if(!stopped)
			return;
		try
		{
			if(latch != null)
				latch.release();
			latch = new Latch();
			latch.acquire();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * @param cmdAccessor
	 * @param command
	 * @throws InvalidMessageException
	 */
	protected void handleResult(CommandAccessor cmdAccessor)
	{
		try
		{
			if(suspended)
			{
				if(suspensionHandler != null)
					suspensionHandler.suspended();
			}
			else
			{
				// cmdAccessor.setReply(command.getAn());
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.log(e);
		}
	}

	/**
	 * @param nthFibonacciNumber
	 * @return
	 */
	protected BigInteger calculateFibonacci()
	{
		int nthFibonacciNumber = command.getInput();
		if (nthFibonacciNumber < 3)
			return BigInteger.ZERO;
		
		BigInteger a1 = command.getA1();
		BigInteger a2 = command.getA2();
		BigInteger an = BigInteger.ZERO; 
		int completedIterations = command.getCompletedIterations();
		
		for (int i = completedIterations; i < nthFibonacciNumber; i++)
		{			
			checkStopped();
			if(suspended)
			{				
				command.setA1(a1);
				command.setA2(a2);
				break;
			}
			else
			{
				an = a1.add(a2);
				a1 = a2;
				a2 = an;
				command.setCompletedIterations(i + 1);
			}
		}
		return an;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public StringVector commandNames()
	{
		return new StringVector(FibonacciBalancingCommand.QualifiedName);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException
	{
		FibonacciBalancingInterpreter  obj = (FibonacciBalancingInterpreter) super.clone();
		return obj;
	}
	
	public void suspend()
	{
		stopped = true;
	}
	
	public void resume()
	{
		stopped = false;
		latch.release();
		latch = null;
	}
}
