/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.distributor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.almendra.commons.events.IVetoableCallbackHandler;
import org.almendra.commons.events.IVetoableEventRegistry;
import org.almendra.commons.number.Long;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.collection.Association;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.MapUtility;
import org.almendra.commons.util.swing.BillboardPanelController;
import org.almendra.janet.ade.commands.ExecuteWorkloadCommand;
import org.almendra.janet.ade.executor.AbstractAnchor;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.handlers.AskVetoShutdownEventParams;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.CapabilityPath;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LocalCapabilityPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeName;
import org.almendra.janet.cas.node.SystemApplication;

import EDU.oswego.cs.dl.util.concurrent.ReentrantWriterPreferenceReadWriteLock;



public class DistributorAnchor extends AbstractAnchor
{
	public static final String NODE_LABEL = "AD Distributor"; //$NON-NLS-1$
	public static final String NODE_VIEW_PREFIX = "AD Distributor on"; //$NON-NLS-1$
	
	public static final String LABEL = "ADE_LOAD_DISTRIBUTOR_ANCHOR"; //$NON-NLS-1$
	public static final String CAPABILITY_NAME = "ADE_LOAD_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String BALANCING_AGENT_NAME = "BALANCING_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String SHARING_AGENT_NAME = "SHARING_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String LOAD_ADMINISTRATOR_AGENT_NAME = "LOAD_ADMINISTRATOR"; //$NON-NLS-1$
	
	public static final String BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY = ".QueueSizeCategory"; //$NON-NLS-1$
	public static final String BILLBOARD_ATTRIBUTE_COMMANDS_WAITING_TIME = ".CommandsWaitingTime"; //$NON-NLS-1$
	
	public static final String LOG_EVENT_SHARING_COMMAND_DISTRIBUTED = "SHARING_COMMAND_DISTRIBUTED"; //$NON-NLS-1$
	public static final String LOG_EVENT_BALANCING_COMMAND_DISTRIBUTED = "BALANCING_COMMAND_DISTRIBUTED"; //$NON-NLS-1$
	public static final String LOG_EVENT_PANIC_CANNOT_FIND_ANY_AGENT = "PANIC_CANNOT_FIND_ANY_AGENT"; //$NON-NLS-1$
	
	protected Map workstationLoadImages = new Hashtable();
	protected ReentrantWriterPreferenceReadWriteLock imagesLock = new ReentrantWriterPreferenceReadWriteLock();
	protected Logger logger = null;
	protected String lastSelectedSharingNodeName = null;
	protected long lastSelectedSharingNodeNameTime = -1;
	protected Descriptor descriptor = null;
	protected Node node = null;
	protected Map billboardMasks = new Hashtable();
	protected IAgentProxy loadAdministratorAgent = null;
	protected IAgentProxy balancerAgent = null;
	protected CapabilityQueueSizeStore waitingQueuesStore = new CapabilityQueueSizeStore();
	protected CapabilityQueueSizeStore executingQueuesStore = new CapabilityQueueSizeStore();
	protected Map agentsCountByCapability = new Hashtable();
	protected Map workstationLoadImageByNodeName = new Hashtable();
	protected String lastSelectNodeName = null;
	
	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	protected DistributorAnchor()
	{
		super();
	}
	
	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	public DistributorAnchor(Node node, Logger parentLogger)
	{
		super();
		this.node = node;
		this.systemApplication = node.getSystemApplication();
		setLogger(parentLogger);
		registerForEvents();
	}

	/**
	 * 
	 */
	protected void registerForEvents()
	{
		IVetoableEventRegistry vetoableEventRegistry = node.getVetoableEventRegistry();
		IVetoableCallbackHandler handler = new AskVetoShutdownCallbackHandler(this);
		vetoableEventRegistry.register(AskVetoShutdownEventParams.EVENT_NAME, handler);
	}

	/**
	 * @param parentLogger
	 */
	protected void setLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setAgentName(SHARING_AGENT_NAME);
		logger.getDefaultValues().setCapabilityName(CAPABILITY_NAME);
		logger.getDefaultValues().setApplicationName(SystemApplication.NAME);
		logger.registerForEvents(registeredLogEvents());
		logger.subscribeForEvents(registeredLogEvents());
	}

	/**
	 * @param i
	 * @return
	 */
	public String getLastSelectedSharingNode()
	{
		if(lastSelectedSharingNodeNameTime == -1)
			return null;
		if(lastSelectedSharingNodeName == null)
			return null;
		long timeNow = System.currentTimeMillis();
		if(timeNow - getSharingDelay() <= lastSelectedSharingNodeNameTime)
			return lastSelectedSharingNodeName;
		return null;
	}

	/**
	 * @param string
	 */
	public void setLastSelectedSharingNode(String nodeName)
	{
		lastSelectedSharingNodeName = nodeName;
		lastSelectedSharingNodeNameTime = System.currentTimeMillis();
	}
	
	/**
	 * @param desc
	 */
	public void setDescriptor(Descriptor descriptor)
	{
		this.descriptor = descriptor;
	}
	
	protected long getSharingDelay()
	{
		long defaultDelay = 2000;
		if(descriptor == null)
			return defaultDelay;
		long delay = descriptor.getSharingDelay();
		if(delay == -1)
			return defaultDelay;
		return delay;
	}
	
	public List getAvailableNodes()
	{
		return getAvailableNodes(null);
	}
	
	public List getAvailableNodes(String ignoreWorkstationNameOrNull)
	{
		List availableNodes = new ArrayList();

		try
		{
			imagesLock.readLock().acquire();
			Iterator it = getAvailableWorkstations().iterator();
			while (it.hasNext())
			{
				WorkstationLoadImage workstation = (WorkstationLoadImage) it.next();
				if(ignoreWorkstationNameOrNull != null && !workstation.getWorkstationName().equals(ignoreWorkstationNameOrNull))
					availableNodes.addAll(workstation.getNodeNames());
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		
		return availableNodes;
	}
	
	/**
	 * @return
	 */
	public List getAvailableWorkstations()
	{
		try
		{
			List availableWorkstations = new ArrayList();
			imagesLock.readLock().acquire();
			Iterator it = workstationLoadImages.values().iterator();
			while (it.hasNext())
			{
				WorkstationLoadImage image = (WorkstationLoadImage) it.next();
				if(!image.isAboveThreshold())
					availableWorkstations.add(image);
			}
			return availableWorkstations;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return new ArrayList();
	}
	
	public void addWorkstationLoadImage(WorkstationLoadImage image)
	{
		try
		{
			imagesLock.writeLock().acquire();
			AssertUtility.assertFalse(workstationLoadImages.containsKey(image.getWorkstationName()));
			workstationLoadImages.put(image.getWorkstationName(), image);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}		
	}
	
	/**
	 * @param nodeName
	 * @param newQueueSizeCategory
	 */
	protected void displayQueueSizeCategory(String nodeName, int newQueueSizeCategory)
	{
		displayValue(nodeName, BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY, String.valueOf(newQueueSizeCategory));
	}
	
	/**
	 * @param nodeName
	 * @param newQueueSizeCategory
	 */
	protected void displayValue(String nodeName, String attribute, String value)
	{
		BillboardPanelController billboard = node.getBillboard();
		if(billboard == null)
			return;
		String fullAttributeName = nodeName + attribute;
		List mask = (List) billboardMasks.get(nodeName);
		if(mask == null)
		{
			mask = new LinkedList();
			if(billboardMasks.isEmpty())
				mask.add(BillboardPanelController.BLANK_LINE);
			mask.add(nodeName + BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY);
			AssertUtility.assertTrue(mask.contains(fullAttributeName));
			billboardMasks.put(nodeName, mask);
			billboard.setMask(nodeName, mask);
		}
		billboard.setValue(nodeName, fullAttributeName, value);
	}

	public int getQueueSizeCategory(String nodeName)
	{
		try
		{
			imagesLock.readLock().acquire();
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			return nodeLoadImage.getQueueSizeCategory();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return -1;
	}
		
	public WorkstationLoadImage getWorkstationLoadImage(String workstationName)
	{
		WorkstationLoadImage workstationLoadImage = null;
		try
		{
			imagesLock.readLock().acquire();
			workstationLoadImage = (WorkstationLoadImage) workstationLoadImages.get(workstationName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		
		if(workstationLoadImage == null)
		{
			try
			{
				imagesLock.writeLock().acquire();
				workstationLoadImage = new WorkstationLoadImage(this, workstationName);
				addWorkstationLoadImage(workstationLoadImage);
			}
			catch (InterruptedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
			finally
			{
				imagesLock.writeLock().release();
			}			
		}
		
		return workstationLoadImage;
	}
	
	public NodeLoadImage getNodeLoadImage(String nodeName)
	{
		WorkstationLoadImage workstationLoadImage = (WorkstationLoadImage) workstationLoadImageByNodeName.get(nodeName); 
		return workstationLoadImage.getNodeLoadImage(nodeName);
	}

	/**
	 * @param localAgentPath
	 * @return
	 */
	public Map getAgentsQueueSizes(LocalAgentPath localAgentPath)
	{
		return node.getClusterImage().getAgentsQueueSizes(localAgentPath);
	}

	/**
	 * @param string
	 * @return
	 */
	protected int getWorkstationQueueSizeCategoryProduct(String nodeName)
	{
		try
		{
			imagesLock.readLock().acquire();
			WorkstationLoadImage workstationLoadImage = getWorkstationLoadImage(new NodeName(nodeName).getWorkstationName());
			return workstationLoadImage.getProduct();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return -1;
	}

	/**
	 * @param string
	 * @return
	 */
	protected int getNodeQueueSizeCategory(String nodeName)
	{
		try
		{
			imagesLock.readLock().acquire();
			return getNodeLoadImage(nodeName).getQueueSizeCategory();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return -1;
	}

	/**
	 * @param string
	 * @param i
	 */
	public void setNonJanetCPULoad(String workstationName, int nonJanetCPULoad)
	{
		try
		{
			imagesLock.writeLock().acquire();
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			image.setNonJanetCPULoad(nonJanetCPULoad);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}
	}
	
	/**
	 * @param string
	 * @param i
	 */
	public void setIsAboveThreshold(String workstationName, boolean isAboveThreshold, int nonJanetCPULoad)
	{
		try
		{
			imagesLock.writeLock().acquire();
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			image.setAboveThreshold(isAboveThreshold);
			image.setNonJanetCPULoad(nonJanetCPULoad);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}
	}
	
	/**
	 * @param string
	 * @param i
	 */
	public int setNewNonJanetCPULoad(String workstationName, int nonJanetCPULoad)
	{
		try
		{
			imagesLock.writeLock().acquire();
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			int previousNonJanetCPULoad = image.getNonJanetCPULoad(); 
			image.setNonJanetCPULoad(nonJanetCPULoad);
			return previousNonJanetCPULoad;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}
		return -1;
	}

	/**
	 * @param string
	 * @param currentQueueSizeCategory
	 * @param currentCommandsWaitingTime
	 */
	public void setNewNodeValues(String nodeName, int newQueueSizeCategory, long waitingTime, Map waitingQueuesByCapability, Map executingQueuesByCapability)
	{
		try
		{
			imagesLock.writeLock().acquire();
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			if(nodeLoadImage == null)
			{
				System.out.println("NodeLoadImage for " + nodeName + " not found!");
				return;
			}
			nodeLoadImage.setQueueSizeCategory(newQueueSizeCategory);
			nodeLoadImage.setWaitingTime(waitingTime);
			displayQueueSizeCategory(nodeName, newQueueSizeCategory);
			waitingQueuesStore.update(waitingQueuesByCapability, nodeName);
			executingQueuesStore.update(executingQueuesByCapability, nodeName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}
	}
	
	/**
	 * @return Returns the loadAdministratorAgent.
	 */
	public IAgentProxy getBalancerAgent()
	{
		if(balancerAgent == null)
		{			
			try
			{
				AgentPath agentPath = new AgentPath(AgentPath.LOCAL_NODE_TAG, SystemApplication.NAME, CAPABILITY_NAME, BALANCING_AGENT_NAME);
				balancerAgent = node.getSystemApplication().getAgent(agentPath);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				node.getLogger().exception(e);
			}
		}
		return balancerAgent;
	}
	
	/**
	 * @return Returns the loadAdministratorAgent.
	 */
	public IAgentProxy getLoadAdministratorAgent()
	{
		if(loadAdministratorAgent == null)
		{			
			try
			{
				AgentPath agentPath = new AgentPath(AgentPath.LOCAL_NODE_TAG, SystemApplication.NAME, CAPABILITY_NAME, LOAD_ADMINISTRATOR_AGENT_NAME);
				loadAdministratorAgent = node.getSystemApplication().getAgent(agentPath);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				node.getLogger().exception(e);
			}
		}
		return loadAdministratorAgent;
	}
	
	/**
	 * @return Returns the node.
	 */
	public Node getNode()
	{
		return node;
	}
	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}
	
	public boolean writeProtectedDo(IExecutable executable)
	{
		try
		{
			imagesLock.writeLock().acquire();
			executable.execute(this);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return false;
		}
		finally
		{
			imagesLock.writeLock().release();
		}
		return true;
	}

	/**
	 * @param workstationName
	 * @return
	 */
	public List getAllSystemAgents(String workstationName)
	{
		List agentProxies = new ArrayList();
		
		try
		{
			imagesLock.readLock().acquire();			
			Iterator it = getWorkstationLoadImage(workstationName).getNodeNames().iterator();
			while (it.hasNext())
			{
				String nodeName = (String) it.next();
				agentProxies.add(node.getSystemApplication().getSystemAgent(nodeName));				
			}
			return agentProxies;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		
		return agentProxies;
	}
	/**
	 * @return Returns the executingQueuesStore.
	 */
	public CapabilityQueueSizeStore getExecutingQueuesStore()
	{
		return executingQueuesStore;
	}
	/**
	 * @return Returns the waitingQueuesStore.
	 */
	public CapabilityQueueSizeStore getWaitingQueuesStore()
	{
		return waitingQueuesStore;
	}

	/**
	 * @param nodeName
	 * @param capabilityPath
	 * @return
	 */
	public long getAgentsCount(String nodeName, String capabilityPath)
	{
		String key = capabilityPath + "." + nodeName; //$NON-NLS-1$
		Long countObject = (Long) agentsCountByCapability.get(key);
		if(countObject == null)
		{
			int count = node.getSystemApplication().getAgentCount(nodeName, capabilityPath);
			countObject = new Long(count);
			agentsCountByCapability.put(key, countObject);
		}
		return countObject.longValue();
	}

	/**
	 * 
	 */
	public void clearCaches()
	{
		agentsCountByCapability = new Hashtable();
	}

	/**
	 * @param string
	 */
	public void registerWorkstation(String workstationName, List nodeNames)
	{
		try
		{
			imagesLock.writeLock().acquire();
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			Iterator it = nodeNames.iterator();
			while(it.hasNext())
			{
				String nodeName = (String) it.next();
				image.addNodeLoadImage(nodeName);	
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}		
		
				
	}

	/**
	 * @param string
	 */
	public void deregisterWorkstation(String workstationName, List nodeNames)
	{
		try
		{
			imagesLock.writeLock().acquire();
			WorkstationLoadImage workstationLoadImage = getWorkstationLoadImage(workstationName);
			if(workstationLoadImage != null)
				workstationLoadImage.removeNodeLoadImagesNamed(nodeNames);
			if(!workstationLoadImage.hasNodes())
				workstationLoadImages.remove(workstationName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().release();
		}		
	}

	/**
	 * @param string
	 * @param string2
	 * @return
	 */
	public List getAllNodesWithCapability(LocalCapabilityPath path, List nodeNamestoConsider)
	{
		return systemApplication.getAllNodeNamesWithCapability(path, nodeNamestoConsider);
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	public List getLeastLoadedCapabilitiesAtLeastOne(List nodeNamesOfNodesWithCapability, LocalCapabilityPath path, LeastLoadedNodeDecisionInformation decisionInformation)
	{
		decisionInformation.setNodeNamesWithCapability(nodeNamesOfNodesWithCapability);
		if(nodeNamesOfNodesWithCapability.isEmpty())
			return null;
		
		if(nodeNamesOfNodesWithCapability.size() == 1)
			return ListUtility.toList(nodeNamesOfNodesWithCapability.get(0));
		
		try
		{
			imagesLock.readLock().acquire();
			List nodeImages = getNodeImages(nodeNamesOfNodesWithCapability);
			if(nodeImages.isEmpty())
				return null;
			List availableNodeLoadImages = getAvailableNodeLoadAtLeastOne(nodeImages, decisionInformation);
			if(availableNodeLoadImages.size() == 1)
				return ListUtility.toList(availableNodeLoadImages.get(0));
			return getLeastLoadedNodes(availableNodeLoadImages, path, decisionInformation);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		
		return null;		
	}

	/**
	 * @param availableNodeLoadImages
	 * @param path
	 * @return node name of least loaded node for capability
	 */
	protected List getLeastLoadedNodes(List availableNodeLoadImages, LocalCapabilityPath path, LeastLoadedNodeDecisionInformation decisionInformation)
	{
		if(availableNodeLoadImages.size() == 1)
			return ListUtility.toList(((NodeLoadImage) availableNodeLoadImages.get(0)).getNodeName());
		
		Map workstationQSCByNode = new Hashtable();
		Iterator it = availableNodeLoadImages.iterator();
		while (it.hasNext())
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) it.next();
			int qsc = nodeLoadImage.getWorkstationLoadImage().getQSCProduct();
			decisionInformation.addWorkstationQSCByNodeName(nodeLoadImage.getNodeName(), new Integer(qsc));
			workstationQSCByNode.put(nodeLoadImage, new Integer(qsc));
		}
		
		List nodesWithSmallestWorkstationQSC = MapUtility.getKeysWithSmallestIntegerValue(workstationQSCByNode);
		decisionInformation.addNodeNamesWithSmallestWorkstationQSC(NodeLoadImage.getNodeNames(nodesWithSmallestWorkstationQSC));
		
		Map nodeQSCByNode = new Hashtable();
		it = nodesWithSmallestWorkstationQSC.iterator();
		while (it.hasNext())
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) it.next();
			nodeQSCByNode.put(nodeLoadImage, new Integer(nodeLoadImage.getQueueSizeCategory()));
		}

		List nodesWithSmallestOverAllQSC = MapUtility.getKeysWithSmallestIntegerValue(nodeQSCByNode);
		decisionInformation.addNodeNamesWithSmallestOverAllQSC(NodeLoadImage.getNodeNames(nodesWithSmallestOverAllQSC));
		
		if(nodesWithSmallestOverAllQSC.size() == 1)
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) nodesWithSmallestOverAllQSC.get(0);
			decisionInformation.setNodeNameWithSmallestOverAllQSC(nodeLoadImage.getNodeName());
			return ListUtility.toList(nodeLoadImage);
		}
		
		Map cqsByNode = new Hashtable();
		it = nodesWithSmallestOverAllQSC.iterator();
		while (it.hasNext())
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) it.next();
			int cqs = nodeLoadImage.getCapabilityQueueSize(path);
			decisionInformation.addCQSByNode(nodeLoadImage.getNodeName(), new Integer(cqs));
			cqsByNode.put(nodeLoadImage, new Integer(cqs));
		}
		
		List leastLoadedNodes = MapUtility.getKeysWithSmallestIntegerValue(cqsByNode);
		decisionInformation.setLeastLoadedNodes(NodeLoadImage.getNodeNames(leastLoadedNodes));
		return leastLoadedNodes;
	}

	/**
	 * @param nodeImages
	 * @return
	 */
	protected List getAvailableNodeLoadAtLeastOne(List nodeImages, LeastLoadedNodeDecisionInformation decisionInformation)
	{
		List availableNodeLoadImages = new ArrayList();
		List availableNodeNames = new ArrayList();
		List notAvailableNodeLoadImages = new ArrayList();
		Iterator it = nodeImages.iterator();
		while (it.hasNext())
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) it.next();
			if(nodeLoadImage.isAboveThreshold())
				notAvailableNodeLoadImages.add(nodeLoadImage);
			else
			{
				availableNodeLoadImages.add(nodeLoadImage);
				availableNodeNames.add(nodeLoadImage.getNodeName());
			}
		}
		decisionInformation.setAvailableNodeNames(availableNodeNames);
		if(availableNodeLoadImages.size() >= 1)
			return availableNodeLoadImages;
		
		int minNonJanetCPULoad = Integer.MAX_VALUE;
		NodeLoadImage leastLoadedNotAvailableNodeLoadImage = null;
		it = notAvailableNodeLoadImages.iterator();
		while (it.hasNext())
		{
			NodeLoadImage nodeLoadImage = (NodeLoadImage) it.next();
			if(nodeLoadImage.getNonJanetCPULoad() < minNonJanetCPULoad)
			{
				leastLoadedNotAvailableNodeLoadImage = nodeLoadImage;
				minNonJanetCPULoad = nodeLoadImage.getNonJanetCPULoad();
			}
		}
		decisionInformation.setLeastLoadedNotAvailableNodeName(leastLoadedNotAvailableNodeLoadImage.getNodeName());
		return ListUtility.toList(leastLoadedNotAvailableNodeLoadImage);
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	protected List getNodeImages(List nodeNames)
	{
		List nodeLoadImages = new ArrayList();
		Iterator it = nodeNames.iterator();
		while(it.hasNext())
		{
			String nodeName = (String) it.next();
			WorkstationLoadImage workstationLoadImage = (WorkstationLoadImage) workstationLoadImageByNodeName.get(nodeName);
			if(workstationLoadImage != null)
			{
				NodeLoadImage nodeLoadImage = workstationLoadImage.getNodeLoadImage(nodeName);
				nodeLoadImages.add(nodeLoadImage);
			}
			else
			{
				System.out.println(getClass().getName() + ".getNodeImages: no workstationLoadImage for node " + nodeName);
			}
		}		
		return nodeLoadImages;
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	protected List getWorkstationImages(List nodeNames)
	{
		List nodeImages = new ArrayList();
		Iterator it = nodeNames.iterator();
		while(it.hasNext())
		{
			String nodeName = (String) it.next();
			nodeImages.add(workstationLoadImageByNodeName.get(nodeName));
		}		
		return nodeImages;
	}

	/**
	 * @param leastLoadedNodeForCapability
	 * @param originAgentPath
	 * @return
	 */
	public AgentPath getAgentPathToNode(String nodeName, AgentPath originAgentPath)
	{
		CapabilityPath path = new CapabilityPath(nodeName, originAgentPath.getApplicationName(), originAgentPath.getCapabilityName()); 
		List agentPaths = systemApplication.getAllAgentPaths(path);
		return (AgentPath) agentPaths.get(0);
	}

	/**
	 * @param command
	 * @param path
	 * @param selectedAgentPath
	 */
	public void logCommandDistributed(ExecuteWorkloadCommand command, AgentPath originAgentPath, AgentPath selectedAgentPath)
	{
		LogEntry entry = (LogEntry) logger.getNewLogEntry();
		String eventName = command.getLogEventName();
		String typeName = command.getLogTypeName();
		entry.setEvent(eventName);
		StringBuffer buffer = new StringBuffer();
		buffer.append("distributed " + typeName + command.toLogString() + " from " + originAgentPath.getQualifiedPathName()); //$NON-NLS-1$ //$NON-NLS-2$
		buffer.append(" to " + selectedAgentPath.getQualifiedPathName()); //$NON-NLS-1$
		entry.setArgument(buffer.toString());
		logger.info(entry);
	}

	/**
	 * @param nodeName
	 * @param path
	 * @return
	 */
	public int getCapabilityQueueSize(String nodeName, String qualifiedCapabilityPath)
	{
		return waitingQueuesStore.getQueueSize(nodeName, qualifiedCapabilityPath) + executingQueuesStore.getQueueSize(nodeName, qualifiedCapabilityPath);
	}
	
	/**
	 * @param nodeName
	 * @param path
	 * @return
	 */
	public int getCapabilityQueueSize(String nodeName, LocalCapabilityPath path)
	{
		return waitingQueuesStore.getQueueSize(nodeName, path.getQualifiedPathName()) + executingQueuesStore.getQueueSize(nodeName, path.getQualifiedPathName());
	}

	/**
	 * @param string
	 * @param nodeName
	 * @return
	 */
	public List getAlternativeCapabilitiesOnOtherNodes(String workstationName, String nodeName)
	{
		List nodeNames = getWorkstationLoadImage(workstationName).getNodeNames();
		List localCapabilityPaths = systemApplication.getAllApplicationCapabilitesOn(nodeName);
		return systemApplication.existingCapabilitiesOnOtherNodes(localCapabilityPaths, nodeNames);
	}

	/**
	 * @return
	 */
	public List getAvailableNodeNamesWithQSC0()
	{
		try
		{
			imagesLock.readLock().acquire();
			List nodeNamesWithQSC0 = new ArrayList();
			Iterator it = workstationLoadImages.values().iterator();
			while(it.hasNext())
			{
				WorkstationLoadImage workstationLoadImage = (WorkstationLoadImage) it.next();
				nodeNamesWithQSC0.addAll(workstationLoadImage.getAvailableNodeNamesWithQSC0());
			}
			return nodeNamesWithQSC0;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}		
		return new ArrayList();
	}

	/**
	 * @param nodeNameAboveQSC1
	 * @param nodeNameWithQSC0
	 * @return
	 */
	public List getCapabilityIntersection(String nodeName1, String nodeName2)
	{
		List localCapabilityPaths1 = systemApplication.getAllApplicationCapabilitesOn(nodeName1);
		List localCapabilityPaths2 = systemApplication.getAllApplicationCapabilitesOn(nodeName2);
		return ListUtility.getIntersection(localCapabilityPaths1, localCapabilityPaths2);
	}
	
	/**
	 * @param nodeNameToMigrateTo name of node to migrate command to
	 * @param capabilities
	 * @return name of node to evict command from
	 */
	public String getNodeNameToEvictCommandFrom(String nodeNameToMigrateTo, List nodeNamesAboveQSC1, List capabilities, NodeDroppedToQSC1DecisionInformation decisionInformation)
	{
		List destinationCapabilities = systemApplication.getAllApplicationCapabilitesOn(nodeNameToMigrateTo);
		decisionInformation.setDestinationCapabilities(LocalCapabilityPath.getQualifiedPathNames(destinationCapabilities));
		Map capabilitiesByCandidateNodeName = new Hashtable(); 
		Set allCandidateNodes = new HashSet();
		Iterator it = destinationCapabilities.iterator();
		while(it.hasNext())
		{
			LocalCapabilityPath path = (LocalCapabilityPath) it.next();
			List nodeNames = getAllNodesWithCapability(path, nodeNamesAboveQSC1);
			if(!nodeNames.isEmpty())
			{
				MapUtility.putAll(capabilitiesByCandidateNodeName, path.getQualifiedPathName(), nodeNames);
				allCandidateNodes.addAll(nodeNames);
			}
		}
		if(capabilitiesByCandidateNodeName.isEmpty())
			return null;
		decisionInformation.setCandidateNodeNamesWithRequiredCapabilities(allCandidateNodes);
		String heaviestLoadedNodeName = getHeaviestLoadedNodeName(allCandidateNodes, capabilitiesByCandidateNodeName, capabilities, decisionInformation);
		decisionInformation.setHeaviestLoadedNodeName(heaviestLoadedNodeName);
		capabilities.addAll((Collection) capabilitiesByCandidateNodeName.get(heaviestLoadedNodeName));
		decisionInformation.setCapabilitiesOfHeaviestLoadedNode(capabilities);
		return heaviestLoadedNodeName;
	}

	/**
	 * @param allCandidateNodes
	 * @param capabilitiesByCandidateNodeName
	 * @return
	 */
	protected String getHeaviestLoadedNodeName(Set allCandidateNodes, Map capabilitiesByCandidateNodeName, List heaviestLoadedCapabilityPaths, NodeDroppedToQSC1DecisionInformation decisionInformation)
	{
		Map workstationQSCByNodeName = new Hashtable();
		Iterator it = allCandidateNodes.iterator();
		while (it.hasNext())
		{
			String nodeName = (String) it.next();
			int workstationQSC = getNodeLoadImage(nodeName).getWorkstationLoadImage().getQSCProduct();
			workstationQSCByNodeName.put(nodeName, new Integer(workstationQSC));
		}
		List smallestWorkstationQSCNodeNames = MapUtility.getKeysWithSmallestIntegerValue(workstationQSCByNodeName);
		Map nodeQSCByNodeName = new Hashtable();
		it = smallestWorkstationQSCNodeNames.iterator();
		while (it.hasNext())
		{
			String nodeName = (String) it.next();
			int nodeQSC = getNodeLoadImage(nodeName).getQueueSizeCategory();
			nodeQSCByNodeName.put(nodeName, new Integer(nodeQSC));
		}
		List smallestNodesQSCNodeNames = MapUtility.getKeysWithSmallestIntegerValue(nodeQSCByNodeName);
		Map smallestCQSByNodeCapabilityPathAssoc = new Hashtable();
		it = smallestNodesQSCNodeNames.iterator();
		while(it.hasNext())
		{
			String nodeName = (String) it.next();
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			List capabilities = (List) capabilitiesByCandidateNodeName.get(nodeName);
			Iterator innerIt = capabilities.iterator();
			while(innerIt.hasNext())
			{
				String capabilityPath = (String) innerIt.next();
				int cqs = nodeLoadImage.getCapabilityQueueSize(capabilityPath);
				smallestCQSByNodeCapabilityPathAssoc.put(new Association(nodeName, capabilityPath), new Integer(cqs));
			}
		}
		Association smallestNodeCapabilityPathAssocs = (Association) MapUtility.getKeyWithGreatestIntegerValue(smallestCQSByNodeCapabilityPathAssoc);
		if(smallestNodeCapabilityPathAssocs == null)
			return null;
		String heaviestLoadedNode = (String) smallestNodeCapabilityPathAssocs.getKey();
		String heaviestLoadedCapabilityPath = (String) smallestNodeCapabilityPathAssocs.getValue();
		heaviestLoadedCapabilityPaths.add(heaviestLoadedCapabilityPath);
		return heaviestLoadedNode;
	}

	/**
	 * @param nodeNameDroppedToQSC1
	 * @param capabilityPath
	 * @return
	 */
	public AgentPath getAnyAgent(String nodeName, String capabilityPath)
	{
		LocalCapabilityPath path = new LocalCapabilityPath(capabilityPath);
		List agentPaths = systemApplication.getAllAgentPaths(new CapabilityPath(nodeName, path.getApplicationName(), path.getCapabilityName()));
		if(agentPaths == null || agentPaths.isEmpty())
			return null;
		return (AgentPath) agentPaths.get(0);
	}

	/**
	 * @param nodeNameAboveQSC1
	 * @param capabilitiesIntersection
	 * @return
	 */
	public String getHeaviestLoadedCapabilityPath(String nodeName, List capabilities)
	{
		// TODO - refactor getHeaviestLoadedNode to use this meethod
		NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
		if(nodeLoadImage == null)
			return null;
		Map cqsByCapabilityPath = new Hashtable();
		Iterator it = capabilities.iterator();
		while(it.hasNext())
		{
			LocalCapabilityPath capabilityPath = (LocalCapabilityPath) it.next();
			int cqs = nodeLoadImage.getCapabilityQueueSize(capabilityPath.getQualifiedPathName());
			cqsByCapabilityPath.put(capabilityPath.getQualifiedPathName(), new Integer(cqs));
		}
		return (String) MapUtility.getKeyWithGreatestIntegerValue(cqsByCapabilityPath);
	}
	
	public List getAllConnectedNodeNames()
	{
		return ListUtility.toList(workstationLoadImageByNodeName.keySet());
	}

	public List getAvailableNodeNamesAboveQSC(int qsc)
	{
		try
		{
			List nodeNames = new ArrayList();
			imagesLock.readLock().acquire();
			Iterator it = workstationLoadImages.values().iterator();
			while(it.hasNext())
			{
				WorkstationLoadImage workstationLoadImage = (WorkstationLoadImage) it.next();
				if(workstationLoadImage.isBelowThreshold())
				{
					Iterator innerIt = workstationLoadImage.getNodeLoadImages().iterator();
					while (innerIt.hasNext())
					{
						NodeLoadImage nodeLoadImage = (NodeLoadImage) innerIt.next();
						if(nodeLoadImage.getQueueSizeCategory() > qsc)
							nodeNames.add(nodeLoadImage.getNodeName());
					}
				}
			}
			return nodeNames;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}		
		return new ArrayList();
	
	}
	
	/**
	 * @return
	 */
	public boolean existsNodeAboveQSC0()
	{
		return !getAvailableNodeNamesAboveQSC(0).isEmpty();
	}
	
	/**
	 * @return
	 */
	public boolean isDisableEviction()
	{
		return descriptor.isDisableEviction();
	}
	/**
	 * @return Returns the lastSelectedNodeName.
	 */
	public String getLastSelectNodeName()
	{
		return lastSelectNodeName;
	}
	/**
	 * @param lastSelectedNodeName The lastSelectedNodeName to set.
	 */
	public void setLastSelectNodeName(String lastSelectNodeName)
	{
		this.lastSelectNodeName = lastSelectNodeName;
	}

	/**
	 * @param nodeNameHeaviestLoaded
	 * @param capabilities
	 * @return
	 */
	public String getHeaviestLoadedCapability(String nodeNameHeaviestLoaded, List capabilityPathNames, NodeDroppedToQSC1DecisionInformation decisionInformation)
	{
		Map cqsByCapabilityName = new Hashtable();
		Iterator it = capabilityPathNames.iterator();
		while(it.hasNext())
		{
			String capabilityPathName = (String) it.next();
			int cqs = getCapabilityQueueSize(nodeNameHeaviestLoaded, capabilityPathName);
			cqsByCapabilityName.put(capabilityPathName, new Integer(cqs));
		}
		decisionInformation.setCQSByCapabilityName(cqsByCapabilityName);
		// TODO - improve not to pick an element at random if cqsByCapabilityName.size() > 1 
		String heaviestLoadedCapabilityName = (String) MapUtility.getKeyWithGreatestIntegerValue(cqsByCapabilityName);
		decisionInformation.setHeaviestLoadedCapabilityName(heaviestLoadedCapabilityName);
		return heaviestLoadedCapabilityName;
	}

	/**
	 * 
	 */
	public void acquireReadLock() throws InterruptedException
	{
		imagesLock.readLock().acquire();
	}

	/**
	 * 
	 */
	public void releaseReadLock()
	{
		imagesLock.readLock().release();
	}
	/**
	 * @return
	 */
	public boolean isDisableFullEviction()
	{
		return descriptor.isDisableFullEviction();
	}
	/**
	 * @return
	 */
	public boolean isDisableNodeDroppedToQSC1()
	{
		return descriptor.isDisableNodeDroppedToQSC1();
	}
	/**
	 * @return
	 */
	public boolean isDisableNodesWithQSC0Exist()
	{
		return descriptor.isDisableNodesWithQSC0Exist();
	}
	/**
	 * @return
	 */
	public boolean isDisablePartialEviction()
	{
		return descriptor.isDisablePartialEviction();
	}
	/**
	 * @return Returns the descriptor.
	 */
	public Descriptor getDescriptor()
	{
		return descriptor;
	}

	/**
	 * @return
	 */
	public boolean hasWorkstations()
	{
		try
		{
			imagesLock.readLock().acquire();
			return !workstationLoadImages.isEmpty();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return true;
	}

	/**
	 * @return
	 */
	public List getWorkstationNames()
	{
		try
		{
			imagesLock.readLock().acquire();
			return ListUtility.toList(workstationLoadImages.keySet());
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().release();
		}
		return new ArrayList();
	}
}
