/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.ade.distributor.interpreters;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.StringVector;
import org.almendra.commons.util.logging.ErrorLogger;
import org.almendra.commons.util.swing.TextAreaPanelController;
import org.almendra.janet.ade.commands.ProcessQSCChangedCommand;
import org.almendra.janet.ade.commands.ReduceLoadCommand;
import org.almendra.janet.ade.distributor.DistributorAnchor;
import org.almendra.janet.ade.distributor.NodeDroppedToQSC1DecisionInformation;
import org.almendra.janet.ade.distributor.PartialEvictionAssignment;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.scheduling.CommandAccessor;


/**
 * @author Oliver Plohmann
 *  
 */
public class ProcessQSCChangedInterpreter extends DistributorInterpreter
{
	/**
	 * Creates a new <code>ProcessQSCChangedInterpreter</code> object.
	 */
	public ProcessQSCChangedInterpreter()
	{
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#execute(org.almendra.commons.janet.cas.scheduling.CommandAccessor)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if (cmdAccessor.getCommand() instanceof ProcessQSCChangedCommand)
		{
			ProcessQSCChangedCommand command = (ProcessQSCChangedCommand) cmdAccessor.getCommand();
			if (command.getChange() == ProcessQSCChangedCommand.NodesWithQSC0Exist)
			{
				reflectNodesWithQSC0Exist(cmdAccessor, command);
				return;
			}
			if (command.getChange() == ProcessQSCChangedCommand.NodeDroppedToQSC0)
			{
				reflectNodeDroppedToQSC0(cmdAccessor, command);
				return;
			}
		}
	}

	/**
	 * @param cmdAccessor
	 * @param command
	 */
	protected void reflectNodeDroppedToQSC0(CommandAccessor cmdAccessor, ProcessQSCChangedCommand command)
	{
		DistributorAnchor anchor = getDistributorAnchor(cmdAccessor);
		String nodeNameDroppedToQSC0 = checkOnQSC0(command.getNodeName(), anchor);
		if(nodeNameDroppedToQSC0 == null)
			return;
		List nodeNamesAboveQSC1 = command.getNodeNames();
		if(nodeNamesAboveQSC1.isEmpty())
			return;
		NodeDroppedToQSC1DecisionInformation decisionInformation = new NodeDroppedToQSC1DecisionInformation(nodeNameDroppedToQSC0, nodeNamesAboveQSC1);
		String nodeNameHeaviestLoaded = null;
		String capabilityPath = null;
		List capabilities = new ArrayList();
		try
		{
			anchor.acquireReadLock();			
			nodeNameHeaviestLoaded = anchor.getNodeNameToEvictCommandFrom(nodeNameDroppedToQSC0, nodeNamesAboveQSC1, capabilities, decisionInformation);
			if (capabilities.isEmpty() || nodeNameHeaviestLoaded == null)
				return;
			capabilityPath = anchor.getHeaviestLoadedCapability(nodeNameHeaviestLoaded, capabilities, decisionInformation);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			ErrorLogger.exception(e);
		}
		finally
		{
			anchor.releaseReadLock();
		}
		TextAreaPanelController freeText = cmdAccessor.getNodeAccessor().getFreeTextArea();
		decisionInformation.printOn(freeText);
		evict1Command(anchor, nodeNameHeaviestLoaded, nodeNameDroppedToQSC0, capabilityPath);
	}

	/**
	 * @param string
	 * @param anchor
	 * @return
	 */
	protected String checkOnQSC0(String nodeName, DistributorAnchor anchor)
	{
		if(anchor.getNodeLoadImage(nodeName).getQueueSizeCategory() == 0)
			return nodeName;
		return null;
	}

	/**
	 * @param anchor
	 * @param nodeNameHeaviestLoaded
	 * @param nodeNameDroppedToQSC0
	 * @param capabilities
	 */
	protected void evict1Command(DistributorAnchor anchor, String nodeNameHeaviestLoaded, String nodeNameDroppedToQSC0,	String capabilityPath)
	{		
		try
		{
			AgentPath destinationAgent = anchor.getAnyAgent(nodeNameDroppedToQSC0, capabilityPath);
			PartialEvictionAssignment assignment = new PartialEvictionAssignment(capabilityPath, destinationAgent);
			anchor.getNodeLoadImage(nodeNameDroppedToQSC0).setMinQueueSizeCategory(1);
			ReduceLoadCommand command = new ReduceLoadCommand(false, ListUtility.toList(assignment));
			IAgentProxy systemAgent = anchor.getSystemApplication().getSystemAgent(nodeNameHeaviestLoaded);
			systemAgent.accept(command);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			ErrorLogger.exception(e);
		}
	}

	/**
	 * @param cmdAccessor
	 * @param command
	 */
	protected void reflectNodesWithQSC0Exist(CommandAccessor cmdAccessor, ProcessQSCChangedCommand command)
	{
		DistributorAnchor anchor = getDistributorAnchor(cmdAccessor);
		List nodesNamesWithQSC0 = checkAllOnQSC0(command.getNodeNames(), anchor);
		String nodeNameAboveQSC1 = command.getNodeName();		
		Iterator it = nodesNamesWithQSC0.iterator();
		while (it.hasNext())
		{
			String nodeNameWithQSC0 = (String) it.next();
			List capabilitiesIntersection = anchor.getCapabilityIntersection(nodeNameAboveQSC1, nodeNameWithQSC0);
			if (capabilitiesIntersection.isEmpty())
				return;
			String capabilityPath = anchor.getHeaviestLoadedCapabilityPath(nodeNameAboveQSC1, capabilitiesIntersection);
			if(capabilityPath == null)
				return;
			evict1Command(anchor, nodeNameAboveQSC1, nodeNameWithQSC0, capabilityPath);
		}
	}

	/**
	 * @param list
	 * @param anchor
	 * @return
	 */
	protected List checkAllOnQSC0(List nodeNames, DistributorAnchor anchor)
	{
		List nodeNamesOnQSC0 = new ArrayList();
		Iterator it = nodeNames.iterator();
		while (it.hasNext())
		{
			String nodeName = (String) it.next();
			if(checkOnQSC0(nodeName, anchor) != null)
				nodeNamesOnQSC0.add(nodeName);
		}
		return nodeNamesOnQSC0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public StringVector commandNames()
	{
		return new StringVector(ProcessQSCChangedCommand.QualifiedName);
	}
}