/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.distributor.interpreters;

import java.io.File;
import java.io.IOException;

import org.almendra.commons.events.ICallbackHandler;
import org.almendra.commons.events.IEventHandler;
import org.almendra.commons.util.collection.StringVector;
import org.almendra.commons.util.xml.XMLParseException;
import org.almendra.janet.ade.commands.StartCommand;
import org.almendra.janet.ade.distributor.Descriptor;
import org.almendra.janet.ade.distributor.DistributorAnchor;
import org.almendra.janet.ade.distributor.gui.OpenSettingsViewHandler;
import org.almendra.janet.ade.distributor.gui.SettingsView;
import org.almendra.janet.ade.distributor.handlers.NodeDeregisterHandler;
import org.almendra.janet.cas.events.EventHandlerEnvelope;
import org.almendra.janet.cas.events.HandlerList;
import org.almendra.janet.cas.handlers.NodeDeregisteredParams;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.gui.NodeViewTitleAccessor;
import org.almendra.janet.cas.node.gui.NodeViewsRegistry;
import org.almendra.janet.cas.scheduling.CommandAccessor;
import org.almendra.janet.cas.scheduling.IInterpreter;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;



public class StartInterpreter extends DistributorInterpreter implements IInterpreter
{
	/**
	 * Creates a new <code>StartInterpreter</code> object.
	 */
	public StartInterpreter()
	{
		super();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#execute(org.almendra.commons.janet.cas.scheduling.CommandAccessor)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if(cmdAccessor.getCommand() instanceof StartCommand)
		{
			cmdAccessor.setNodeLabel(DistributorAnchor.NODE_LABEL);
			DistributorAnchor anchor = new DistributorAnchor(cmdAccessor.getNodeIfPrivilege(), cmdAccessor.getLogger()); 
			cmdAccessor.getApplicationObjectSpace().put(DistributorAnchor.LABEL, anchor);
			readDescriptor(anchor, cmdAccessor);
			prependNodeViewTitle(cmdAccessor);			
			registerForEvents(cmdAccessor, anchor);
			setNodeViewsRegistry(cmdAccessor, anchor.getDescriptor());
		}
	}

	/**
	 * @param cmdAccessor
	 * @param descriptor
	 */
	protected void setNodeViewsRegistry(CommandAccessor cmdAccessor, Descriptor descriptor)
	{
		NodeViewsRegistry nodeViewsRegistry = (NodeViewsRegistry) cmdAccessor.getNodeObjectSpace().get(NodeViewsRegistry.LABEL);
		ICallbackHandler handler = new OpenSettingsViewHandler(descriptor);
		nodeViewsRegistry.addViewEntry(SettingsView.LABEL, handler);
	}

	/**
	 * @param cmdAccessor
	 * @param anchor
	 */
	protected void registerForEvents(CommandAccessor cmdAccessor, DistributorAnchor anchor)
	{
		AgentPath systemAgentPath = cmdAccessor.getSystemApplicationIfPrivilege().getSystemAgentPath();
		IEventHandler handler = new NodeDeregisterHandler(anchor);
		EventHandlerEnvelope env = new EventHandlerEnvelope(handler, systemAgentPath);
		HandlerList handlerList = new HandlerList(env);
		cmdAccessor.getNodeEventRegistry().register(NodeDeregisteredParams.EVENT_NAME, handlerList, cmdAccessor.getApplication());
	}

	/**
	 * 
	 */
	protected void prependNodeViewTitle(CommandAccessor cmdAccessor)
	{
		NodeViewTitleAccessor accessor = (NodeViewTitleAccessor) 
			cmdAccessor.getSystemApplicationIfPrivilege().getObjectSpace().get(NodeViewTitleAccessor.LABEL);
		accessor.prepend(DistributorAnchor.NODE_VIEW_PREFIX);
	}

	/**
	 * @param anchor
	 * @param capability
	 */
	protected void readDescriptor(DistributorAnchor anchor, CommandAccessor cmdAccessor)
	{
		String descriptorFilePathName = cmdAccessor.getCapability().getDescriptorFilePathName();
		if(descriptorFilePathName == null)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append("\n" + getClass().getName() + ".readDescriptor(...): "); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append("descriptor null!"); //$NON-NLS-1$
			return;
		}
		
		File file = getDescriptorFile(cmdAccessor, descriptorFilePathName);
		if(file == null)
			return;
		
		SAXBuilder builder = new SAXBuilder();
		Document doc = null;
		
		try
		{
			doc = builder.build(file);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.getLogger().exception(e);
			return;
		}
		
		Element root = doc.getRootElement();
		Descriptor desc = null;
		
		try
		{
			desc = Descriptor.create(root);
			anchor.setDescriptor(desc);
		}
		catch (XMLParseException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.getLogger().exception(e);
		}
	}

	protected File getDescriptorFile(CommandAccessor cmdAccessor, String descriptorFilePathName)
	{
		File file = new File(descriptorFilePathName);		
		if (!file.exists())
		{			
			String path = null;
			try
			{
				path = file.getCanonicalPath();
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				cmdAccessor.getLogger().exception(e);
				return null;
			}
			if(path != null)
				cmdAccessor.getLogger().fatalText("ADE distributor descriptor " + descriptorFilePathName + " not found! Startup directory: " + path); //$NON-NLS-1$ //$NON-NLS-2$
		}
		return file;
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public StringVector commandNames()
	{
		return new StringVector(StartCommand.QualifiedName);
	}
}
