/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.executor;

import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.almendra.commons.util.collection.ListUtility;
import org.almendra.janet.ade.distributor.DistributorAnchor;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.agents.MultiAgentProxy;
import org.almendra.janet.cas.node.AbstractApplication;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.SystemApplication;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;


public abstract class AbstractAnchor
{
	public static final String LOG_EVENT_NO_DISTRIBUTOR_AGENT_FOUND = "NO_DISTRIBUTOR_AGENT_FOUND"; //$NON-NLS-1$
	
	protected List handlers = new LinkedList();
	protected SystemApplication systemApplication = null;
	protected Logger logger = null;
	protected Map agents = new Hashtable();
	protected ReentrantLock agentsLock = new ReentrantLock();
	
	/**
	 * Creates a new <code>AbstractAnchor</code> object.
	 */
	public AbstractAnchor()
	{
		super();
	}
	
	/**
	 * Creates a new <code>AbstractAnchor</code> object.
	 */
	public AbstractAnchor(SystemApplication systemApplication)
	{
		super();
		this.systemApplication = systemApplication;
	}
	
	public void clearAgentsCache()
	{
		try
		{
			agentsLock.acquire();
			agents = new Hashtable();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			agentsLock.release();
		}
	}
	
	/**
	 * @return Returns the handlers.
	 */
	public List getHandlers()
	{
		return handlers;
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.ade.executor.AbstractAnchor#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		List registeredLogEvents = new LinkedList();
		registeredLogEvents.add(LOG_EVENT_NO_DISTRIBUTOR_AGENT_FOUND);
		return registeredLogEvents;
	}
	
	/**
	 * @return
	 */
	public IAgentProxy getDistributorAgent()
	{
		try
		{
			agentsLock.acquire();
			IAgentProxy distributorAgent = (IAgentProxy) agents.get(DistributorAnchor.BALANCING_AGENT_NAME);
			if(distributorAgent == null)
			{
				distributorAgent = getDistributorAgentUnprotected();
			}
			return distributorAgent;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		finally
		{
			agentsLock.release();
		}
		return null;
	}
	
	protected IAgentProxy getDistributorAgentUnprotected()
	{
		LocalAgentPath agentPath = new LocalAgentPath(SystemApplication.NAME, DistributorAnchor.CAPABILITY_NAME, DistributorAnchor.BALANCING_AGENT_NAME);
		List agentPaths = systemApplication.getAllCrossAgentPaths(agentPath);
		if(agentPaths == null && agentPaths.isEmpty())
		{
			logNoDistributorAgentFound();
			return null;
		}
		IAgentProxy distributorAgent = getDistributorAgent(agentPaths);
		if(distributorAgent != null)
			agents.put(DistributorAnchor.BALANCING_AGENT_NAME, distributorAgent);
		return distributorAgent;
	}
	
	/**
	 * @param agentPaths
	 * @return
	 */
	protected IAgentProxy getDistributorAgent(List agentPaths)
	{
		try
		{
			if(agentPaths.size() == 1)
			{
				AgentPath agentPath = (AgentPath) agentPaths.get(0);
				return systemApplication.getCrossAgent(agentPath);	
			}
			// In case several distributor nodes are present, pick the agent on the node with
			// node name that is alphabetically the first one.
			Map agentPathsByNodeName = new Hashtable();
			Iterator it = agentPaths.iterator();
			while (it.hasNext())
			{
				AgentPath agentPath = (AgentPath) it.next();
				agentPathsByNodeName.put(agentPath.getNodeName(), agentPath);
			}
			List nodeNamesSorted = ListUtility.toList(agentPathsByNodeName.keySet());
			Collections.sort(nodeNamesSorted);
			AgentPath agentPath = (AgentPath) agentPathsByNodeName.get(nodeNamesSorted.get(0));
			return systemApplication.getCrossAgent(agentPath);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}		
		return null;
	}

	/**
	 * 
	 */
	protected void logNoDistributorAgentFound()
	{
		LogEntry entry = (LogEntry) logger.getNewLogEntry();
		entry.setEvent(LOG_EVENT_NO_DISTRIBUTOR_AGENT_FOUND);
		entry.setArgument("no distribution balancer agent found in cluster!"); //$NON-NLS-1$
		logger.fatal(entry);
	}

	/**
	 * @return Returns the systemApplication.
	 */
	public SystemApplication getSystemApplication()
	{
		return systemApplication;
	}
	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}

	/**
	 * @param agentPath
	 * @param app
	 * @return
	 */
	protected IAgentProxy getCrossAgent(AgentPath agentPath, AbstractApplication app)
	{
		try
		{
			return app.getCrossAgent(agentPath);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		return null;
	}
	
	protected List getAgentsList(List agentPaths)
	{
		if(agentPaths.size() == 0)
			return null;
		List agentProxies = new LinkedList();
		Iterator it = agentPaths.iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			IAgentProxy agent = getCrossAgent(agentPath, systemApplication);
			agentProxies.add(agent);
		}
		return agentProxies;
	}
	
	protected Map getAgentsMap(List agentPaths)
	{
		if(agentPaths.size() == 0)
			return null;
		Map agentProxies = new Hashtable();
		Iterator it = agentPaths.iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			IAgentProxy agent = getCrossAgent(agentPath, systemApplication);
			agentProxies.put(agentPath, agent);
		}
		return agentProxies;
	}
	
	protected IAgentProxy getAgents(String keyName, List agentPaths)
	{
		if(agentPaths.size() == 0)
			return null;
		List agentProxies = new LinkedList();
		Iterator it = agentPaths.iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			IAgentProxy agent = getCrossAgent(agentPath, systemApplication);
			agentProxies.add(agent);
		}
		IAgentProxy proxy = new MultiAgentProxy(agentProxies); 
		if(proxy != null)
			agents.put(keyName, proxy);
		return proxy;
	}

	public IAgentProxy getLoadAdministratorAgents()
	{
		try
		{
			agentsLock.acquire();
			IAgentProxy loadAdministratorAgents = (IAgentProxy) agents.get(DistributorAnchor.LOAD_ADMINISTRATOR_AGENT_NAME);
			if(loadAdministratorAgents == null)
			{
				LocalAgentPath agentPath = new LocalAgentPath(
						SystemApplication.NAME, 
						DistributorAnchor.CAPABILITY_NAME, 
						DistributorAnchor.LOAD_ADMINISTRATOR_AGENT_NAME);
				List agentPaths = systemApplication.getAllCrossAgentPaths(agentPath);
				loadAdministratorAgents = getAgents(DistributorAnchor.LOAD_ADMINISTRATOR_AGENT_NAME, agentPaths);
			}
			return loadAdministratorAgents;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		finally
		{
			agentsLock.release();
		}	
		return null;
	}
}
