/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.ade.executor;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Timer;

import org.almendra.commons.events.IVetoableCallbackHandler;
import org.almendra.commons.events.IVetoableEventRegistry;
import org.almendra.commons.number.Long;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.timer.TimerTask;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.CircularIterator;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.MapUtility;
import org.almendra.commons.util.globalId.UniqueIdGenerator;
import org.almendra.commons.util.swing.BillboardPanelController;
import org.almendra.commons.util.swing.TextAreaPanelController;
import org.almendra.janet.ade.commands.ExecuteWorkloadBalancingCommand;
import org.almendra.janet.ade.commands.ExecuteWorkloadCommand;
import org.almendra.janet.ade.commands.ExecuteWorkloadSharingCommand;
import org.almendra.janet.ade.commands.NotifyQSCChangedCommand;
import org.almendra.janet.ade.commands.WorkloadDistributionCommand;
import org.almendra.janet.ade.distributor.DistributorAnchor;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.handlers.AskVetoShutdownEventParams;
import org.almendra.janet.cas.monitor.commands.DisplayAttributeCommand;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.CapabilityPath;
import org.almendra.janet.cas.node.IExecutorAnchor;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LocalCapabilityPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeAnchor;
import org.almendra.janet.cas.node.SystemApplication;
import org.almendra.janet.cas.scheduling.CommandEnvelope;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.IWorkLoadBalancingCommand;
import org.almendra.janet.cas.scheduling.IWorkLoadSharingCommand;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;
import EDU.oswego.cs.dl.util.concurrent.ReentrantWriterPreferenceReadWriteLock;

public class ExecutorAnchor extends AbstractAnchor implements IExecutorAnchor
{
	public static final String INFO_FILE_SUFFIX = "-ADE.Executor.info"; //$NON-NLS-1$
	public static final String LOG_EVENT_INTERRUPTED_REDUCE_LOAD_COMMAND = "INTERRUPTED_REDUCE_LOAD_COMMAND"; //$NON-NLS-1$
	public static final String LOG_EVENT_NO_WAITING_COMMAND_TO_EVICT = "NO_WAITING_COMMAND_TO_EVICT"; //$NON-NLS-1$
	public static final String LOG_EVENT_EVICTED_WAITING_COMMAND = "EVICTED_WAITING_COMMAND"; //$NON-NLS-1$
	public static final String LOG_EVENT_NO_EXECUTING_COMMAND_TO_EVICT = "NO_EXECUTING_COMMAND_TO_EVICT"; //$NON-NLS-1$
	public static final String LOG_EVENT_EVICTED_EXECUTING_COMMAND = "EVICTED_EXECUTING_COMMAND"; //$NON-NLS-1$
	public static final String LOG_EVENT_SUSPENDING_BALANCING_COMMAND = "SUSPENDING_BALANCING_COMMAND"; //$NON-NLS-1$
	public static final String LOG_EVENT_SUSPENDED_BALANCING_COMMAND = "SUSPENDED_BALANCING_COMMAND"; //$NON-NLS-1$
	public static final String LOG_EVENT_INVALID_OBSERVER_ASSIGNMENT = "INVALID_OBSERVER_ASSIGNMENT"; //$NON-NLS-1$
	public static final String LOG_EVENT_OBSERVER_ASSIGNMENT_CHANGED = "OBSERVER_ASSIGNMENT_CHANGED"; //$NON-NLS-1$
	public static final String LOG_EVENT_NEW_QSC = "NEW_QSC"; //$NON-NLS-1$
	public static final String LOG_EVENT_CHANGED_QSC = "DISTRIBUTOR_NOTIFIED_CHANGED_QSC"; //$NON-NLS-1$
	public static final String LOG_EVENT_QSC_LOGEVALUATION = "QSC_LOGEVALUATION"; //$NON-NLS-1$
	public static final String CAPABILITY_NAME = "ADE_EXECUTOR"; //$NON-NLS-1$
	public static final String AGENT_NAME = "EXECUTOR"; //$NON-NLS-1$
	public static final String APPLICATION_NAME = "ADE"; //$NON-NLS-1$
	public static final String BILLBOARD_LABEL = "AD"; //$NON-NLS-1$
	public static final String QUEUE_SIZE_CATEGORY = "AD.QUEUE_SIZE_CATEGORY"; //$NON-NLS-1$
	public static final String WAITING_WORKLOAD_SHARING_COMMANDS = "AD.WAITING_WORKLOAD_SHARING_COMMANDS"; //$NON-NLS-1$
	public static final String WAITING_WORKLOAD_BALANCING_COMMANDS = "AD.WAITING_WORKLOAD_BALANCING_COMMANDS"; //$NON-NLS-1$
	public static final String EXECUTING_WORKLOAD_SHARING_COMMANDS = "AD.EXECUTING_WORKLOAD_SHARING_COMMANDS"; //$NON-NLS-1$
	public static final String EXECUTING_WORKLOAD_BALANCING_COMMANDS = "AD.EXECUTING_WORKLOAD_BALANCING_COMMANDS"; //$NON-NLS-1$
	public static final String WORKLOAD_COMMANDS_WAITING_TIME = "AD.WORKLOAD_COMMANDS_WAITING_TIME"; //$NON-NLS-1$
	public static final String KEY_NEXT_LOAD_SHARING_AGENT = "KEY_NEXT_LOAD_SHARING_AGENT"; //$NON-NLS-1$
	protected ReentrantLock lock = new ReentrantLock();
	protected ReentrantWriterPreferenceReadWriteLock counterLock = new ReentrantWriterPreferenceReadWriteLock();
	protected Descriptor descriptor = null;
	protected int previousTotalCommandsCount = 0;
	protected int totalCommandsCount = 0;
	protected IAgentProxy systemAgent = null;
	protected int currentQueueSizeCategory = 0;
	protected int waitingCommandsCount = 0;
	protected int executingCommandsCount = 0;
	protected int waitingSharingCommandsCount = 0;
	protected int executingSharingCommandsCount = 0;
	protected int waitingBalancingCommandsCount = 0;
	protected int executingBalancingCommandsCount = 0;
	protected List queueSizeCategoriesSorted = null;
	protected Map waitingTimesByCommands = new Hashtable();
	protected Map waitingCommandsCountByCapability = new Hashtable();
	protected Map waitingCommandsByCapability = new Hashtable();
	protected Map executingCommandsCountByCapability = new Hashtable();
	protected Map executingCommandsByCapability = new Hashtable();
	protected long commandsWaitingTime = 0;
	protected Timer timer = null;
	protected List reduceLoadTokens = new LinkedList();
	protected ReentrantWriterPreferenceReadWriteLock reduceLoadTokensLock = new ReentrantWriterPreferenceReadWriteLock();
	protected volatile boolean isEvictionInProcess = false;
	protected volatile boolean isAboveThreshold = false;
	protected volatile String connectedObserverNodeName = null;
	protected Map loadSharingAgentsByAgentPath = new Hashtable();

	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	public ExecutorAnchor(IAgentProxy systemAgent, SystemApplication systemApplication, Logger logger)
	{
		super(systemApplication);
		this.systemAgent = systemAgent;
		setLogger(logger);
		setBillboard();
		writeValue(0);
		registerForEvents();
	}

	/**
	 * 
	 */
	protected void registerForEvents()
	{
		IVetoableEventRegistry vetoableEventRegistry = systemApplication.getNode().getVetoableEventRegistry();
		IVetoableCallbackHandler handler = new AskVetoShutdownCallbackHandler(this);
		vetoableEventRegistry.register(AskVetoShutdownEventParams.EVENT_NAME, handler);
	}

	public static AgentPath getADEAgentPath()
	{
		return new AgentPath(AgentPath.LOCAL_NODE_TAG, SystemApplication.NAME, ExecutorAnchor.CAPABILITY_NAME,
				ExecutorAnchor.AGENT_NAME);
	}

	/**
	 *  
	 */
	protected void setBillboard()
	{
		List mask = new LinkedList();
		mask.add(BillboardPanelController.BLANK_LINE);
		mask.add(QUEUE_SIZE_CATEGORY);
		mask.add(WAITING_WORKLOAD_SHARING_COMMANDS);
		mask.add(WAITING_WORKLOAD_BALANCING_COMMANDS);
		mask.add(EXECUTING_WORKLOAD_SHARING_COMMANDS);
		mask.add(EXECUTING_WORKLOAD_BALANCING_COMMANDS);
		mask.add(WORKLOAD_COMMANDS_WAITING_TIME);
		getBillboard().setMask(BILLBOARD_LABEL, mask);
		getBillboard().setValue(BILLBOARD_LABEL, QUEUE_SIZE_CATEGORY, "0"); //$NON-NLS-1$
		getBillboard().setValue(BILLBOARD_LABEL, WAITING_WORKLOAD_SHARING_COMMANDS, "0"); //$NON-NLS-1$
		getBillboard().setValue(BILLBOARD_LABEL, WAITING_WORKLOAD_BALANCING_COMMANDS, "0"); //$NON-NLS-1$
		getBillboard().setValue(BILLBOARD_LABEL, EXECUTING_WORKLOAD_SHARING_COMMANDS, "0"); //$NON-NLS-1$
		getBillboard().setValue(BILLBOARD_LABEL, EXECUTING_WORKLOAD_BALANCING_COMMANDS, "0"); //$NON-NLS-1$
		getBillboard().setValue(BILLBOARD_LABEL, WORKLOAD_COMMANDS_WAITING_TIME, "0"); //$NON-NLS-1$
	}

	/**
	 * @param logger2
	 */
	protected void setLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setApplicationName(APPLICATION_NAME);
		logger.getDefaultValues().setObjectName(LABEL);
		logger.registerForEvents(registeredLogEvents());
		logger.subscribeForEvents(registeredLogEvents());
	}

	protected void workloadCommandArrived(CommandEnvelope env, String localCapabilityPath)
	{
		assertValid(env);
		AssertUtility.assertTrue(waitingCommandsCount >= 0);
		AssertUtility.assertTrue(totalCommandsCount >= 0);
		waitingCommandsCount++;
		totalCommandsCount++;
		incrementCount(waitingCommandsCountByCapability, localCapabilityPath);
		addCommandEnvelope(waitingCommandsByCapability, env, localCapabilityPath);
		waitingTimesByCommands.put(env, new Long(0));
		checkTimer();
		checkQueueSizeCategories();
	}

	/**
	 * @param waitingCommandsByCapability
	 * @param env
	 * @param localCapabilityPath
	 */
	protected void addCommandEnvelope(Map map, CommandEnvelope env, String localCapabilityPath)
	{
		List envelopes = (List) map.get(localCapabilityPath);
		if (envelopes == null)
		{
			envelopes = new ArrayList();
			map.put(localCapabilityPath, envelopes);
		}
		envelopes.add(env);
	}

	/**
	 * @param waitingCommandsByCapability2
	 * @param env
	 * @param localCapabilityPath
	 */
	protected void removeCommandEnvelope(Map map, CommandEnvelope env, String localCapabilityPath)
	{
		if (env.isSuspended())
			return;
		List envelopes = (List) map.get(localCapabilityPath);
		boolean found = envelopes.remove(env);
		AssertUtility.assertTrue(found);
		if (envelopes.isEmpty())
			map.remove(localCapabilityPath);
	}

	/**
	 * @param string
	 */
	public void workloadSharingCommandArrived(CommandEnvelope env, String localCapabilityPath)
	{
		try
		{
			counterLock.writeLock().acquire();
			AssertUtility.assertTrue(waitingSharingCommandsCount >= 0);
			waitingSharingCommandsCount++;
			workloadCommandArrived(env, localCapabilityPath);
			updateBillboardWaitingSharingCommandsCount();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	protected void addWaitingTimes(long period)
	{
		try
		{
			counterLock.writeLock().acquire();
			Iterator it = waitingTimesByCommands.values().iterator();
			while (it.hasNext())
			{
				Long waitingTime = (Long) it.next();
				waitingTime.add(period);
				commandsWaitingTime += period;
			}
			updateWaitingTime();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 *  
	 */
	protected void updateBillboardWaitingSharingCommandsCount()
	{
		getBillboard().setValue(BILLBOARD_LABEL, WAITING_WORKLOAD_SHARING_COMMANDS,
				String.valueOf(waitingSharingCommandsCount));
	}

	/**
	 *  
	 */
	protected void updateBillboardQSC(int newQueueSizeCategory)
	{
		getBillboard().setValue(BILLBOARD_LABEL, QUEUE_SIZE_CATEGORY, String.valueOf(newQueueSizeCategory));
	}

	/**
	 *  
	 */
	protected void updateBillboardExecutingSharingCommandsCount()
	{
		getBillboard().setValue(BILLBOARD_LABEL, EXECUTING_WORKLOAD_SHARING_COMMANDS,
				String.valueOf(executingSharingCommandsCount));
	}

	/**
	 *  
	 */
	protected void updateWaitingTime()
	{
		getBillboard().setValue(BILLBOARD_LABEL, WORKLOAD_COMMANDS_WAITING_TIME, String.valueOf(commandsWaitingTime));
	}

	protected void workloadCommandStarted(CommandEnvelope env, String localCapabilityPath)
	{
		assertValid(env);
		AssertUtility.assertTrue(executingCommandsCount >= 0);
		waitingCommandsCount--;
		executingCommandsCount++;
		removedFromWaitingQueuePrim(env, localCapabilityPath);
	}

	/**
	 * @param string
	 */
	public void workloadSharingCommandStarted(CommandEnvelope env, String localCapabilityPath)
	{
		try
		{
			counterLock.writeLock().acquire();
			workloadCommandStarted(env, localCapabilityPath);
			AssertUtility.assertTrue(waitingSharingCommandsCount > 0);
			waitingSharingCommandsCount--;
			AssertUtility.assertTrue(executingSharingCommandsCount >= 0);
			executingSharingCommandsCount++;
			checkQueueSizeCategories();
			updateBillboardWaitingSharingCommandsCount();
			updateBillboardExecutingSharingCommandsCount();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 *  
	 */
	public void checkQueueSizeCategories()
	{
		int newQueueSizeCategory = getNewQueueSizeCateogry();
		if (newQueueSizeCategory == -1)
			return;
		int currentQueueSizeCategory = getCurrentQueueSizeCategory();
		if (currentQueueSizeCategory != newQueueSizeCategory)
		{
			setCurrentQueueSizeCategory(newQueueSizeCategory);
			notifiyNewQueueSizeCategory(currentQueueSizeCategory, newQueueSizeCategory);
		}
	}

	/**
	 * @param newQueueSizeCategory
	 */
	protected void notifiyNewQueueSizeCategory(int currentQueueSizeCategory, int newQueueSizeCategory)
	{
		try
		{
			updateBillboardQSC(newQueueSizeCategory);
			monitorQSC(newQueueSizeCategory);
			if (!isConnectedWithAnObserver())
				return;
			logNotifyChangedQSC(currentQueueSizeCategory, newQueueSizeCategory);
			writeValue(newQueueSizeCategory);
			if (isEvictionInProcess())
				return;
			IAgentProxy agents = getLoadAdministratorAgents();
			if (agents == null)
				return;
			NotifyQSCChangedCommand cmd = new NotifyQSCChangedCommand(systemApplication.getNodeName(),
					newQueueSizeCategory, commandsWaitingTime, new Hashtable(waitingCommandsCountByCapability),
					new Hashtable(executingCommandsCountByCapability));
			agents.accept(cmd);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
	}

	/**
	 * @param currentQueueSizeCategory2
	 * @param newQueueSizeCategory
	 */
	public void logNotifyChangedQSC(int currentQueueSizeCategory, int newQueueSizeCategory)
	{
		if (!isLogQSCChanges())
			return;
		if (logger.isSubscribed(LOG_EVENT_CHANGED_QSC))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_CHANGED_QSC);
			entry.setArgument("{previousQSC=" + currentQueueSizeCategory + " | newQSC=" + newQueueSizeCategory + "}"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			logger.info(entry);
		}
	}

	/**
	 * @param newQueueSizeCategory
	 */
	protected void monitorQSC(int newQueueSizeCategory)
	{
		Node node = systemApplication.getNode();
		List monitorAgents = node.getMonitorAgents();
		if (monitorAgents != null)
		{
			DisplayAttributeCommand cmd = new DisplayAttributeCommand(node.getName());
			DisplayAttributeCommand.Entry entry = cmd.new Entry();
			entry.attribute = NodeAnchor.QUEUE_SIZE_CATEGORY;
			entry.value = String.valueOf(newQueueSizeCategory);
			cmd.add(entry);
			node.sendToMonitorAgents(cmd);
		}
	}

	/**
	 *  
	 */
	protected void writeValue(int newQueueSizeCategory)
	{
		String fileName = systemApplication.getNodeName() + INFO_FILE_SUFFIX;
		try
		{
			java.io.FileOutputStream fis = new java.io.FileOutputStream(fileName);
			java.io.ObjectOutputStream ois = new java.io.ObjectOutputStream(fis);
			ois.writeUTF(String.valueOf(newQueueSizeCategory));
			ois.close();
			fis.close();
		}
		catch (Exception e)
		{
			String msg = e.getClass().getName() + ": " + e.getMessage() + " - file: " + fileName; //$NON-NLS-1$ //$NON-NLS-2$
			logger.exception(e);
			System.out.println(msg);
		}
	}

	protected Set getCommands(Map map, String key)
	{
		Set values = (Set) map.get(key);
		if (values == null)
		{
			values = new HashSet();
			map.put(key, values);
		}
		return values;
	}

	/**
	 * @param suspendWaitingCommands
	 */
	public void executeLocked(IExecutable executable)
	{
		try
		{
			lock.acquire();
			executable.execute(this);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
	}

	/**
	 * @param desc
	 */
	public void setDescriptor(Descriptor descriptor)
	{
		this.descriptor = descriptor;
		if (!descriptor.isLogEvaluation())
			logger.removeSubscription(LOG_EVENT_QSC_LOGEVALUATION);
	}

	/**
	 * @return Returns the descriptor.
	 */
	public Descriptor getDescriptor()
	{
		return descriptor;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.node.IExecutorAnchor#sendToSharingDistributor(org.almendra.commons.janet.cas.agents.AgentProxy,
	 *      org.almendra.commons.janet.cas.agents.IAgentProxy,
	 *      org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand)
	 */
	public ICommand getWorkloadDistributionSharingCommandEnvelope(AgentPath initialAgentPath, IWorkLoadSharingCommand sharingCommand, String capabilityPath) 
	{
		ExecuteWorkloadSharingCommand executeCommand = new ExecuteWorkloadSharingCommand(sharingCommand, capabilityPath);
		return new WorkloadDistributionCommand(initialAgentPath, executeCommand);
	}
	
	/*
	 *  (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.node.IExecutorAnchor#getWorkloadDistributionSharingCommandEnvelope(org.almendra.commons.janet.cas.node.AgentPath, org.almendra.commons.janet.cas.agents.IAgentProxy, org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand)
	 */
	public ICommand getWorkloadDistributionBalancingCommandEnvelope(AgentPath initialAgentPath, IWorkLoadBalancingCommand balancingCommand, String capabilityPath)
	{
		ExecuteWorkloadBalancingCommand executeCommand = new ExecuteWorkloadBalancingCommand(balancingCommand, capabilityPath); 
		return new WorkloadDistributionCommand(initialAgentPath, executeCommand);
	}

	public AgentPath getLoadAdministratorAgentPath(IAgentProxy agent)
	{
		return (AgentPath) MapUtility.getKeyOfValue(loadSharingAgentsByAgentPath, agent);
	}

	public IAgentProxy getNextLoadSharingAgent()
	{
		try
		{
			agentsLock.acquire();
			CircularIterator it = (CircularIterator) agents.get(KEY_NEXT_LOAD_SHARING_AGENT);
			if (it == null)
			{
				List agentPaths = getLoadSharingAgentPaths();
				if (agentPaths.size() == 0)
					return null;
				loadSharingAgentsByAgentPath = getAgentsMap(agentPaths);
				if (loadSharingAgentsByAgentPath == null || loadSharingAgentsByAgentPath.isEmpty())
					return null;
				it = new CircularIterator(ListUtility.toList(loadSharingAgentsByAgentPath.values()));
				agents.put(KEY_NEXT_LOAD_SHARING_AGENT, it);
			}
			return (IAgentProxy) it.next();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		finally
		{
			agentsLock.release();
		}
		return null;
	}

	/**
	 * @return
	 */
	protected List getLoadSharingAgentPaths()
	{
		LocalAgentPath agentPath = new LocalAgentPath(SystemApplication.NAME, DistributorAnchor.CAPABILITY_NAME,
				DistributorAnchor.LOAD_ADMINISTRATOR_AGENT_NAME); //$NON-NLS-1$ //$NON-NLS-2$
		return systemApplication.getAllCrossAgentPaths(agentPath);
	}

	/**
	 * @return
	 */
	protected List getBalancerAgentPaths()
	{
		LocalAgentPath agentPath = new LocalAgentPath(SystemApplication.NAME, DistributorAnchor.CAPABILITY_NAME,
				DistributorAnchor.BALANCING_AGENT_NAME); //$NON-NLS-1$ //$NON-NLS-2$
		List agentPaths = systemApplication.getAllCrossAgentPaths(agentPath);
		return agentPaths;
	}

	/**
	 * @return
	 */
	public List getQueueSizeCategories()
	{
		return descriptor.getQueueSizeCategories();
	}

	/**
	 * @return Returns the currentQueueSizeCategory.
	 */
	public int getCurrentQueueSizeCategory()
	{
		return currentQueueSizeCategory;
	}

	/**
	 * @param currentQueueSizeCategory
	 *            The currentQueueSizeCategory to set.
	 */
	public void setCurrentQueueSizeCategory(int currentQueueSizeCategory)
	{
		this.currentQueueSizeCategory = currentQueueSizeCategory;
	}

	/**
	 * @param executingCommandsCount
	 *            The executingCommandsCount to set.
	 */
	public void setExecutingCommandsCount(int executingCommandsCount)
	{
		this.executingCommandsCount = executingCommandsCount;
	}

	/**
	 * @return
	 */
	public List getQueueSizeCategoriesReverseSorted()
	{
		// Sort descending. Category with highest category name first in list.
		if (queueSizeCategoriesSorted == null)
		{
			queueSizeCategoriesSorted = new LinkedList();
			List queueSizeCategories = getDescriptor().getQueueSizeCategories();
			int length = queueSizeCategories.size();
			if (length > 0)
			{
				QueueSizeCategory[] categories = new QueueSizeCategory[length];
				for (int i = 0; i < length; i++)
				{
					QueueSizeCategory category = (QueueSizeCategory) queueSizeCategories.get(i);
					if (category.getName() < length)
						categories[category.getName()] = category;
				}
				queueSizeCategoriesSorted = ListUtility.toList(categories);
				Collections.reverse(queueSizeCategoriesSorted);
			}
		}
		return queueSizeCategoriesSorted;
	}

	protected void workloadCommandFinished(CommandEnvelope env, String localCapabilityPath)
	{
		// AssertUtility.assertTrue(executingCommandsCount > 0);
		// AssertUtility.assertTrue(totalCommandsCount > 0);
		executingCommandsCount--;
		totalCommandsCount--;
	}

	/**
	 *  
	 */
	public void workloadSharingCommandFinished(CommandEnvelope env, String localCapabilityPath)
	{
		try
		{
			counterLock.writeLock().acquire();
			executingSharingCommandsCount--;
			workloadCommandFinished(env, localCapabilityPath);
			AssertUtility.assertTrue(executingSharingCommandsCount >= 0);
			updateBillboardExecutingSharingCommandsCount();
			checkQueueSizeCategories();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 * @param list
	 * @return
	 */
	protected int getNewQueueSizeCateogry()
	{
		List categories = getQueueSizeCategoriesReverseSorted();
		if (categories.isEmpty())
		{
			// TODO - display warning, makes no sense to define no queue size
			// categories
		}
		// Check whether command is just about to be taken out of waiting queue.
		// If so, ignore case and wait till the command will be executin.
		if (waitingCommandsCount > 0 && executingCommandsCount == 0)
			return -1;
		if (waitingCommandsCount == 0 && executingCommandsCount == 0)
			return 0;
		Iterator it = categories.iterator();
		int matchingCategory = categories.size();
		while (it.hasNext())
		{
			QueueSizeCategory category = (QueueSizeCategory) it.next();
			if (category.match(waitingCommandsCount, executingCommandsCount))
			{
				matchingCategory = category.getName();
			}
			else
			{
				logNewQSC(matchingCategory, waitingCommandsCount, executingCommandsCount);
				return matchingCategory;
			}
		}
		logNewQSC(matchingCategory, waitingCommandsCount, executingCommandsCount);
		return matchingCategory;
	}

	/**
	 * @param matchingCategory
	 * @param i
	 * @param j
	 */
	public void logNewQSC(int matchingCategory, int waitingCommandsCount, int executingCommandsCount)
	{
		if (!isLogQSCChanges())
			return;
		if (logger.isSubscribed(LOG_EVENT_NEW_QSC))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_NEW_QSC);
			entry.setArgument("{newQSC=" + matchingCategory + " | waitingCommandsCount=" + waitingCommandsCount
					+ " | executingCommandsCount=" + executingCommandsCount + "}"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
			logger.info(entry);
		}
	}

	protected BillboardPanelController getBillboard()
	{
		return systemApplication.getNode().getBillboard();
	}

	protected TextAreaPanelController getFreeTextArea()
	{
		return systemApplication.getNode().getFreeTextArea();
	}

	/**
	 * @param cmdAccessor
	 */
	protected void checkTimer()
	{
		if (waitingSharingCommandsCount > 0 && timer == null)
		{
			timer = new Timer(true);
			long period = getDescriptor().getPeriod();
			IExecutable executable = new ObserveWaitingCommandsExecutable(this, period);
			TimerTask task = new TimerTask(executable);
			timer.scheduleAtFixedRate(task, getDescriptor().getDelay(), period);
			return;
		}
		if (waitingSharingCommandsCount == 0 && timer != null)
		{
			timer.cancel();
			timer = null;
			return;
		}
	}

	/**
	 *  
	 */
	public void removeQueueCategorySizeFile()
	{
		String fileName = systemApplication.getNodeName() + INFO_FILE_SUFFIX;
		File file = new File(fileName);
		if (file.exists())
			file.delete();
	}

	protected void incrementCount(Map capabilityMap, String quailifiedLocalCapabilityPath)
	{
		Long number = (Long) capabilityMap.get(quailifiedLocalCapabilityPath);
		if (number == null)
		{
			number = new Long();
			capabilityMap.put(quailifiedLocalCapabilityPath, number);
		}
		number.increment();
	}

	protected void decrementCount(Map capabilityMap, String quailifiedLocalCapabilityPath)
	{
		Long number = (Long) capabilityMap.get(quailifiedLocalCapabilityPath);
		AssertUtility.assertTrue(number.longValue() > 0);
		number.decrement();
		if (number.isZero())
			capabilityMap.remove(quailifiedLocalCapabilityPath);
	}

	/**
	 * @param cmd
	 * @param capabilityPath
	 */
	public void workloadBalancingCommandStarted(CommandEnvelope env, String localCapabilityPath)
	{
		try
		{
			counterLock.writeLock().acquire();
			workloadCommandStarted(env, localCapabilityPath);
			AssertUtility.assertTrue(waitingBalancingCommandsCount > 0);
			waitingBalancingCommandsCount--;
			AssertUtility.assertTrue(executingBalancingCommandsCount >= 0);
			executingBalancingCommandsCount++;
			addCommandEnvelope(executingCommandsByCapability, env, localCapabilityPath);
			incrementCount(executingCommandsCountByCapability, localCapabilityPath);
			checkQueueSizeCategories();
			updateBillboardWaitingBalancingCommandsCount();
			updateBillboardExecutingBalancingCommandsCount();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 * @param cmd
	 * @param capabilityPath
	 */
	public void workloadBalancingCommandFinished(CommandEnvelope env, String localCapabilityPath)
	{
		try
		{
			assertValid(env);
			counterLock.writeLock().acquire();
			workloadCommandFinished(env, localCapabilityPath);
			executingBalancingCommandsCount--;
			// AssertUtility.assertTrue(executingBalancingCommandsCount
			// >= 0);
			removeCommandEnvelope(executingCommandsByCapability, env, localCapabilityPath);
			decrementCount(executingCommandsCountByCapability, localCapabilityPath);
			updateBillboardExecutingBalancingCommandsCount();
			checkQueueSizeCategories();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 * @param command
	 * @param string
	 */
	public void evictedWaitingCommand(CommandEnvelope env, String capabilityPath)
	{
		try
		{
			counterLock.writeLock().acquire();
			boolean kindDefined = false;
			waitingCommandsCount--;
			removedFromWaitingQueuePrim(env, capabilityPath);
			monitorEvictedWaitingCommand(env);
			if (env.getKind() == CommandEnvelope.COMMAND_TYPE_BALANCING)
			{
				waitingBalancingCommandsCount--;
				updateBillboardWaitingBalancingCommandsCount();
				kindDefined = true;
				return;
			}
			if (env.getKind() == CommandEnvelope.COMMAND_TYPE_SHARING)
			{
				waitingSharingCommandsCount--;
				updateBillboardWaitingSharingCommandsCount();
				kindDefined = true;
				return;
			}
			AssertUtility.assertTrue("command envelope kind undefined", kindDefined); //$NON-NLS-1$											
			checkQueueSizeCategories();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 * @param command
	 * @param string
	 */
	public void evictedExecutingCommand(CommandEnvelope env, String capabilityPath)
	{
		if (env.getKind() == CommandEnvelope.COMMAND_TYPE_BALANCING)
		{
			workloadBalancingCommandFinished(env, capabilityPath);
			return;
		}
		if (env.getKind() == CommandEnvelope.COMMAND_TYPE_SHARING)
		{
			workloadSharingCommandFinished(env, capabilityPath);
			return;
		}
		AssertUtility.assertTrue("command envelope kind undefined", false); //$NON-NLS-1$
	}

	/**
	 * @param command
	 * @param string
	 */
	public void workloadBalancingCommandArrived(CommandEnvelope env, String capabilityPath)
	{
		try
		{
			assertValid(env);
			counterLock.writeLock().acquire();
			waitingBalancingCommandsCount++;
			workloadCommandArrived(env, capabilityPath);
			AssertUtility.assertTrue(waitingBalancingCommandsCount >= 0);
			updateBillboardWaitingBalancingCommandsCount();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	/**
	 * @param env
	 */
	private void assertValid(CommandEnvelope env)
	{
		org.almendra.commons.util.AssertUtility.assertTrue(env.getCommand() instanceof ExecuteWorkloadCommand);
	}

	/**
	 *  
	 */
	protected void updateBillboardExecutingBalancingCommandsCount()
	{
		getBillboard().setValue(BILLBOARD_LABEL, EXECUTING_WORKLOAD_BALANCING_COMMANDS,
				String.valueOf(executingBalancingCommandsCount));
	}

	/**
	 *  
	 */
	protected void updateBillboardWaitingBalancingCommandsCount()
	{
		getBillboard().setValue(BILLBOARD_LABEL, WAITING_WORKLOAD_BALANCING_COMMANDS,
				String.valueOf(waitingBalancingCommandsCount));
	}

	public boolean writeProtectedDo(IExecutable executable)
	{
		try
		{
			counterLock.writeLock().acquire();
			executable.execute(this);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return false;
		}
		finally
		{
			counterLock.writeLock().release();
		}
		return true;
	}

	public void removeToken(String token)
	{
		try
		{
			reduceLoadTokensLock.writeLock().acquire();
			reduceLoadTokens.remove(token);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			reduceLoadTokensLock.writeLock().release();
		}
	}

	public void removeTokens()
	{
		try
		{
			reduceLoadTokensLock.writeLock().acquire();
			reduceLoadTokens = new LinkedList();
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			reduceLoadTokensLock.writeLock().release();
		}
	}

	public boolean isEvictionInProcess()
	{
		return isEvictionInProcess;
	}

	public void setEvictionInProcess(boolean isEvictionInProcess)
	{
		this.isEvictionInProcess = isEvictionInProcess;
	}

	public String getInsertNewToken()
	{
		try
		{
			reduceLoadTokensLock.writeLock().acquire();
			String newToken = UniqueIdGenerator.next().getValue();
			reduceLoadTokens = new LinkedList();
			reduceLoadTokens.add(newToken);
			return newToken;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			reduceLoadTokensLock.writeLock().release();
		}
		return null;
	}

	public boolean existsToken(String token)
	{
		try
		{
			reduceLoadTokensLock.readLock().acquire();
			return reduceLoadTokens.contains(token);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			reduceLoadTokensLock.readLock().release();
		}
		return false;
	}

	/**
	 * @param string
	 */
	public void logInterruptedReduceLoadCommand(String commandId)
	{
		if (logger.isSubscribed(LOG_EVENT_INTERRUPTED_REDUCE_LOAD_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_INTERRUPTED_REDUCE_LOAD_COMMAND);
			entry.setArgument("{id=" + commandId + "}"); //$NON-NLS-1$ //$NON-NLS-2$
			logger.info(entry);
		}
	}

	/**
	 * @param string
	 */
	public void logNoWaitingCommandToEvict()
	{
		if (logger.isSubscribed(LOG_EVENT_NO_WAITING_COMMAND_TO_EVICT))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_NO_WAITING_COMMAND_TO_EVICT);
			logger.info(entry);
		}
	}

	/**
	 * @param string
	 */
	public void logNoExecutingCommandToEvict()
	{
		if (logger.isSubscribed(LOG_EVENT_NO_EXECUTING_COMMAND_TO_EVICT))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_NO_EXECUTING_COMMAND_TO_EVICT);
			logger.info(entry);
		}
	}

	/**
	 * @param string
	 */
	public void logEvictedExecutingCommand(String commandLogString, AgentPath evictionAgentPath)
	{
		if (logger.isSubscribed(LOG_EVENT_EVICTED_EXECUTING_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_EVICTED_EXECUTING_COMMAND);
			entry.setArgument("{command=" + commandLogString + " to agent " + evictionAgentPath.getQualifiedPathName()
					+ "}"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			logger.info(entry);
		}
	}

	public List registeredLogEvents()
	{
		List events = super.registeredLogEvents();
		events.add(LOG_EVENT_INTERRUPTED_REDUCE_LOAD_COMMAND);
		events.add(LOG_EVENT_NO_WAITING_COMMAND_TO_EVICT);
		events.add(LOG_EVENT_NO_EXECUTING_COMMAND_TO_EVICT);
		events.add(LOG_EVENT_EVICTED_WAITING_COMMAND);
		events.add(LOG_EVENT_EVICTED_EXECUTING_COMMAND);
		events.add(LOG_EVENT_SUSPENDING_BALANCING_COMMAND);
		events.add(LOG_EVENT_SUSPENDED_BALANCING_COMMAND);
		events.add(LOG_EVENT_INVALID_OBSERVER_ASSIGNMENT);
		events.add(LOG_EVENT_OBSERVER_ASSIGNMENT_CHANGED);
		events.add(LOG_EVENT_QSC_LOGEVALUATION);
		events.add(LOG_EVENT_NEW_QSC);
		events.add(LOG_EVENT_CHANGED_QSC);
		return events;
	}

	/**
	 * @return Returns the executingCommandsByCapability.
	 */
	public Map getExecutingCommandsByCapability()
	{
		return executingCommandsByCapability;
	}

	/**
	 * @return Returns the waitingCommandsByCapability.
	 */
	public Map getWaitingCommandsByCapability()
	{
		return waitingCommandsByCapability;
	}

	/**
	 * @param env
	 */
	protected void removedFromWaitingQueuePrim(CommandEnvelope env, String localCapabilityPath)
	{
		decrementCount(waitingCommandsCountByCapability, localCapabilityPath);
		removeCommandEnvelope(waitingCommandsByCapability, env, localCapabilityPath);
		Long commandTotalWaitingTime = (Long) waitingTimesByCommands.remove(env);
		commandsWaitingTime -= commandTotalWaitingTime.longValue();
		checkTimer();
		updateWaitingTime();
	}

	/**
	 * @param string
	 */
	public void logEvictedWaitingCommand(String commandLogString)
	{
		if (logger.isSubscribed(LOG_EVENT_EVICTED_WAITING_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_EVICTED_WAITING_COMMAND);
			entry.setArgument("{command=" + commandLogString + "}"); //$NON-NLS-1$ //$NON-NLS-2$
			logger.info(entry);
		}
	}

	/**
	 * @param string
	 */
	public void logEvictedSuspendedExecutingCommand(String commandLogString)
	{
		if (logger.isSubscribed(LOG_EVENT_SUSPENDED_BALANCING_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_SUSPENDED_BALANCING_COMMAND);
			entry.setArgument("{command=" + commandLogString + "}"); //$NON-NLS-1$ //$NON-NLS-2$
			logger.info(entry);
		}
	}

	public void logInvalidObserverAssignment(String argument)
	{
		if (logger.isSubscribed(LOG_EVENT_INVALID_OBSERVER_ASSIGNMENT))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_INVALID_OBSERVER_ASSIGNMENT);
			entry.setArgument(argument);
			logger.error(entry);
		}
	}

	public void logObserverAssignmentChanged(String argument)
	{
		if (logger.isSubscribed(LOG_EVENT_OBSERVER_ASSIGNMENT_CHANGED))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_OBSERVER_ASSIGNMENT_CHANGED);
			entry.setArgument(argument);
			logger.info(entry);
		}
	}

	/**
	 * @param string
	 */
	public void logSuspendingBalancingCommand(String commandLogString)
	{
		if (logger.isSubscribed(LOG_EVENT_SUSPENDING_BALANCING_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_SUSPENDING_BALANCING_COMMAND);
			entry.setArgument("{command=" + commandLogString + "}"); //$NON-NLS-1$ //$NON-NLS-2$
			logger.info(entry);
		}
	}

	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}

	/**
	 *  
	 */
	public void monitorEvictedExecutingCommand(CommandEnvelope env)
	{
		Node node = systemApplication.getNode();
		List agents = node.getMonitorAgents();
		if (agents == null)
			return;
		ICommand command = node.getMonitorProducerAnchor().getEvictedExecutingCommandNotification(node.getName(), env);
		node.sendToMonitorAgents(command);
	}

	/**
	 *  
	 */
	public void monitorEvictedWaitingCommand(CommandEnvelope env)
	{
		Node node = systemApplication.getNode();
		List agents = node.getMonitorAgents();
		if (agents == null)
			return;
		ICommand command = node.getMonitorProducerAnchor().getEvictedWaitingCommandNotification(node.getName(), env);
		node.sendToMonitorAgents(command);
	}

	/**
	 * @return
	 */
	protected List getLoadAdministratorAgentPaths()
	{
		LocalAgentPath agentPath = new LocalAgentPath(SystemApplication.NAME, DistributorAnchor.CAPABILITY_NAME,
				DistributorAnchor.LOAD_ADMINISTRATOR_AGENT_NAME); //$NON-NLS-1$ //$NON-NLS-2$
		List agentPaths = systemApplication.getAllCrossAgentPaths(agentPath);
		return agentPaths;
	}

	public IAgentProxy getLoadSharingAgents()
	{
		try
		{
			agentsLock.acquire();
			IAgentProxy loadSharingAgents = (IAgentProxy) agents.get(DistributorAnchor.SHARING_AGENT_NAME);
			if (loadSharingAgents == null)
			{
				loadSharingAgents = getAgents(DistributorAnchor.SHARING_AGENT_NAME, getLoadSharingAgentPaths());
			}
			return loadSharingAgents;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		finally
		{
			agentsLock.release();
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.node.IExecutorAnchor#isConnectedWithAnObserver()
	 */
	public boolean isConnectedWithAnObserver()
	{
		return StringUtility.isValid(connectedObserverNodeName);
	}

	/**
	 * @return Returns the connectedObserverNodeName.
	 */
	public String getConnectedObserverNodeName()
	{
		return connectedObserverNodeName;
	}

	protected void connectObserverNodeName(String connectedObserverNodeName)
	{
		if (this.connectedObserverNodeName != null && !this.connectedObserverNodeName.equals(connectedObserverNodeName))
		{
			logInvalidObserverAssignment("executor still connected with observer " + this.connectedObserverNodeName //$NON-NLS-1$
					+ "! Connection with " + connectedObserverNodeName + " refused!"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			return;
		}
		logObserverAssignmentChanged("connected with observer " + connectedObserverNodeName); //$NON-NLS-1$
		this.connectedObserverNodeName = connectedObserverNodeName;
	}

	protected void disconnectObserverNodeName(String connectedObserverNodeName)
	{
		if (this.connectedObserverNodeName == null)
		{
			logInvalidObserverAssignment("executor already disconnected from observer!"); //$NON-NLS-1$
			return;
		}
		if (this.connectedObserverNodeName != null && !this.connectedObserverNodeName.equals(connectedObserverNodeName))
		{
			logInvalidObserverAssignment("executor still connected with observer " + this.connectedObserverNodeName //$NON-NLS-1$
					+ "! Connection with observer  " + connectedObserverNodeName + " refused!"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			return;
		}
		logObserverAssignmentChanged("disconnected from observer " + this.connectedObserverNodeName); //$NON-NLS-1$
		this.connectedObserverNodeName = connectedObserverNodeName;
	}

	/**
	 * @param connectedObserverNodeName
	 *            The connectedObserverNodeName to set.
	 */
	public void setConnectedObserverNodeName(boolean isConnect, String connectedObserverNodeName)
	{
		if (isConnect)
			connectObserverNodeName(connectedObserverNodeName);
		else
			disconnectObserverNodeName(connectedObserverNodeName);
	}

	protected boolean isLogQSCChanges()
	{
		return descriptor.isLogEvaluation();
	}

	/**
	 * @param path
	 * @return
	 */
	public String getLeastLoadedAgentName(LocalCapabilityPath path)
	{
		CapabilityPath capabilityPath = new CapabilityPath(systemApplication.getNodeName(), path.getApplicationName(),
				path.getCapabilityName());
		List agentPaths = systemApplication.getAllAgentPaths(capabilityPath);
		return systemApplication.getAgentNameShortestQueue(agentPaths);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.ade.executor.AbstractAnchor#clearAgentsCache()
	 */
	public void clearAgentsCache()
	{
		super.clearAgentsCache();
		loadSharingAgentsByAgentPath = new Hashtable();
	}

	/**
	 * @return Returns the isAboveThreshold.
	 */
	public boolean isAboveThreshold()
	{
		return isAboveThreshold;
	}

	/**
	 * @param isAboveThreshold
	 *            The isAboveThreshold to set.
	 */
	public void setAboveThreshold(boolean isAboveThreshold)
	{
		this.isAboveThreshold = isAboveThreshold;
	}

	/**
	 * @return
	 */
	public String getNodeName()
	{
		return systemApplication.getNodeName();
	}

	public void suspendExecutingCommands()
	{
		try
		{
			counterLock.writeLock().acquire();
			Iterator it = executingCommandsByCapability.values().iterator();
			while (it.hasNext())
			{
				List commandEnvs = (List) it.next();
				Iterator innerIt = commandEnvs.iterator();
				while (innerIt.hasNext())
				{
					CommandEnvelope env = (CommandEnvelope) innerIt.next();
					env.invokeOnCommand("suspend"); //$NON-NLS-1$
				}								
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}

	public void resumeExecutingCommands()
	{
		try
		{
			counterLock.writeLock().acquire();
			Iterator it = executingCommandsByCapability.values().iterator();
			while (it.hasNext())
			{
				List commandEnvs = (List) it.next();
				Iterator innerIt = commandEnvs.iterator();
				while (innerIt.hasNext())
				{
					CommandEnvelope env = (CommandEnvelope) innerIt.next();
					env.invokeOnCommand("resume"); //$NON-NLS-1$
				}		
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			counterLock.writeLock().release();
		}
	}
	
	public ICommand getExecutebalancingCommand(IWorkLoadBalancingCommand command, String capabilityPath)
	{
		return new ExecuteWorkloadBalancingCommand(command, capabilityPath);
	}
	
}