/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.ade.observer;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Timer;

import org.almendra.commons.events.IVetoableCallbackHandler;
import org.almendra.commons.events.IVetoableEventRegistry;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.timer.TimerTask;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.swing.TextAreaPanelController;
import org.almendra.janet.ade.commands.NotifyCPULoadChangedCommand;
import org.almendra.janet.ade.commands.ObserveWorkstationLoadCommand;
import org.almendra.janet.ade.commands.ObserverConnectionStatusChangedCommand;
import org.almendra.janet.ade.commands.ObserverStateChangedCommand;
import org.almendra.janet.ade.distributor.DistributorAnchor;
import org.almendra.janet.ade.executor.AbstractAnchor;
import org.almendra.janet.ade.executor.ExecutorAnchor;
import org.almendra.janet.ade.observer.gui.IDisplayController;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.handlers.AskVetoShutdownEventParams;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.node.SystemApplication;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.SendCommandExecutable;


public class ObserverAnchor extends AbstractAnchor
{
	public static final String LABEL = "ADE_LOAD_OBSERVER_ANCHOR"; //$NON-NLS-1$
	public static final String CAPABILITY_NAME = "ADE_LOAD_OBSERVER"; //$NON-NLS-1$
	public static final String AGENT_NAME = "OBSERVER"; //$NON-NLS-1$
	public static final String VIEW_TITLE_PREFIX = "AD.Observer for"; //$NON-NLS-1$
	public static final String NODE_LABEL = "AD.Observer"; //$NON-NLS-1$
	public static final String ALL_OTHERS_LABEL = "AllOthers"; //$NON-NLS-1$
	protected Descriptor descriptor = null;
	protected TextAreaPanelController textArea = null;
	protected IAgentProxy observerAgent = null;
	protected Timer timer = null;
	protected IDisplayController displayController = null;
	protected int significantWhenChangeExceeding = -1;
	protected int evictionThreshold = -1;
	protected List nodeNames = new LinkedList();
	protected volatile int nonJanetCPULoad = 0;
	protected String workstationName = null;
	protected Object nodeNamesLock = new Object();

	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	protected ObserverAnchor()
	{
		super();
	}

	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	public ObserverAnchor(SystemApplication systemApplication, Logger logger)
	{
		super(systemApplication);
		setLogger(logger);
		registerForEvents();
	}

	/**
	 * 
	 */
	protected void registerForEvents()
	{
		IVetoableEventRegistry vetoableEventRegistry = systemApplication.getNode().getVetoableEventRegistry();
		IVetoableCallbackHandler handler = new AskVetoShutdownCallbackHandler(this);
		vetoableEventRegistry.register(AskVetoShutdownEventParams.EVENT_NAME, handler);
	}

	/**
	 * 
	 */
	public void notifyWorkstationName()
	{
		AssertUtility.assertTrue(workstationName != null);
		
		try
		{
			IAgentProxy loadAdministratorAgents = getLoadAdministratorAgents();
			if (loadAdministratorAgents == null)
				return;
			ICommand cmd = new ObserverStateChangedCommand(
				workstationName,
				getNodeNames(),
				ObserverStateChangedCommand.Up);
			loadAdministratorAgents.accept(cmd);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
	}

	/**
	 * @param logger
	 */
	protected void setLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.registerForEvents(registeredLogEvents());
		logger.subscribeForEvents(registeredLogEvents());
	}

	/**
	 * @param desc
	 */
	public void setDescriptor(Descriptor descriptor)
	{
		this.descriptor = descriptor;
	}

	/**
	 * @return Returns the descriptor.
	 */
	public Descriptor getDescriptor()
	{
		return descriptor;
	}

	/**
	 * @param proxy
	 */
	public void startTimer(IAgentProxy observerAgent)
	{
		this.observerAgent = observerAgent;
		cancelTimer();
		timer = new Timer(true);
		ICommand command = new ObserveWorkstationLoadCommand(this);
		IExecutable executable = new SendCommandExecutable(observerAgent, command, logger);
		TimerTask task = new TimerTask(executable);
		timer.scheduleAtFixedRate(task, getDescriptor().getPeriod(), getDescriptor().getPeriod());
	}

	/**
	 * @param textArea
	 *            The textArea to set.
	 */
	public void setTextArea(TextAreaPanelController textArea)
	{
		this.textArea = textArea;
	}

	/**
	 *  
	 */
	public void cancelTimer()
	{
		if (timer != null)
		{
			timer.cancel();
			timer = null;
		}
	}

	/**
	 * @param loadTable
	 */
	public void displayValues(Map loadTable)
	{
		if (displayController != null)
		{
			Map values = new Hashtable();
			Iterator it = loadTable.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				Integer value = (Integer) entry.getValue();
				values.put(entry.getKey(), value.toString());
			}
			displayController.displayValues(values);
		}
	}

	/**
	 * @param displayController
	 *            The displayController to set.
	 */
	public void setDisplayController(IDisplayController displayController)
	{
		this.displayController = displayController;
	}

	/**
	 * @param currentAllOtherCPULoad
	 * @param newAllOtherCPULoad
	 */
	public void notifyDistributorAgent(int currentAllOtherCPULoad, int newAllOtherCPULoad, int cpuLoadAboveOrBelow)
	{
		try
		{
			IAgentProxy loadAdministratorAgents = getLoadAdministratorAgents();
			if (loadAdministratorAgents == null)
				return;
			ICommand cmd = new NotifyCPULoadChangedCommand(
					workstationName, 
					currentAllOtherCPULoad,
					newAllOtherCPULoad, 
					cpuLoadAboveOrBelow);
			loadAdministratorAgents.accept(cmd);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
	}

	/**
	 * @return Returns the significantWhenChangeExceeding.
	 */
	public int getSignificantWhenChangeExceeding()
	{
		if (significantWhenChangeExceeding == -1)
		{
			significantWhenChangeExceeding = descriptor.getSignificantWhenChangeExceeding();
			if(significantWhenChangeExceeding == -1)
				significantWhenChangeExceeding = 10;
		}
		return significantWhenChangeExceeding;
	}
	
	public int getEvictionThreshold()
	{
		if (evictionThreshold == -1)
		{
			evictionThreshold = descriptor.getEvictionThreshold();
			if(evictionThreshold == -1)
				evictionThreshold = 30;
		}
		return evictionThreshold;
	}

	/**
	 * @return Returns the allNodeNamesOnWorkstation.
	 */
	public List getNodeNames()
	{
		synchronized(nodeNamesLock)
		{
			return nodeNames;
		}
	}

	/**
	 * @return Returns the nonJanetCPULoad.
	 */
	public int getNonJanetCPULoad()
	{
		return nonJanetCPULoad;
	}

	/**
	 * @param allNodeNamesOnWorkstation
	 *            The allNodeNamesOnWorkstation to set.
	 */
	public void setNodeNames(List nodes)
	{
		synchronized(nodeNamesLock)
		{
			this.nodeNames = new LinkedList(nodes);
		}
	}

	/**
	 * @param nonJanetCPULoad
	 *            The nonJanetCPULoad to set.
	 */
	public void setNonJanetCPULoad(int nonJanetCPULoad)
	{
		this.nonJanetCPULoad = nonJanetCPULoad;
	}
	/**
	 * @return Returns the workstationName.
	 */
	public String getWorkstationName()
	{
		return workstationName;
	}
	/**
	 * @param workstationName The workstationName to set.
	 */
	public void setWorkstationName(String workstationName)
	{
		this.workstationName = workstationName;		
	}

	/**
	 * @param nodeName
	 */
	public String notifyNodeAdded(String nodeName)
	{
		synchronized(nodeNamesLock)
		{
			nodeNames.add(nodeName);
		}
		
		try
		{
			IAgentProxy loadAdministratorAgents = getLoadAdministratorAgents();
			if (loadAdministratorAgents == null)
				return "Assignment change not possible: no distributor present!"; //$NON-NLS-1$
			
			ObserverStateChangedCommand commandToDistributor = new ObserverStateChangedCommand(
				workstationName,
				ObserverStateChangedCommand.NodeAdded);
			commandToDistributor.setNodeName(nodeName);
			loadAdministratorAgents.accept(commandToDistributor);
			IAgentProxy executorAnchor = getExecutorAgent(nodeName);
			ObserverConnectionStatusChangedCommand commandToExecutor = new ObserverConnectionStatusChangedCommand(systemApplication.getNodeName(), true);
			executorAnchor.accept(commandToExecutor);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		return "";
	}
	
	/**
	 * @param nodeName
	 * @return
	 */
	protected IAgentProxy getExecutorAgent(String nodeName) throws NoSuchAgentException, IllegalAgentAccessException
	{
		AgentPath agentPath = new AgentPath(nodeName, SystemApplication.NAME, ExecutorAnchor.CAPABILITY_NAME, ExecutorAnchor.AGENT_NAME); 
		return systemApplication.getAgent(agentPath);
	}

	/**
	 * @param nodeName
	 */
	public String notifyNodeRemoved(String nodeName)
	{
		synchronized(nodeNamesLock)
		{
			nodeNames.remove(nodeName);
		}
		
		try
		{
			IAgentProxy loadAdministratorAgents = getLoadAdministratorAgents();
			if (loadAdministratorAgents == null)
				return "Assignment change not possible: no distributor present!"; //$NON-NLS-1$
			
			ObserverStateChangedCommand command = new ObserverStateChangedCommand(
					workstationName,
					ObserverStateChangedCommand.NodeRemoved);
			command.setNodeName(nodeName);
			loadAdministratorAgents.accept(command);
			IAgentProxy executorAnchor = getExecutorAgent(nodeName);
			ObserverConnectionStatusChangedCommand commandToExecutor = new ObserverConnectionStatusChangedCommand(systemApplication.getNodeName(), false);
			executorAnchor.accept(commandToExecutor);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		return "";
	}

	/**
	 * @param string
	 */
	public void checkRegisteredNodeIsDistributor(String nodeName)
	{
		if(systemApplication.hasCapability(nodeName, SystemApplication.NAME, DistributorAnchor.CAPABILITY_NAME))
			notifyWorkstationName();
	}
}