/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.monitor.interpreters;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.almendra.commons.events.IEventHandler;
import org.almendra.commons.util.collection.StringVector;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.events.EventHandlerEnvelope;
import org.almendra.janet.cas.events.HandlerList;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.handlers.NodeShutdownParams;
import org.almendra.janet.cas.monitor.MonitorProducerAnchor;
import org.almendra.janet.cas.monitor.commands.DisplayAttributeCommand;
import org.almendra.janet.cas.monitor.commands.InstallProducerCommand;
import org.almendra.janet.cas.monitor.commands.SetMaskCommand;
import org.almendra.janet.cas.monitor.handlers.NodeShutdownHandler;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.ApplicationAccessor;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.scheduling.CommandAccessor;
import org.almendra.janet.cas.scheduling.IInterpreter;



public class InstallProducerInterpreter extends AbstractProducerInterpreter implements IInterpreter
{
	/**
	 * Creates a new <code>InstallProducerInterpreter</code> object.
	 */
	public InstallProducerInterpreter()
	{
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#execute(org.almendra.commons.janet.cas.scheduling.CommandAccessor)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if (cmdAccessor.getCommand() instanceof InstallProducerCommand)
		{
			InstallProducerCommand cmd = (InstallProducerCommand) cmdAccessor.getCommand();
			AgentPath consumerAgentPath = cmd.getAgentPath();
			MonitorProducerAnchor anchor = getAnchor(cmdAccessor);			
			
			if(isAlreadyInstalled(anchor, consumerAgentPath))
				return;			
			
			IAgentProxy consumerAgent = getConsumerAgent(cmdAccessor, consumerAgentPath);			
			if(consumerAgent == null)
				return;
						
			anchor.getConsumerAgentPaths().put(consumerAgentPath.getQualifiedPathName(), consumerAgent);
			
			if(!sendNodeMask(cmdAccessor, consumerAgent))
				return;
			
			cmdAccessor.getNodeIfPrivilege().addMonitorAgent(consumerAgent);			
			registerShutdownHandler(cmdAccessor, consumerAgent, consumerAgentPath);
		}
	}

	/**
	 * @param consumerAgentPath
	 */
	protected void registerShutdownHandler(CommandAccessor cmdAccessor, IAgentProxy consumerAgent, AgentPath consumerAgentPath)
	{
		ILocalEventRegistry registry = cmdAccessor.getNodeEventRegistry();
		AgentPath agentPath = cmdAccessor.getAgentPath();
		ApplicationAccessor app = cmdAccessor.getApplication();		
		IAgentProxy localAgent = cmdAccessor.getAgentProxy();
		
		IEventHandler handler = new NodeShutdownHandler(cmdAccessor.getNodeAccessor().getName(),localAgent, consumerAgent, consumerAgentPath);
		EventHandlerEnvelope env = new EventHandlerEnvelope(handler, agentPath);
		HandlerList handlerList = new HandlerList(env);
		registry.register(NodeShutdownParams.EVENT_NAME, handlerList, app);
		
		MonitorProducerAnchor anchor = getAnchor(cmdAccessor);
		anchor.getShutdownHandlers().put(consumerAgentPath.getQualifiedPathName(), handlerList);
	}

	/**
	 * @param cmdAccessor
	 * @param consumerAgent
	 */
	protected boolean sendNodeMask(CommandAccessor cmdAccessor, IAgentProxy consumerAgent)
	{
		try
		{
			String nodeName = cmdAccessor.getNodeAccessor().getName();
			Map monitorMaskDefaultValues = cmdAccessor.getApplication().getMonitorMaskDefaultValues();
			SetMaskCommand setMaskCmd = new SetMaskCommand(cmdAccessor.getApplication().getMonitorMask(), nodeName);
			DisplayAttributeCommand attrCmd = new DisplayAttributeCommand(nodeName, getDefaultValues(nodeName, monitorMaskDefaultValues));
			consumerAgent.accept(setMaskCmd);
			consumerAgent.accept(attrCmd);
			return true;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.log(e);
		}
		return false;
	}

	/**
	 * @param nodeName
	 * @param entries
	 * @param attrCmd
	 */
	protected List getDefaultValues(String nodeName, Map monitorMaskDefaultValues)
	{
		List entries = new LinkedList();
		DisplayAttributeCommand attrCmd = new DisplayAttributeCommand(nodeName);
		
		Iterator it = monitorMaskDefaultValues.entrySet().iterator();
		while (it.hasNext())
		{
			DisplayAttributeCommand.Entry entry = attrCmd.new Entry();
			Map.Entry mapEntry = (Map.Entry) it.next();
			entry.attribute = (String) mapEntry.getKey();
			entry.value = (String) mapEntry.getValue();
			entries.add(entry);
		}
		
		return entries;
	}

	protected boolean isAlreadyInstalled(MonitorProducerAnchor anchor, AgentPath consumerAgentPath)
	{
		return anchor.getConsumerAgentPaths().containsKey(consumerAgentPath.getQualifiedPathName());
	}
	
	protected IAgentProxy getConsumerAgent(CommandAccessor cmdAccessor, AgentPath consumerAgentPath)
	{
		try
		{
			return cmdAccessor.getAgent(consumerAgentPath);
		}
		catch (NoSuchAgentException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.log(e);
		}
		catch (IllegalAgentAccessException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.log(e);			
		}
		return null;		
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public StringVector commandNames()
	{
		return new StringVector(InstallProducerCommand.QualifiedName);
	}
}
