/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.agents;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.almendra.commons.events.IHistorizableCallbackHandler;
import org.almendra.commons.threads.CountDown;
import org.almendra.commons.threads.FutureResult;
import org.almendra.commons.util.AssertUtility;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.InvalidCommandException;
import org.almendra.janet.cas.scheduling.Acknowledge;
import org.almendra.janet.cas.scheduling.CommandEnvelope;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.IWorkLoadBalancingCommand;
import org.almendra.janet.cas.scheduling.IWorkLoadSharingCommand;


/**
 * @author Oliver
 *  
 */
public class MultiAgentProxy extends AbstractAgentProxy implements IAgentProxy
{
	protected List agentProxies = new ArrayList();

	/**
	 * Creates a new <code>MultiAgentProxy</code> object.
	 */
	public MultiAgentProxy()
	{
		super();
	}

	public MultiAgentProxy(List agentProxies)
	{
		super();
		this.agentProxies.addAll(agentProxies);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.logger.node.IAgentProxy#accept(org.almendra.commons.janet.cas.logger.schedulers.ICommand)
	 */
	public void accept(ICommand command) throws RemoteException, InvalidCommandException
	{
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				ICommand clonedCommand = (ICommand) command.clone();
				agentProxy.accept(clonedCommand);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

	public void add(IAgentProxy agentProxy)
	{
		agentProxies.add(agentProxy);
	}

	public boolean remove(IAgentProxy agentProxy)
	{
		return agentProxies.remove(agentProxy);
	}

	public int size()
	{
		return agentProxies.size();
	}

	/**
	 * @param command
	 * @param ackHandler
	 */
	public void accept(ICommand command, Acknowledge ack) throws RemoteException, InvalidCommandException
	{
		if(agentProxies.isEmpty())
			return;
		ack.setCallbackHandler(new CountDown(size()));
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				ICommand clonedCommand = (ICommand) command.clone();
				agentProxy.accept(clonedCommand, ack);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
		ack.acquire();
	}

	/**
	 * Send <code>commmand</code> to an agent expecting a result to be sent
	 * back.
	 * 
	 * @param command
	 *            to be sent to another agent
	 * @param replyToAgentPath
	 *            path to agent that should receive result
	 * @return a future that blocks until the result has arrived or times out.
	 * @throws IllegalArgumentException
	 */
	public void accept(ICommand command, FutureResult futureResult) 
	throws IllegalArgumentException
	{
		//TODO -- not yet implemented
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.logger.node.AbstractAgentProxy#accept(org.almendra.commons.janet.cas.logger.schedulers.ICommand,
	 *      org.almendra.commons.commons.events.ICallbackHandler)
	 */
	public void accept(ICommand command, IHistorizableCallbackHandler handler) throws RemoteException, InvalidCommandException
	{
		//TODO -- not yet tested
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				ICommand clonedCommand = (ICommand) command.clone();
				agentProxy.accept(clonedCommand, handler);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.janet.cas.logger.node.AbstractAgentProxy#replyPathSameAgent()
	 */
	protected AgentPath replyPathSameAgent()
	{
		//TODO -- not yet implemented
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
		return null;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.CommandEnvelope)
	 */
	public void accept(CommandEnvelope env) throws RemoteException, InvalidCommandException
	{
		// TODO -- not yet tested
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				CommandEnvelope clonedEnv = (CommandEnvelope) env.clone(); 
				agentProxy.accept(clonedEnv);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand)
	 */
	public void accept(IWorkLoadSharingCommand command) throws RemoteException, InvalidCommandException
	{
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				IWorkLoadSharingCommand clonedCommand = (IWorkLoadSharingCommand) command.clone();
				agentProxy.accept(clonedCommand);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadBalancingCommand)
	 */
	public void accept(IWorkLoadBalancingCommand command) throws RemoteException, InvalidCommandException
	{
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				IWorkLoadBalancingCommand clonedCommand = (IWorkLoadBalancingCommand) command.clone();
				agentProxy.accept(clonedCommand);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand, org.almendra.commons.janet.cas.scheduling.Acknowledge)
	 */
	public void accept(IWorkLoadSharingCommand command, Acknowledge ack) throws RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand, org.almendra.commons.commons.threads.FutureResult)
	 */
	public void accept(IWorkLoadSharingCommand command, FutureResult futureResult) throws IllegalArgumentException, RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadBalancingCommand, org.almendra.commons.commons.events.IHistorizableCallbackHandler)
	 */
	public void accept(IWorkLoadBalancingCommand command, IHistorizableCallbackHandler handler) throws RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadBalancingCommand, org.almendra.commons.janet.cas.scheduling.Acknowledge)
	 */
	public void accept(IWorkLoadBalancingCommand command, Acknowledge ack) throws RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadBalancingCommand, org.almendra.commons.commons.threads.FutureResult)
	 */
	public void accept(IWorkLoadBalancingCommand command, FutureResult futureResult) throws IllegalArgumentException, RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.IWorkLoadSharingCommand, org.almendra.commons.commons.events.IHistorizableCallbackHandler)
	 */
	public void accept(IWorkLoadSharingCommand command, IHistorizableCallbackHandler handler) throws RemoteException, InvalidCommandException
	{
		AssertUtility.assertTrue("not yet implemented", false); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.agents.IAgentProxy#accept(org.almendra.commons.janet.cas.scheduling.ICommand, int)
	 */
	public void accept(ICommand command, int kind) throws RemoteException, InvalidCommandException
	{
		Iterator it = agentProxies.iterator();
		while (it.hasNext())
		{
			IAgentProxy agentProxy = (IAgentProxy) it.next();
			try
			{
				ICommand clonedCommand = (ICommand) command.clone();
				agentProxy.accept(clonedCommand, kind);
			}
			catch (CloneNotSupportedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
		}
	}

}
