/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.events;

import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.almendra.commons.events.Event;
import org.almendra.commons.events.EventNotifier;
import org.almendra.commons.events.EventRegistry;
import org.almendra.commons.threads.LockUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.IListEvaluator;
import org.almendra.commons.util.exception.PropagatedException;
import org.almendra.commons.util.inet.RMIUtility;
import org.almendra.commons.util.inet.RegistryLocation;



public class ClusterEventRegistry extends EventRegistry implements IClusterEventRegistry
{
	public static final String RMI_PREFIX = "Janet.clusterEventRegistry."; //$NON-NLS-1$
	
	protected Map localRegistriesByNodeName = new Hashtable();
	/**
	 * Creates a new <code>ClusterEventRegistry</code> object.
	 */
	public ClusterEventRegistry()
	{
		super();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.IClusterEventRegistry#register(java.lang.String)
	 */
	public void register(String localEventRegistryName, RegistryLocation registryLocation)
	{				
		try
		{
			LockUtility.acquire(lock);
			Registry registry = RMIUtility.getRegistry(registryLocation);
			ILocalClusterEventRegistryRemote localRegistry = (ILocalClusterEventRegistryRemote) registry.lookup(localEventRegistryName);
			String nodeName = StringUtility.remove(localEventRegistryName, LocalClusterEventRegistry.RMI_PREFIX);
			localRegistriesByNodeName.put(nodeName, localRegistry);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}				
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.IClusterEventRegistry#deregister(java.lang.String)
	 */
	public boolean deregister(String nodeName)
	{
		boolean found = false;
		try
		{
			LockUtility.acquire(lock);
			localRegistriesByNodeName.remove(nodeName);
			Iterator it = registry.values().iterator();
			while (it.hasNext())
			{
				List notifiers = (List) it.next();
				Iterator notifiersIt = new ArrayList(notifiers).iterator();
				while (notifiersIt.hasNext())
				{
					ClusterEventNotifier notifier = (ClusterEventNotifier) notifiersIt.next();
					found |= notifier.removeHandlers(nodeName);
				}			
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();	
		}
		
		return found;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.IClusterEventRegistry#register(java.lang.String, org.almendra.commons.janet.cas.events.HandlerList)
	 */
	public void register(String eventName, ClusterHandlerList handlerList) throws IllegalArgumentException
	{
		super.register(eventName, handlerList);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.IClusterEventRegistry#deregister(java.lang.String, java.lang.String)
	 */
	public boolean deregister(String eventName, String handlerListId) throws NoSuchElementException, IllegalArgumentException
	{
		LockUtility.acquire(lock);
		List notifiers = (List) registry.get(eventName);
		if (notifiers == null)
		{	
			lock.release();
			throw new NoSuchElementException("no event registered named " + eventName); //$NON-NLS-1$
		}
		Iterator it = new ArrayList(notifiers).iterator();
		while (it.hasNext())
		{
			ClusterEventNotifier notifier = (ClusterEventNotifier) it.next();
			if (notifier.hasHandlerList(handlerListId))
			{	
				boolean found = notifiers.remove(notifier);
				lock.release();				
				return found;
			}
		}
		lock.release();
		throw new NoSuchElementException("list evaluator for event named " + eventName + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.events.EventRegistry#getEventNotifier(org.almendra.commons.commons.util.collection.IListEvaluator)
	 */
	protected EventNotifier getEventNotifier(IListEvaluator listEvaluator)
	{
		return new ClusterEventNotifier(listEvaluator);
	}
	
	public boolean isRegistered(String originNodeName, String eventName)
	{
		List notifiers = (List) registry.get(eventName);
		Iterator it = notifiers.iterator();
		while (it.hasNext())
		{
			ClusterEventNotifier notifier = (ClusterEventNotifier) it.next();
			if(notifier.isRegistered(originNodeName))
				return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.events.ILocalEventRegistry#occured(org.almendra.commons.commons.events.Event)
	 */
	public void occured(String originNodeName, Event event) throws PropagatedException
	{
		try
		{
			lock.acquire();
			Iterator it = localRegistriesByNodeName.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String nodeName = (String) entry.getKey();
				if(!originNodeName.equals(nodeName))
				{
					if(isRegistered(nodeName, event.getName()))
					{	
						ILocalClusterEventRegistryRemote localRegistry = (ILocalClusterEventRegistryRemote) entry.getValue();
						localRegistry.occured(event);
					}
				}
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			throw new PropagatedException(e);
		}		
		finally
		{
			lock.release();
		}
		
		
	}

}
