/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.cas.events;

import java.rmi.RemoteException;
import java.util.Iterator;

import org.almendra.commons.events.Event;
import org.almendra.commons.events.EventNotifier;
import org.almendra.commons.events.EventRegistry;
import org.almendra.commons.util.collection.IListEvaluator;
import org.almendra.commons.util.exception.PropagatedException;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.ApplicationAccessor;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;



public class LocalClusterEventRegistry extends EventRegistry implements ILocalClusterEventRegistry
{
	public static final String RMI_PREFIX = "Janet.localClusterEventRegistry."; //$NON-NLS-1$
	public static final String ObjectName = "LocalClusterEventRegistry"; //$NON-NLS-1$
	
	protected IClusterEventRegistry remoteClusterEventRegistry = null;
	protected Logger logger = null;
	protected Node node = null;
	
	/**
	 * Creates a new <code>LocalClusterEventRegistry</code> object.
	 */
	protected LocalClusterEventRegistry()
	{
		super();
	}
	
	/**
	 * Creates a new <code>LocalClusterEventRegistry</code> object.
	 */
	public LocalClusterEventRegistry(Node node, IClusterEventRegistry remoteClusterEventRegistry, Logger parentLogger)
	{
		// TODO - throw exception if node null
		super();
		this.remoteClusterEventRegistry = remoteClusterEventRegistry;
		this.node = node;
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setObjectName(ObjectName);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.events.EventRegistry#getEventNotifier(org.almendra.commons.commons.util.collection.IListEvaluator)
	 */
	protected EventNotifier getEventNotifier(IListEvaluator listEvaluator)
	{
		return new LocalClusterEventNotifier(listEvaluator);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.ILocalClusterEventRegistry#register(java.lang.String, org.almendra.commons.janet.cas.events.HandlerList)
	 */
	public void register(String eventName, HandlerList list, ApplicationAccessor app)
	{
		registerLocal(eventName, list, app);
		registerRemote(eventName, list);
	}

	/**
	 * @param eventName
	 * @param list
	 */
	protected void registerLocal(String eventName, HandlerList list, ApplicationAccessor app)
	{
		// TODO - code partially identical to EventRegistryEnvelope.register()
		Iterator it = list.iterator();
		while (it.hasNext())
		{			
			EventHandlerEnvelope handlerEnv = (EventHandlerEnvelope) it.next();
			if(handlerEnv.getAgentPath() == null)
				throw new IllegalArgumentException("no agent path supplied!"); //$NON-NLS-1$
			handlerEnv.setNode(node);
			handlerEnv.setApplication(app);
			// Make sure the user has no way to provide an agent path to an 
			// agent on a different node or a different application. 
			handlerEnv.getAgentPath().setNodeName(AgentPath.LOCAL_NODE_TAG);
			handlerEnv.getAgentPath().setApplicationName(app.getName());
		}
		register(eventName, list);
	}

	/**
	 * @param eventName
	 * @param list
	 */
	protected void registerRemote(String eventName, HandlerList list)
	{
		try
		{
			ClusterHandlerList clusterHandlerList = makeRemoteList(list);
			remoteClusterEventRegistry.register(eventName, clusterHandlerList);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
	}

	/**
	 * @param list
	 * @return
	 */
	protected ClusterHandlerList makeRemoteList(HandlerList list)
	{
		ClusterHandlerList makeRemoteList = new ClusterHandlerList();
		makeRemoteList.id = list.getId();
		Iterator it = list.iterator();
		while (it.hasNext())
		{
			EventHandlerEnvelope env = (EventHandlerEnvelope) it.next();			
			AgentPath agentPath = new AgentPath(env.getAgentPath());
			agentPath.setNodeName(node.getName());
			ClusterEventHandlerEnvelope remoteHandler = new ClusterEventHandlerEnvelope(env.getHandler(), agentPath);
			makeRemoteList.add(remoteHandler);
		}
		return makeRemoteList;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.ILocalClusterEventRegistry#deregister(java.lang.String, org.almendra.commons.janet.cas.events.HandlerList)
	 */
	public boolean deregister(String eventName, HandlerList list)
	{	
		// order of calls important! 
		boolean result = false;
		
		try
		{			
			result = remoteClusterEventRegistry.deregister(eventName, list.getId());
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return false;
		}
		result &= super.deregister(eventName, list);
		return result;
	}

	/**
	 * @return Returns the logger.
	 */
	protected Logger getLogger()
	{
		return logger;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.events.ILocalClusterEventRegistry#occured(org.almendra.commons.commons.events.Event, boolean)
	 */
	protected void occured(Event event, boolean propagate)
	{
		super.occured(event);
		
		if(!propagate)
			return;
		
		try
		{
			remoteClusterEventRegistry.occured(node.getName(), event);
		}
		catch (RemoteException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		catch (PropagatedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.events.ILocalEventRegistry#occured(org.almendra.commons.commons.events.Event)
	 */
	public void occured(Event event)
	{
		occured(event, true);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.events.ILocalEventRegistry#getHandlerCount(java.lang.String)
	 */
	public int getHandlerCount(String eventName)
	{
		try
		{
			return remoteClusterEventRegistry.getHandlerCount(eventName);
		}
		catch (RemoteException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			return 0;
		}
	}

}
