/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.cas.node;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.almendra.commons.INamedObject;
import org.almendra.commons.events.EventRegistry;
import org.almendra.commons.events.IEventHandler;
import org.almendra.commons.events.IEventRegistry;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.CollectionUtility;
import org.almendra.commons.util.collection.OrderedHashtable;
import org.almendra.commons.util.collection.OrderedMap;
import org.almendra.commons.util.collection.TreeNode;
import org.almendra.commons.util.exception.InsufficientPrivilegeException;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.commons.util.swing.BillboardPanelController;
import org.almendra.janet.cas.agents.AgentProxy;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.agents.MultiAgentProxy;
import org.almendra.janet.cas.agents.SystemAgent;
import org.almendra.janet.cas.events.EventHandlerEnvelope;
import org.almendra.janet.cas.events.EventRegistryEnvelope;
import org.almendra.janet.cas.events.HandlerList;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.IInterpreter;
import org.almendra.janet.cas.spaces.IObjectSpace;
import org.almendra.janet.cas.spaces.IObjectSpaceProvider;
import org.almendra.janet.cas.spaces.ObjectSpace;



public abstract class AbstractApplication implements INamedObject, ILoggable, IObjectSpaceProvider
{
	public static final String LOG_EVENT_REGISTERED_AGENT = "REGISTERED_AGENT"; //$NON-NLS-1$
	
	protected OrderedMap capabilities = new OrderedHashtable();
	protected IObjectSpace objectSpace = new ObjectSpace();
	protected IEventRegistry eventRegistry = new EventRegistry();
	protected List exportedEvents = new LinkedList();
	protected Logger logger = null;
	protected String name = null;
	protected Node node = null;
	protected boolean started = false;
	protected int numOfSchedulers = -1;

	/**
	 * Creates a new <code>AbstractApplication</code> object.
	 */
	public AbstractApplication()
	{
		super();
	}
	
	public AbstractApplication(Logger logger, Node node, String name)
	{
		super();
		this.name = name;
		this.node = node;
		setAndInitLogger(Logger.getChildLogger(logger));
	}
	
	public AbstractApplication(String name)
	{
		super();
		this.name = name;
	}
	
	public void addCapabilities(List newCapabilities)
	{
		Iterator it = newCapabilities.iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			capabilities.put(cap.getName(), cap);
		}
	}
	/**
	 * @param capability
	 */
	public void addCapability(Capability capability)
	{
		AssertUtility.assertNotNull(capability);
		String key = capability.getName();
		AssertUtility.assertTrue(!capabilities.containsKey(key));
		capabilities.put(key, capability);
	}
	
	/**
	 * @param list
	 */
	public void addExportedEventDescriptors(List eventDescriptors)
	{
		exportedEvents.addAll(eventDescriptors);
	}
	
	public boolean checkValidity()
	throws IllegalArgumentException
	{
		if(!StringUtility.isValid(getName()))
			throw new IllegalArgumentException("application has no name!"); //$NON-NLS-1$
		
		if(getNode() == null)
			throw new IllegalArgumentException("no node assigned to application " + StringUtility.print(getName()));  //$NON-NLS-1$
		
		if(capabilities.isEmpty())
			throw new IllegalArgumentException("application " + getName() + " has no capabilities!"); //$NON-NLS-1$ //$NON-NLS-2$
		
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			cap.checkValidity();
		}
		
		return true;
	}
	/**
	 * @return
	 */
	public Set getCapabilities()
	{
		return CollectionUtility.toSet(capabilities.valuesInOrder());
	}
	
	public ILocalEventRegistry getEventRegistry()
	{
		return new EventRegistryEnvelope(node, eventRegistry);
	}
	
	/**
	 * @return Returns the exportedEvents.
	 */
	public List getExportedEvents()
	{
		return exportedEvents;
	}
	/**
	 * @param capability
	 * @return
	 */
	public boolean hasCapability(Capability capability)
	{
		return capabilities.containsValue(capability);
	}
	/**
	 * @param capability
	 * @return
	 */
	public boolean hasCapability(String nodeName, String applicationName, String capabilityName)
	{
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		return anchor.getClusterImage().hasCapability(nodeName, applicationName, capabilityName);
	}
	
	/**
	 * @param agentPath
	 * @return
	 */
	public IAgentDispatcher getAgentDispatcher(AgentPath agentPath)
	throws NoSuchAgentException, IllegalAgentAccessException
	{
		return getAgentDispatcher(agentPath, true);
	}
	
	/**
	 * @param agentPath
	 * @return
	 */
	public IAgentDispatcher getAgentDispatcher(AgentPath agentPath, boolean restrictedAccess)
	throws NoSuchAgentException, IllegalAgentAccessException
	{
		if(restrictedAccess && (agentPath.getApplicationName() == null || !agentPath.getApplicationName().equals(getName())))
			throw new IllegalAgentAccessException(
				"no access from application " + agentPath.getApplicationName() + " to application " + getName() + "!"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		return node.getAgentDispatcher(agentPath);
	}
	
	/**
	 * @param capabilities
	 * @return
	 */
	public boolean hasCapabilities(Set capabilities)
	{
		return capabilities.containsAll(capabilities);
	}
	/**
	 * @return
	 */
	protected synchronized boolean isStarted()
	{
		return started;
	}
	protected void primRegisterEvents()
	{
		Iterator it = getExportedEvents().iterator();
		while (it.hasNext())
		{			
			try
			{
				EventDescriptor desc = (EventDescriptor) it.next();
				IEventHandler handler = (IEventHandler) (Class.forName(desc.getHandlerName()).newInstance());
				EventHandlerEnvelope env = new EventHandlerEnvelope(handler, desc.getAgentPath());
				HandlerList handlers = new HandlerList(env);
				getNodeEventRegistryEnvelope().register(desc.getName(), handlers, getAccessor());
			}
			catch (InstantiationException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
			catch (IllegalAccessException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}
			catch (ClassNotFoundException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}			
			
		}
	}
	
	/**
	 * @return
	 */
	protected ILocalEventRegistry getNodeEventRegistryEnvelope()
	{
		return node.getNodeEventRegistry();
	}

	protected void primStart()
	{
		primRegisterEvents();
	}
	/**
	 * 
	 */
	protected void primStop()
	{
		// TODO - not yet implemented
		// all nodes need to deregister this application, then all agents have to be stopped
		// what if application is busy and is not at a safe point to shut down?
		// tricky issue, skip for now
		// For the time being remove agents with brute force - kludge for now
		Iterator it = getAllLocalAgentNames().iterator();
		while (it.hasNext())
		{
			String localAgentName = (String) it.next();
			node.getLocalAgentDispatcher().stopAndRemove(localAgentName);
		}		
	}
	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		return null;
	}
	public void setAndInitLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setApplicationName(getName());
		List eventNames = new ArrayList();
		eventNames.add(LOG_EVENT_REGISTERED_AGENT);
		logger.registerForEvents(eventNames);
	}
	/**
	 * @param string
	 */
	public void setName(String string)
	{
		name = string;
	}
	/**
	 * @param node The node to set.
	 */
	public void setNode(Node node)
	{
		this.node = node;
	}
	/**
	 * @param b
	 */
	protected synchronized void setStarted(boolean started)
	{
		this.started = started; 
	}
	/**
	 * Start all the agents of the capability if not already started.
	 * 
	 * @return
	 */
	public boolean start()
	{
		if (isStarted())
			return false;
		primStart();
		setStarted(true);
		return true;
	}
	/**
	 * Stop all the agents of the capability if started.
	 * 
	 * @return
	 */
	public boolean stop()
	{
		if (!isStarted())
			return false;
		primStop();
		setStarted(false);
		return true;
	}
	/**
	 * @return
	 */
	public List getAllNodeNames()
	{
		return node.getSystemAnchor().getClusterImage().getAllForeignNodeNames();
	}
	
	protected List getAllQualifiedAgentNames()
	{
		List qualifiedAgentNames = new ArrayList(); 
		Iterator it = getAllAgentPaths().iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			qualifiedAgentNames.add(agentPath.getQualifiedPathName());
		}
		return qualifiedAgentNames;
	}
	/**
	 * @return
	 */
	public List getAllForeignNodeNames()
	{
		List nodeNames = getAllNodeNames();
		nodeNames.remove(getName());
		nodeNames.remove(CentralAnchor.NAME);
		return nodeNames;
	}
	/**
	 * @return
	 */
	public List getAllForeignNodeNamesWithCentral()
	{
		List nodeNames = getAllNodeNames();
		nodeNames.remove(getName());
		return nodeNames;
	}
	protected List getAllLocalAgentNames()
	{
		List qualifiedAgentNames = new ArrayList(); 
		Iterator it = getAllAgentPaths().iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			qualifiedAgentNames.add(agentPath.getLocalAgentName());
		}
		return qualifiedAgentNames;
	}
	
	public List getAgentPaths(String capabilityName, String agentName)
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		return anchor.getClusterImage().getAgentPaths(getName(), capabilityName, agentName);		
	}
	
	public List getAllAgentPaths(CapabilityPath path)
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		return anchor.getClusterImage().getAgentPaths(path);		
	}
	
	public int getAgentCount(String nodeName, String capabilityPath)
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		LocalCapabilityPath path = new LocalCapabilityPath(capabilityPath);
		AbstractNodeImage nodeImage = anchor.getClusterImage().getNodeImage(nodeName);
		if(nodeImage == null)
			return 0;
		return nodeImage.getAgentCount(path.getApplicationName(), path.getCapabilityName());		
	}
	
	public List getAgentPaths(String capabilityName)
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		return anchor.getClusterImage().getAgentPaths(getName(), capabilityName);		
	}
	
	public IAgentProxy getAllAgentsWithCapability(String capabilityName, String agentName)
	{
		List agentPaths = getAgentPaths(capabilityName, agentName);
		List agents = new LinkedList();
		Iterator it = agentPaths.iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			try
			{
				agents.add(getAgent(agentPath));
			} 
			catch (Exception e)
			{
				// no exception but logical error if this happens - do not handle
				System.out.println(e.getMessage());
				e.printStackTrace();
				getLogger().exception(e);
			}
		}
		return new MultiAgentProxy(agents);
	}
	
	public Capability getCapability(String capabilityName)
	{
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability capability = (Capability) it.next();
			if(capability.getName().equals(capabilityName))
				return capability;
		}
		return null;
	}
	
	/**
	 * @return Returns the objectSpace.
	 */
	public IObjectSpace getObjectSpace()
	{
		return objectSpace;
	}
	
	/**
	 * @return Returns the objectSpace.
	 */
	public IObjectSpace getNodeObjectSpace()
	{
		return node.getNodeObjectSpace();
	}
	
	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}
	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}
	/**
	 * @return Returns the node.
	 */
	public Node getNode()
	{
		return node;
	}
	/**
	 * Answer the path to the system agent of node <nodeName>.
	 * 
	 * @param nodeName
	 * @return
	 */
	protected AgentPath getSystemAgentPath(String nodeName)
	{
		return new AgentPath(nodeName, SystemApplication.NAME, SystemApplication.CORE_CAPABILITY_NAME, SystemAgent.NAME);
	}

	public List getAllAgentPaths()
	{
		List agentPaths = new ArrayList(); 
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			Iterator agentNamesIt = cap.getAgentNames().iterator();
			while (agentNamesIt.hasNext())
			{
				String agentName = (String) agentNamesIt.next();
				agentPaths.add(new AgentPath(getNode().getName(), getName(), cap.getName(), agentName));
			}						
		}
		return agentPaths;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.node.AbstractApplication#getNameDescriptor()
	 */
	protected String getNameDescriptor()
	{
		return "application name=" + StringUtility.print(getName()); //$NON-NLS-1$
	}

	public AgentPath getLocalAgentPath(String capabilityName, String agentName) 
	throws NoSuchAgentException 
	{
		AgentPath agentPath = new AgentPath(getNode().getName(), getName(), capabilityName, agentName);
		if(!hasAgent(capabilityName, agentName))
			throw new NoSuchAgentException("no such agent " + agentPath); //$NON-NLS-1$
		return agentPath;
	}
	
	/**
	 * @param capabilityName
	 * @param agentName
	 * @return
	 */
	public boolean hasAgent(String capabilityName, String agentName)
	{
		Capability cap = getCapability(capabilityName);
		if(cap == null)
			return false;
		return cap.hasAgent(agentName);
	}

	public IAgentProxy getLocalAgent(String capabilityName, String agentName) 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return getAgent(new AgentPath(getNode().getName(), getName(), capabilityName, agentName));
	}

	public IAgentProxy getAgent(AgentPath path)
	throws NoSuchAgentException, IllegalAgentAccessException
	{
		return getAgent(path, true);
	}
	
	protected IAgentProxy getAgent(AgentPath path, boolean restrictedAccess)
	throws NoSuchAgentException, IllegalAgentAccessException
	{
		path.checkValidity();
		AgentPath verifiedPath = new AgentPath(path);
		IAgentDispatcher agentDispatcher = getAgentDispatcher(verifiedPath, restrictedAccess);
		return new AgentProxy(this, agentDispatcher, verifiedPath, logger);
	}
	
	public AgentProxy getCrossAgent(AgentPath path)
	throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		if(!getName().equals(path.getApplicationName()) && !isSystemApplication())
			throw new InsufficientPrivilegeException("privilege only granted to system application"); //$NON-NLS-1$
		path.checkValidity();
		AgentPath verifiedPath = new AgentPath(path);
		IAgentDispatcher agentDispatcher = getAgentDispatcher(verifiedPath, false);
		return new AgentProxy(this, agentDispatcher, verifiedPath, logger);
	}
	
	
	
	/**
	 * @return
	 */
	public List getForeignNodeNames()
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		return anchor.getClusterImage().getForeignNodeNamesWithApplication(getName());
	}
	
	/**
	 * Return a list with all the nodes that have an application with the same name registered
	 * including the owner node of this application.
	 * @return
	 */
	public List getNodeNames()
	{
		List nodeNames = getForeignNodeNames();
		nodeNames.add(getNode().getName());
		return nodeNames;
	}
	
	/**
	 * @param path
	 * @return
	 */
	public List getAllCrossAgentPaths(LocalAgentPath localPath)
	{
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		return anchor.getClusterImage().getAllAgents(localPath);		
	}
	
	public IAgentProxy getAllCrossAgents(List agentPaths) 
	throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		List agentProxies = new ArrayList();
		Iterator it = agentPaths.iterator();
		while (it.hasNext())
		{
			AgentPath agentPath = (AgentPath) it.next();
			LocalAgentPath localAgentPath = new LocalAgentPath(agentPath);
			String nodeName = agentPath.getNodeName();
			AbstractNodeImage nodeImage = anchor.getClusterImage().getNodeImage(nodeName);
			if(nodeImage.hasAgent(localAgentPath))
				agentProxies.add(getCrossAgent(agentPath));			
		}
		if(agentProxies.isEmpty())
			return null;
		if(agentProxies.size() == 1)
			return (IAgentProxy) agentProxies.get(0);
		return new MultiAgentProxy(agentProxies);
	}

	/**
	 * @param path
	 * @return
	 */
	public List getAllCrossAgents(LocalAgentPath localPath)
	{
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		List foreignAgentPaths = anchor.getClusterImage().getAllAgents(localPath);		
		Iterator it = foreignAgentPaths.iterator();
		List agentProxies = new LinkedList();
		while (it.hasNext())
		{			
			try
			{
				AgentPath path = (AgentPath) it.next();
				agentProxies.add(getCrossAgent(path));
			} 
			catch (Exception e)
			{
				// logical error if this happens - do not handle
				System.out.println(e.getMessage());
				e.printStackTrace();
				getLogger().exception(e);
			}
		}
		return agentProxies;
	}
	
	public List getAllAgentPaths(LocalAgentPath localPath)
	{
		AssertUtility.assertTrue(localPath.getApplicationName().equals(getName()));
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		return anchor.getClusterImage().getAllAgents(localPath);	
	}
	
	/**
	 * @param path
	 * @return
	 */
	public IAgentProxy getAllAgents(LocalAgentPath localPath)
	{
		AssertUtility.assertTrue(localPath.getApplicationName().equals(getName()));
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		List foreignAgentPaths = anchor.getClusterImage().getAllAgents(localPath);		
		Iterator it = foreignAgentPaths.iterator();
		List agentProxies = new LinkedList();
		while (it.hasNext())
		{
			AgentPath path = (AgentPath) it.next();
			try
			{
				agentProxies.add(getAgent(path));
			} 
			catch (Exception e)
			{
				// logical error if this happens - do not handle
				System.out.println(e.getMessage());
				e.printStackTrace();
				getLogger().exception(e);
			}
		}
		return new MultiAgentProxy(agentProxies);
	}
	
	/**
	 * @param string
	 * @param string2
	 * @return
	 */
	public IAgentProxy getAllForeignAgents(LocalAgentPath localPath) 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		List foreignAgentPaths = getAllForeignAgentPaths(localPath);		
		Iterator it = foreignAgentPaths.iterator();
		List agentProxies = new LinkedList();
		while (it.hasNext())
		{
			AgentPath path = (AgentPath) it.next();
			agentProxies.add(getAgent(path));
		}
		return new MultiAgentProxy(agentProxies);
	}
	
	public boolean hasForeignAgents(LocalAgentPath localPath)
	{
		return getAllForeignAgentPaths(localPath).size() > 0;
	}
	
	public List getAllForeignAgentPaths(LocalAgentPath localPath) 
	throws IllegalArgumentException
	{
		AbstractNodeAnchor anchor = node.getSystemAnchor();
		return anchor.getClusterImage().getAllForeignAgents(localPath);		
	}

	/**
	 * @param ackAgentPath
	 * @return
	 */
	public boolean hasAgent(AgentPath ackAgentPath)
	{
		if(ackAgentPath == null)			
			return false;
		if(!getNode().getName().equals(ackAgentPath.getNodeName()))
			return false;
		if(!getName().equals(ackAgentPath.getApplicationName()))
			return false;
		return hasAgent(ackAgentPath.getCapabilityName(), ackAgentPath.getAgentName());
	}

	/**
	 * @return
	 */
	public boolean isSystemApplication()
	{
		return false;
	}
	
	public String getNodeName()
	{
		return getNode().getName();
	}

	public ApplicationAccessor getAccessor()
	{
		return new ApplicationAccessor(this);
	}

	protected String getNodeLabel()
	{
		String nodeLabel = node.getNodeLabel();
		if(nodeLabel == null)
			return nodeLabel;
		return nodeLabel;
	}

	/**
	 *  
	 */
	protected void primStartInitialCommands()
	{
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			Iterator descIt = cap.getAgentDescriptors().iterator();
			while (descIt.hasNext())
			{
				AgentDescriptor desc = (AgentDescriptor) descIt.next();
				String initialCommandName = desc.getExecuteWhenAppStarted();
				if (initialCommandName != null)
				{
					AgentPath agentPath = cap.getAgentPath(desc.getName());
					try
					{
						IAgentProxy agent = getAgent(agentPath);
						ICommand cmd = (ICommand) (Class.forName(initialCommandName).newInstance());
						agent.accept(cmd);
					}
					catch (Exception e)
					{
						System.out.println(e.getMessage());
						e.printStackTrace();
						getLogger().exception(e);
					}
				}
			}
		}
	}

	/**
	 * 
	 * @return
	 */
	public List toForeignDescriptor()
	{
		List agentPaths = new ArrayList();
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			Iterator itAgentNames = cap.getAgentNames().iterator();
			while (itAgentNames.hasNext())
			{
				String agentName = (String) itAgentNames.next();
				agentPaths.add(new AgentPath(node.getName(), getName(), cap.getName(), agentName));
			}
		}
		return agentPaths;
	}

	public TreeNode toTreeDescriptor()
	{
		TreeNode root = new TreeNode(getNameDescriptor());
		
		if(!getExportedEvents().isEmpty())
		{	
			TreeNode exportedEventsNode = root.addChild("exportedEvents"); //$NON-NLS-1$
		
			Iterator it = getExportedEvents().iterator();
			while (it.hasNext())
			{
				EventDescriptor eventDesc = (EventDescriptor) it.next();
				exportedEventsNode.addChild(eventDesc.toTreeNodeString());
			}
		}
		
		TreeNode capabilitiesNode = root.addChild("capabilities"); //$NON-NLS-1$
		
		Iterator it = getCapabilities().iterator();
		while (it.hasNext())
		{
			Capability cap = (Capability) it.next();
			TreeNode capabilityNode = capabilitiesNode.addChild("capability name=\"" + cap.getName() + "\""); //$NON-NLS-1$ //$NON-NLS-2$
			if(!cap.getAgentDescriptors().isEmpty())
			{
				// agents
				TreeNode agentsNode = capabilityNode.addChild("agents"); //$NON-NLS-1$
				Iterator agentIt = cap.getAgentDescriptors().iterator();
				while (agentIt.hasNext())
				{
					AgentDescriptor agentDesc = (AgentDescriptor) agentIt.next();
					String agentLabel = "agent name=" + "\"" + agentDesc.getName() + "\""; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					if(StringUtility.isValid(agentDesc.getExecuteWhenAppStarted()))
					{
						agentLabel += " executeWhenStarted=" + "\"" + agentDesc.getExecuteWhenAppStarted() + "\""; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					}
					agentsNode.addChild(agentLabel); 
				}
				
				// interpreters
				TreeNode interpretersNode = capabilityNode.addChild("interpreters"); //$NON-NLS-1$
				Iterator interpretersIt = cap.getInterpreters().iterator();
				while (interpretersIt.hasNext())
				{
					IInterpreter interpreter = (IInterpreter) interpretersIt.next();
					interpretersNode.addChild(interpreter.getClass().getName()); //$NON-NLS-1$
				}
			}
		}
	
		return root;
	}
	
	public Map getMonitorMaskDefaultValues()
	{
		Map defaultValues = new Hashtable();
		
		defaultValues.put(NodeAnchor.NODE_NAME, getNode().getName());
		
		String nodeLabel = getNodeLabel();		
		if(nodeLabel != null)
			defaultValues.put(NodeAnchor.NODE_LABEL, nodeLabel); //$NON-NLS-1$
		
		defaultValues.put(NodeAnchor.LOCAL_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.LOCAL_WAITING_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.LOCAL_EXECUTING_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		
		defaultValues.put(NodeAnchor.FOREIGN_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.FOREIGN_WAITING_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.EVICTED_WAITING_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.FOREIGN_EXECUTING_COMMANDS_COUNT, "1"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.EVICTED_EXECUTING_COMMANDS_COUNT, "0"); //$NON-NLS-1$
		defaultValues.put(NodeAnchor.QUEUE_SIZE_CATEGORY, "0"); //$NON-NLS-1$
		
		// defaultValues.put(FOREIGN_EXECUTING_BALANCING_COMMANDS_COUNT, "0");
		
		return defaultValues;
	}


	public List getMonitorMask()
	{
		List mask = new LinkedList();
		
		mask.add(BillboardPanelController.BLANK_LINE);
		mask.add(NodeAnchor.NODE_NAME);
			
		if(getNodeLabel() != null)
			mask.add(NodeAnchor.NODE_LABEL);
		
		mask.add(BillboardPanelController.BLANK_LINE);
		
		mask.add(NodeAnchor.LOCAL_COMMANDS_COUNT);
		mask.add(NodeAnchor.LOCAL_WAITING_COMMANDS_COUNT);
		mask.add(NodeAnchor.LOCAL_EXECUTING_COMMANDS_COUNT);
		
		mask.add(NodeAnchor.FOREIGN_COMMANDS_COUNT);
		mask.add(NodeAnchor.FOREIGN_WAITING_COMMANDS_COUNT);
		mask.add(NodeAnchor.FOREIGN_EXECUTING_COMMANDS_COUNT);
		mask.add(NodeAnchor.EVICTED_WAITING_COMMANDS_COUNT);
		mask.add(NodeAnchor.EVICTED_EXECUTING_COMMANDS_COUNT);
		mask.add(NodeAnchor.QUEUE_SIZE_CATEGORY);
		
		return mask;
	}

	/**
	 * @return
	 */
	public int getNumberOfSchedulers()
	{
		if (numOfSchedulers == -1)
		{
			numOfSchedulers = 0;
			Iterator it = getCapabilities().iterator();
			while (it.hasNext())
			{
				Capability cap = (Capability) it.next();
				numOfSchedulers += cap.getNumberOfSchedulers();
			}
		}
		return numOfSchedulers;
	}

	/**
	 * @param path
	 * @return
	 */
	public List getAllNodeNamesWithCapability(LocalCapabilityPath path, List nodeNamesToConsider)
	{
		NodeAnchor anchor = (NodeAnchor) getNode().getSystemAnchor();
		return anchor.getClusterImage().getAllNodeNamesWithCapability(path, nodeNamesToConsider);
	}	
	
}
