/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.inet.RMIUtility;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;

/**
 * @author Oliver
 */
public abstract class AbstractClusterImage
{
	protected Map nodeImages = new Hashtable();
	protected ReentrantLock lock = new ReentrantLock();
	protected AbstractNodeImage localNodeImage = null;

	public AbstractClusterImage()
	{
		super();
	}

	/**
	 * @param nodeImage
	 */
	public AbstractNodeImage removeNodeImage(String nodeName)
	{
		AbstractNodeImage nodeImage = null;
		try
		{
			lock.acquire();
			nodeImage = (AbstractNodeImage) nodeImages.remove(nodeName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return nodeImage;
	}

	/**
	 * @param nodeImage
	 */
	public void deregisterAllApplications(String nodeName)
	{
		try
		{
			lock.acquire();
			AbstractNodeImage nodeImage = (AbstractNodeImage) nodeImages.get(nodeName);
			nodeImage.deregisterAllApplications();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
	}

	public abstract AbstractNodeImage addNodeImage(IAgentDispatcher agentDispatcher, NodeLocation nodeLocation);

	/**
	 * @param path
	 * @return
	 */
	public IAgentDispatcher getAgentDispatcher(AgentPath path) throws NoSuchAgentException
	{
		AbstractNodeImage nodeImage = getNodeImage(path.getNodeName());
		if (nodeImage == null)
			throw new NoSuchAgentException(path, NoSuchAgentException.NODE_LEVEL, path.getNodeName());
		return nodeImage.getAgentDispatcher();
	}

	/**
	 * @return
	 */
	public List getAllForeignNodeNames()
	{
		List nodeNames = new ArrayList();
		try
		{
			lock.acquire();
			Iterator it = nodeImages.values().iterator();
			while (it.hasNext())
			{
				AbstractNodeImage nodeImage = (AbstractNodeImage) it.next();
				if (nodeImage.isForeign)
					nodeNames.add(nodeImage.getNodeName());
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return nodeNames;
	}

	public AbstractNodeImage getNodeImage(String nodeName)
	{
		try
		{
			lock.acquire();
			return (AbstractNodeImage) nodeImages.get(nodeName);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return null;
	}

	public AbstractNodeImage getNodeImage(NodeLocation locationOrNull, String nodeName) throws AccessException,
			RemoteException, NotBoundException
	{
		try
		{
			lock.acquire();
			AbstractNodeImage nodeImage = (AbstractNodeImage) nodeImages.get(nodeName);
			if (nodeImage == null && locationOrNull != null)
			{
				Registry registry = RMIUtility.getRegistry(locationOrNull.getRegistryLocation());
				IAgentDispatcher dispatcher = (IAgentDispatcher) registry.lookup(Node.RMI_LABEL_PREFIX
						+ locationOrNull.getNodeName());
				nodeImage = addLocalNode(dispatcher, locationOrNull);
			}
			return nodeImage;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return null;
	}

	/**
	 * @param string
	 * @return
	 */
	public List getForeignNodeNamesWithApplication(String applicationName)
	{
		Set nodeNames = new HashSet();
		try
		{
			lock.acquire();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String key = (String) entry.getKey();
				AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
				if (nodeImage.isForeign)
				{
					Iterator appIt = nodeImage.getApplications().entrySet().iterator();
					while (appIt.hasNext())
					{
						entry = (Map.Entry) appIt.next();
						key = (String) entry.getKey();
						if (key.equals(applicationName))
						{
							List paths = (List) entry.getValue();
							Iterator pathsIt = paths.iterator();
							while (pathsIt.hasNext())
							{
								AgentPath path = (AgentPath) pathsIt.next();
								nodeNames.add(path.getNodeName());
							}
						}
					}
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return ListUtility.toList(nodeNames);
	}

	public int getAgentCount(String applicationName, String capabilityName)
	{
		List agentPaths = getAgentPaths(applicationName, capabilityName);
		if (agentPaths == null)
			return 0;
		return agentPaths.size();
	}

	/**
	 * @param string
	 * @return
	 */
	public List getAgentPaths(String applicationName, String capabilityName, String agentName)
	{
		try
		{
			lock.acquire();
			List agentPaths = getAgentPaths(applicationName, capabilityName);
			Iterator it = new LinkedList(agentPaths).iterator();
			while (it.hasNext())
			{
				AgentPath agentPath = (AgentPath) it.next();
				if (!agentPath.getAgentName().equals(agentName))
					agentPaths.remove(agentPath);
			}
			return agentPaths;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return null;
	}

	/**
	 * @param string
	 * @return
	 */
	public List getAgentPaths(String applicationName, String capabilityName)
	{
		List agentPaths = new ArrayList();
		try
		{
			lock.acquire();
			agentPaths = new ArrayList();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String key = (String) entry.getKey();
				AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
				Iterator appIt = nodeImage.getApplications().entrySet().iterator();
				while (appIt.hasNext())
				{
					entry = (Map.Entry) appIt.next();
					key = (String) entry.getKey();
					if (key.equals(applicationName))
					{
						Set paths = (Set) entry.getValue();
						Iterator pathsIt = paths.iterator();
						while (pathsIt.hasNext())
						{
							AgentPath path = (AgentPath) pathsIt.next();
							if (path.getCapabilityName().equals(capabilityName))
								agentPaths.add(new AgentPath(path));
						}
					}
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return agentPaths;
	}

	/**
	 * @param string
	 * @return
	 */
	public List getForeignNodesWithApplication(String applicationName)
	{
		List nodes = new ArrayList();
		try
		{
			lock.acquire();
			nodes = new ArrayList();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String key = (String) entry.getKey();
				AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
				if (nodeImage.isForeign)
				{
					Iterator appIt = nodeImage.getApplications().entrySet().iterator();
					while (appIt.hasNext())
					{
						entry = (Map.Entry) appIt.next();
						key = (String) entry.getKey();
						if (key.equals(applicationName))
						{
							AgentPath path = (AgentPath) entry.getValue();
							nodes.add(path);
						}
					}
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return nodes;
	}

	/**
	 * @return
	 */
	public List getAllForeignNodeLocations()
	{
		List allNodeLocations = new ArrayList();
		try
		{
			lock.acquire();
			allNodeLocations = new ArrayList();
			Iterator it = nodeImages.values().iterator();
			while (it.hasNext())
			{
				AbstractNodeImage nodeImage = (AbstractNodeImage) it.next();
				if (nodeImage.isForeign)
					allNodeLocations.add(nodeImage.getNodeLocation());
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return allNodeLocations;
	}

	public List getAllAgents(LocalAgentPath localPath)
	{
		return getAllForeignAgents(localPath, false);
	}

	public List getAllForeignAgents(LocalAgentPath localPath)
	{
		return getAllForeignAgents(localPath, true);
	}

	/**
	 * @param string
	 * @param capabilityName
	 * @param agentName
	 * @return
	 */
	public List getAllForeignAgents(LocalAgentPath localPath, boolean onlyForeign)
	{
		List allAgentPaths = new ArrayList();
		try
		{
			lock.acquire();
			String applicationName = localPath.getApplicationName();
			String capabilityName = localPath.getCapabilityName();
			String agentName = localPath.getAgentName();
			Iterator it = nodeImages.values().iterator();
			while (it.hasNext())
			{
				AbstractNodeImage nodeImage = (AbstractNodeImage) it.next();
				boolean considerNodeImage = true;
				if (onlyForeign)
					considerNodeImage = nodeImage.isForeign;
				if (considerNodeImage && nodeImage.hasAgent(applicationName, capabilityName, agentName))
				{
					AgentPath agentPath = new AgentPath(nodeImage.getNodeName(), applicationName, capabilityName,
							agentName);
					allAgentPaths.add(agentPath);
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return allAgentPaths;
	}

	public int getForeignNodesCount()
	{
		int size = 0;
		try
		{
			lock.acquire();
			Iterator it = nodeImages.values().iterator();
			while (it.hasNext())
			{
				AbstractNodeImage nodeImage = (AbstractNodeImage) it.next();
				if (nodeImage.isForeign)
					size++;
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return size;
	}

	public List getAllForeignSystemAgentPaths(String applicationName)
	{
		List systemAgents = new ArrayList();
		try
		{
			lock.acquire();
			Iterator it = nodeImages.values().iterator();
			while (it.hasNext())
			{
				AbstractNodeImage nodeImage = (AbstractNodeImage) it.next();
				if (nodeImage.isForeign && nodeImage.hasApplication(applicationName))
				{
					AgentPath agentPath = AgentPath.getSystemPath(nodeImage.getNodeName());
					systemAgents.add(agentPath);
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return systemAgents;
	}

	/**
	 * @param dispatcher
	 * @param nodeLocation
	 */
	public AbstractNodeImage addLocalNode(IAgentDispatcher dispatcher, NodeLocation nodeLocation)
	{
		try
		{
			lock.acquire();
			AbstractNodeImage nodeImage = createLocalNode(dispatcher, nodeLocation);
			localNodeImage = nodeImage;
			AssertUtility.assertFalse(nodeImages.containsKey(nodeImage.getNodeName()));
			nodeImages.put(nodeImage.getNodeName(), nodeImage);
			return nodeImage;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return null;
	}

	/**
	 * @param dispatcher
	 * @param nodeLocation
	 * @return
	 */
	protected abstract AbstractNodeImage createLocalNode(IAgentDispatcher dispatcher, NodeLocation nodeLocation);

	/**
	 * @return
	 */
	public AbstractNodeImage getLocalNodeImage()
	{
		return localNodeImage;
	}

	/**
	 * @param agentPath
	 * @return
	 */
	public Map getAgentsQueueSizes(LocalAgentPath agentPath)
	{
		try
		{
			lock.acquire();
			Map agentQueueSizes = new Hashtable();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String nodeName = (String) entry.getKey();
				AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
				try
				{
					long queueSize = nodeImage.getAgentDispatcher().getQueueSize(agentPath);
					if (queueSize != -1)
						agentQueueSizes.put(new AgentPath(nodeName, agentPath), new Long(queueSize));
				}
				catch (RemoteException e)
				{
					System.out.println(e.getMessage());
					e.printStackTrace();
				}
			}
			return agentQueueSizes;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return new Hashtable();
	}

	/**
	 * @param nodeName
	 * @param capabilityName
	 * @return
	 */
	public boolean hasCapability(String nodeName, String applicationName, String capabilityName)
	{
		try
		{
			lock.acquire();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String currentNodeName = (String) entry.getKey();
				if (currentNodeName.equals(nodeName))
				{
					AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
					return nodeImage.hasCapability(applicationName, capabilityName);
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return false;
	}
	
	public boolean hasApplication(String nodeName, String applicationName)
	{
		try
		{
			lock.acquire();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String currentNodeName = (String) entry.getKey();
				if (currentNodeName.equals(nodeName))
				{
					AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
					return nodeImage.hasApplication(applicationName);
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return false;
	}

	/**
	 * @param path
	 * @return
	 */
	public List getAllNodeNamesWithCapability(LocalCapabilityPath path, List nodeNamesToConsider)
	{
		try
		{
			lock.acquire();
			List nodeNames = new ArrayList();
			Iterator it = nodeNamesToConsider.iterator();
			while (it.hasNext())
			{
				String nodeName = (String) it.next();
				AbstractNodeImage nodeImage = (AbstractNodeImage) nodeImages.get(nodeName);
				if (nodeImage != null && nodeImage.hasCapability(path.getApplicationName(), path.getCapabilityName()))
					nodeNames.add(nodeName);
			}
			return nodeNames;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return new ArrayList();
	}

	/**
	 * @param path
	 * @return
	 */
	public List getAgentPaths(CapabilityPath path)
	{
		try
		{
			lock.acquire();
			List agentPaths = new ArrayList();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String nodeName = (String) entry.getKey();
				if (nodeName.equals(path.getNodeName()))
				{
					AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
					List agentNames = nodeImage.getAgentNames(path.getApplicationName(), path.getCapabilityName());
					Iterator innerIt = agentNames.iterator();
					while (innerIt.hasNext())
					{
						String agentName = (String) innerIt.next();
						agentPaths.add(new AgentPath(path.getNodeName(), path.getApplicationName(), path
								.getCapabilityName(), agentName));
					}
				}
			}
			return agentPaths;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return new ArrayList();
	}

	/**
	 * @param nodeName
	 * @return
	 */
	public List getAllApplicationCapabilitesOn(String nodeName)
	{
		try
		{
			lock.acquire();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String currentNodeName = (String) entry.getKey();
				if (currentNodeName.equals(nodeName))
				{
					AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
					return nodeImage.getAllApplicationCapabilityPaths();
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return new ArrayList();
	}

	/**
	 * @param localCapabilityPaths
	 * @param excludingNodeNames
	 * @return
	 */
	public List existingCapabilitiesOnOtherNodes(List localCapabilityPaths, List excludingNodeNames)
	{
		List existingCapabilitiesOnOtherNodes = new ArrayList();
		Iterator it = localCapabilityPaths.iterator();
		while (it.hasNext())
		{
			LocalCapabilityPath path = (LocalCapabilityPath) it.next();
			if (existsCapabilityOnOtherNodes(path, excludingNodeNames))
				existingCapabilitiesOnOtherNodes.add(path);
		}
		return existingCapabilitiesOnOtherNodes;
	}

	/**
	 * @param path
	 * @param excludingNodeNames
	 * @return
	 */
	protected boolean existsCapabilityOnOtherNodes(LocalCapabilityPath path, List excludingNodeNames)
	{
		try
		{
			lock.acquire();
			Iterator it = nodeImages.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				String nodeName = (String) entry.getKey();
				if (!excludingNodeNames.contains(nodeName))
				{
					AbstractNodeImage nodeImage = (AbstractNodeImage) entry.getValue();
					if (nodeImage.getAgentCount(path.getApplicationName(), path.getCapabilityName()) > 0)
						return true;
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.release();
		}
		return false;
	}
}