/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;


import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.almendra.commons.INamedObject;
import org.almendra.commons.events.Event;
import org.almendra.commons.events.IEventParams;
import org.almendra.commons.util.StringUtility;
import org.almendra.janet.cas.agents.AbstractAgent;
import org.almendra.janet.cas.agents.IAgent;
import org.almendra.janet.cas.handlers.CommandEnvelopeParams;
import org.almendra.janet.cas.scheduling.CommandEnvelope;
import org.almendra.janet.cas.scheduling.CommandResponseEnvelope;
import org.almendra.janet.cas.scheduling.ICommand;

import EDU.oswego.cs.dl.util.concurrent.WriterPreferenceReadWriteLock;

/**
 * @author Oliver Plohmann
 *  
 */
public class AgentDispatcher implements INamedObject, IAgentDispatcher
{
	public static final String LOG_EVENT_ILLEGAL_AGENT_NAME = "ILLEGAL_AGENT_NAME"; //$NON-NLS-1$
	public static final String LOG_EVENT_NO_SUCH_AGENT = "NO_SUCH_AGENT"; //$NON-NLS-1$
	public static final String LOGGER_OBJECT_NAME = "AGENT_DISPATCHER"; //$NON-NLS-1$
	public static final String LOG_EVENT_RECEIVED_COMMAND = "RECEIVED_COMMAND"; //$NON-NLS-1$

	protected Map agentsByName = Collections.synchronizedMap(new Hashtable());
	protected String name = null;
	protected Logger logger = null;
	protected Node node = null;
	protected WriterPreferenceReadWriteLock lock = new WriterPreferenceReadWriteLock();

	protected AgentDispatcher()
	{
		super();
	}

	public AgentDispatcher(Node node, Logger logger, String nodeName)
	{
		super();
		this.node = node;
		name = nodeName;
		setLogger(logger);
	}

	protected void setLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setObjectName(LOGGER_OBJECT_NAME);
		logger.registerForEvents(registeredLogEvents());
	}

	/**
	 * @return
	 */
	protected List registeredLogEvents()
	{
		List events = new ArrayList();
		events.add(LOG_EVENT_ILLEGAL_AGENT_NAME);
		events.add(LOG_EVENT_NO_SUCH_AGENT);
		events.add(LOG_EVENT_RECEIVED_COMMAND);
		return events;
	}

	/**
	 * @param name
	 *            The name to set.
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	public void accept(CommandEnvelope commandEnvelope) throws IllegalArgumentException
	{
		commandEnvelope.arrivedAt(node);
		logReceivedCommand(commandEnvelope);				
		signalCommandArrived(commandEnvelope);
		
		if(commandEnvelope.isIgnore())
		{
			monitorCommandReceived(commandEnvelope);
			return;
		}
		
		monitorCommandReceivedAndWaiting(commandEnvelope);

		String destinationAgentName = commandEnvelope.getLocalizedDestinationAgentName();
		checkDestinationAgentName(destinationAgentName);
		IAgent agent = getAgent(destinationAgentName);
		commandEnvelope.setCurrentAgent(agent);
		
		if(commandEnvelope.isSynchronousResponse())
			acceptResponse(commandEnvelope, agent);
		else
			agent.accept(commandEnvelope);
	}

	/**
	 * @param commandEnvelope
	 */
	protected void signalCommandArrived(CommandEnvelope commandEnvelope)
	{
		IEventParams params = new CommandEnvelopeParams(commandEnvelope);
		Event event = new Event(CommandEnvelopeParams.EVENT_NAME_COMMAND_ARRIVED, params);
		node.getSimpleEventRegistry().occured(event);
	}

	/**
	 * @param commandEnvelope
	 */
	protected void monitorCommandReceived(CommandEnvelope env)
	{
		if(!env.isEventHandlerCommand())
		{
			List monitorAgents = node.getMonitorAgents();
			if(monitorAgents == null)
				return;
			
			ICommand cmd = node.getMonitorProducerAnchor().getReceivedCommandNotification(node.getName(), env);
			node.sendToMonitorAgents(cmd);				
		}
	}
	
	/**
	 * @param commandEnvelope
	 */
	protected void monitorCommandReceivedAndWaiting(CommandEnvelope env)
	{
		if(!env.isEventHandlerCommand())
		{
			List monitorAgents = node.getMonitorAgents();
			if(monitorAgents == null)
				return;
			
			ICommand cmd = node.getMonitorProducerAnchor().getReceivedCommandAndWaitingNotification(node.getName(), env);
			node.sendToMonitorAgents(cmd);				
		}
	}

	/**
	 * @param commandEnvelope
	 */
	protected void acceptResponse(CommandEnvelope commandEnvelope, IAgent agent)
	{
		CommandResponseEnvelope responseEnvelope = (CommandResponseEnvelope) commandEnvelope;
		responseEnvelope.setBack(true);
		AbstractAgent.setCommandEnvelope(responseEnvelope, agent, agent.getLogger(), node);
		node.getScheduler().getResponseScheduler().accept(commandEnvelope);
	}

	/**
	 * @param commandEnvelope
	 */
	protected void logReceivedCommand(CommandEnvelope commandEnvelope)
	{
		if (logger.isSubscribed(LOG_EVENT_RECEIVED_COMMAND))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_RECEIVED_COMMAND);
			String destination = commandEnvelope.getQualifiedDestinationAgentName();
			String origin = commandEnvelope.getOriginDestinationAgentName();
			StringBuffer buffer = new StringBuffer();
			buffer.append("{" + StringUtility.reversePackageName(commandEnvelope.getCommand().getQualifiedName())); //$NON-NLS-1$
			buffer.append("(id:" + commandEnvelope.getGlobalId().getValue() + ")}"); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append("{FROM:" + origin + "}{TO:" + destination + "}"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			entry.setArgument(buffer.toString());
			logger.info(entry);
		}
	}

	protected IAgent getAgent(String destinationAgentName) throws IllegalArgumentException
	{
		IAgent agent = null;
		try
		{
			lock.readLock().acquire();
			agent = (IAgent) agentsByName.get(destinationAgentName);
			if (agent == null)
			{
				if (logger.isSubscribed(LOG_EVENT_NO_SUCH_AGENT))
				{
					LogEntry entry = (LogEntry) logger.getNewLogEntry();
					entry.setEvent(LOG_EVENT_NO_SUCH_AGENT);
					entry.setArgument(destinationAgentName);
					logger.error(entry);
				}
				throw new IllegalArgumentException("no agent exists with name " //$NON-NLS-1$
				+StringUtility.print(destinationAgentName) + "!"); //$NON-NLS-1$
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return agent;
	}

	/**
	 * @param destinationAgentName
	 */
	protected boolean checkDestinationAgentName(String destinationAgentName) throws IllegalArgumentException
	{
		if (!StringUtility.isValid(destinationAgentName))
		{
			if (logger.isSubscribed(LOG_EVENT_ILLEGAL_AGENT_NAME))
			{
				LogEntry entry = (LogEntry) logger.getNewLogEntry();
				entry.setEvent(LOG_EVENT_ILLEGAL_AGENT_NAME);
				entry.setArgument(StringUtility.print(destinationAgentName));
				logger.error(entry);
			}
			throw new IllegalArgumentException("invalid destination agent name: " + destinationAgentName); //$NON-NLS-1$
		}
		return true;
	}

	protected void add(IAgent agent) throws IllegalArgumentException
	{
		String name = agent.getQualifiedName();
		if (!StringUtility.isValid(name))
			throw new IllegalArgumentException("agent name null!"); //$NON-NLS-1$
		
		try
		{
			lock.writeLock().acquire();
			if (agentsByName.containsKey(name))
				throw new IllegalArgumentException("agent with name " //$NON-NLS-1$
				+ StringUtility.print(name) + " already exists!"); //$NON-NLS-1$
			agentsByName.put(name, agent);
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
	}

	protected int exists(LocalAgentPath path)
	{
		try
		{
			lock.readLock().acquire();
			if (agentsByName.containsKey(path.getQualifiedPathName()))
				return NoSuchAgentException.NO_ERROR;
			int compareLevel = -1;
			Iterator it = agentsByName.keySet().iterator();
			while (it.hasNext())
			{
				String key = (String) it.next();
				LocalAgentPath keyPath = new LocalAgentPath(key);
				compareLevel = Math.max(compareLevel, keyPath.comparePath(path));
			}
			return compareLevel;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return NoSuchAgentException.NO_ERROR;
	}
	
	protected boolean existsExactly(LocalAgentPath path)
	{
		try
		{
			lock.readLock().acquire();
			return agentsByName.containsKey(path.getQualifiedPathName());
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return false;
	}

	protected IAgent remove(String qualifiedAgentName) throws IllegalArgumentException
	{
		if (!StringUtility.isValid(qualifiedAgentName))
			throw new IllegalArgumentException("invalid agent name: " //$NON-NLS-1$
			+ StringUtility.print(name) + "!"); //$NON-NLS-1$
		
		try
		{
			lock.writeLock().acquire();
			IAgent agent = (IAgent) agentsByName.remove(name);
			if (agent == null)
				throw new IllegalArgumentException("no agent exists with name " //$NON-NLS-1$
				+ StringUtility.print(name) + "!"); //$NON-NLS-1$
			return agent;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
		return null;
	}
	
	protected IAgent remove(IAgent agent) throws IllegalArgumentException
	{
		// TODO - workaround as agentsByName.remove(agent.getQualifiedName()) doesn't work
		// very strange but too time intensive to deal with right now.
		try
		{
			lock.writeLock().acquire();
			Iterator it = agentsByName.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				if(entry.getValue() == agent)
				{
					agentsByName.remove(entry.getKey());
					return agent;
				}
			}
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
		return null;
	}

	protected void stopAndRemove(String qualifiedAgentName)
	{
		IAgent agent = getAgent(qualifiedAgentName);
		agent.stop();		
		remove(agent);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param string
	 * @param string2
	 * @return
	 */
	protected List getAllAgentNames(String applicationName, String capabilityName)
	{
		try
		{
			lock.readLock().acquire();
			List agentNames = new ArrayList();
			String partialQualifiedName = applicationName + "." + capabilityName; //$NON-NLS-1$
			Iterator it = agentsByName.keySet().iterator();
			while (it.hasNext())
			{
				String key = (String) it.next();
				if(key.indexOf(partialQualifiedName) != -1)
					agentNames.add(key);		
			}
			return agentNames;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.node.IAgentDispatcher#isLocalAndIsIdle(org.almendra.commons.janet.cas.node.AgentPath)
	 */
	public boolean isLocalAndIsIdle(AgentPath agentPath)
	{
		if(!agentPath.isLocalTag())
			return false;
		
		String destinationAgentName = agentPath.getQualifiedPathName();
		checkDestinationAgentName(destinationAgentName);		
		IAgent agent = getAgent(destinationAgentName);
		if(agent.isIdle())
			return true;
		
		return false;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.node.IAgentDispatcher#getQueueSize(org.almendra.commons.janet.cas.node.LocalAgentPath)
	 */
	public long getQueueSize(LocalAgentPath agentPath) throws RemoteException
	{
		try
		{
			lock.readLock().acquire();
			List qualifiedAgentNames = getAllAgentNames(agentPath.getApplicationName(), agentPath.getCapabilityName());
			if(qualifiedAgentNames.isEmpty())
				return -1;
			long[] queueSizes = new long[qualifiedAgentNames.size()];
			for (int i = 0; i < qualifiedAgentNames.size(); i++)
			{
				String agentName = new LocalAgentPath((String) qualifiedAgentNames.get(i)).getAgentName();
				LocalAgentPath localAgentPath = new LocalAgentPath(agentPath);
				localAgentPath.setAgentName(agentName);
				IAgent agent = getAgent(localAgentPath.getQualifiedPathName());
				queueSizes[i] = agent.getQueueSize();
			}
			Arrays.sort(queueSizes);
			return queueSizes[queueSizes.length - 1];
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return -1;
	}

}
