/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.almendra.commons.util.xml.JDOMUtility;
import org.almendra.commons.util.xml.XMLParseException;
import org.jdom.Element;

/**
 * @author Oliver
 *  
 */
public class ApplicationDescriptor extends Descriptor
{
	public static final String XMLParentTag = "applications"; //$NON-NLS-1$
	public static final String XMLNodeTag = "application"; //$NON-NLS-1$
	public static final String XMLNameTag = "name"; //$NON-NLS-1$
	public static final String XMLAgentsTag = "agents"; //$NON-NLS-1$
	public static final String XMLAgentNameTag = "agentName"; //$NON-NLS-1$
	public static final String XMLExportedEventsTag = "exportedEvents"; //$NON-NLS-1$

	protected String name = null;
	protected List capabilityDescriptors = new LinkedList();
	protected List agentNames = new LinkedList();
	protected List exportedEvents = new LinkedList();

	public ApplicationDescriptor()
	{
		super();
	}

	/**
	 * @return Returns the name.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param name
	 *            The name to set.
	 */
	public void setName(String applicationName)
	{
		this.name = applicationName;
	}

	/**
	 * @return Returns the capabilities.
	 */
	public List getCapabilityDescriptors()
	{
		return capabilityDescriptors;
	}

	public List getCapabilities(AbstractApplication app) throws DescriptorParseException 
	{
		List capabilities = new LinkedList();
		Iterator it = capabilityDescriptors.iterator();
		while (it.hasNext())
		{
			CapabilityDescriptor desc = (CapabilityDescriptor) it.next();
			Capability capability = new Capability(app, desc.getName());
			desc.fill(capability);
			capabilities.add(capability);
		}
		return capabilities;
	}

	public void addCapabilityDescriptor(CapabilityDescriptor desc)
	{
		capabilityDescriptors.add(desc);
	}

	public static ApplicationDescriptor createAppDesc(Element element) throws XMLParseException
	{
		ApplicationDescriptor desc = new ApplicationDescriptor();
		desc.fill(element);
		return desc;
	}
	
	public void fill(Element element) throws XMLParseException
	{
		setName(element);
		setExportedEvents(element);
		fillCapabilities(element);
	}
	
	
	/**
	 * @param element
	 */
	protected void setExportedEvents(Element element) throws XMLParseException
	{
		Element eventsNode = JDOMUtility.getChild(element, XMLExportedEventsTag, false);
		if(eventsNode == null)
			return;
		List eventsNodes = JDOMUtility.getChildren(eventsNode, EventDescriptor.XMLEventTag);
		Iterator it = eventsNodes.iterator();
		while (it.hasNext())
		{
			Element eventNode = (Element) it.next();
			exportedEvents.add(createEventDescriptor(eventNode));
		}
	}

	/**
	 * @param eventNode
	 * @return
	 * @throws XMLParseException
	 */
	protected EventDescriptor createEventDescriptor(Element eventNode) throws XMLParseException
	{
		return EventDescriptor.create(eventNode);
	}

	/**
	 * @param element
	 */
	protected void setName(Element element) throws XMLParseException
	{
		name = element.getAttributeValue(XMLNameTag);
		if(name == null)
			throw new XMLParseException("no application name specified!"); //$NON-NLS-1$
	}

	/**
	 * @return
	 */
	protected boolean isMustHaveAgentsDefined()
	{
		return true;
	}

	protected void fillCapabilities(Element element) 
	throws XMLParseException
	{
		fillCapabilities(element, isMustHaveAgentsDefined()); 
	}
	
	protected void fillCapabilities(Element element, boolean mustHaveAgentsDefined) 
	throws XMLParseException
	{
		Element capabilities = JDOMUtility.getChild(element, CapabilityDescriptor.XMLParentTag);
		List capabilityNodes = JDOMUtility.getChildren(capabilities, CapabilityDescriptor.XMLNodeTag);
		List capabilityDescs = new LinkedList();

		for (int i = 0; i < capabilityNodes.size(); i++)
		{
			Element capabilityNode = (Element) capabilityNodes.get(i);
			capabilityDescs.add(CapabilityDescriptor.create(capabilityNode, mustHaveAgentsDefined));
		}
		
		setCapabilityDescriptors(capabilityDescs);
	}

	/**
	 * @param capabilityDescriptors
	 *            The capabilityDescriptors to set.
	 */
	public void setCapabilityDescriptors(List capabilityDescriptors)
	{
		this.capabilityDescriptors = capabilityDescriptors;
	}

	public void addAgentName(String agentName)
	{
		agentNames.add(agentName);
	}

	protected void fill(AbstractApplication app) throws DescriptorParseException
	{
		app.setName(getName());
		app.addCapabilities(getCapabilities(app));
		
		Iterator it = getExportedEvents().iterator();
		while (it.hasNext())
		{
			EventDescriptor desc = (EventDescriptor) it.next();
			desc.setApplicationName(getName());
		}
		
		app.addExportedEventDescriptors(getExportedEvents());
	}


	public SystemApplication toSystemApplication() throws DescriptorParseException
	{
		SystemApplication app = new SystemApplication();
		fill(app);
		return app;
	}

	public Application toApplication() throws DescriptorParseException
	{
		Application app = new Application();
		fill(app);
		return app;
	}
	/**
	 * @return Returns the exportedEvents.
	 */
	public List getExportedEvents()
	{
		return exportedEvents;
	}

	/**
	 * @param values
	 */
	public void visit(Map values)
	{
		Iterator it = capabilityDescriptors.iterator();
		while (it.hasNext())
		{
			CapabilityDescriptor desc = (CapabilityDescriptor) it.next();
			desc.visit(values);
		}
	}

}
