/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.almendra.commons.util.collection.StringVector;
import org.almendra.commons.util.xml.JDOMUtility;
import org.almendra.commons.util.xml.XMLParseException;
import org.jdom.Element;


/**
 * @author Oliver
 *  
 */
public class CapabilityDescriptor extends Descriptor
{
	public static final String XMLParentTag = "capabilities"; //$NON-NLS-1$
	public static final String XMLNodeTag = "capability"; //$NON-NLS-1$
	public static final String XMLNameTag = "name"; //$NON-NLS-1$
	public static final String XMLAgentsTag = "agents"; //$NON-NLS-1$
	public static final String XMLAgentTag = "agent"; //$NON-NLS-1$
	public static final String XMLInterpreterNodeTag = "interpreters"; //$NON-NLS-1$
	public static final String XMLInterpreterTag = "interpreter"; //$NON-NLS-1$
	public static final String XMLDescriptorTag = "descriptor"; //$NON-NLS-1$

	protected String name = null;
	protected String descriptor = null;
	protected List agentDescriptors = new ArrayList();
	protected List interpreterNames = new LinkedList();
	protected String executeWhenAppStarted = null;

	public CapabilityDescriptor()
	{
		super();
	}

	/**
	 * @return Returns the name.
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param name
	 *            The name to set.
	 */
	public void setName(String capabilityName)
	{
		this.name = capabilityName;
	}

	/**
	 * @return Returns the interpreters.
	 */
	public List getInterpretersNames()
	{
		return interpreterNames;
	}

	/**
	 * @return Returns the interpreters.
	 */
	public List getInterpreters()
	throws DescriptorParseException
	{
		List interpreters = new LinkedList();
		Iterator it = interpreterNames.iterator();
		while (it.hasNext())
		{
			String interpreterName = (String) it.next();
			try
			{
				Class interpreterClass = Class.forName(interpreterName);
				interpreters.add(interpreterClass.newInstance());
			}
			catch (ClassNotFoundException e)
			{
				System.out.println("no such interpreter named " + interpreterName + " found in node descriptor!"); //$NON-NLS-1$ //$NON-NLS-2$
				System.out.println(e.getMessage());
				e.printStackTrace();
				throw new DescriptorParseException(e.getMessage(), e.getCause());
			}
			catch (InstantiationException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				throw new DescriptorParseException(e.getMessage(), e.getCause());
			}
			catch (IllegalAccessException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				throw new DescriptorParseException(e.getMessage(), e.getCause());
			}
		}
		return interpreters;
	}

	public void addInterpreter(String fullyQualifiedClassName)
	{
		interpreterNames.add(fullyQualifiedClassName);
	}

	public static CapabilityDescriptor create(Element element, boolean mustHaveAgentsDefined) throws XMLParseException
	{
		CapabilityDescriptor desc = new CapabilityDescriptor();

		String name = element.getAttributeValue(XMLNameTag);
		if (name == null)
			throw new XMLParseException("no capability name specified!"); //$NON-NLS-1$
		else
			desc.setName(name);
		
		desc.descriptor = element.getAttributeValue(XMLDescriptorTag); // may be null	

		Element agentsNode = JDOMUtility.getChild(element, XMLAgentsTag, mustHaveAgentsDefined);

		if (mustHaveAgentsDefined && agentsNode == null)
			desc.throwNoAgentsDefined(name);
		
		if (agentsNode != null)
		{
			List agentNodes = JDOMUtility.getChildren(agentsNode, XMLAgentTag, mustHaveAgentsDefined);
			Iterator it = agentNodes.iterator();
			while (it.hasNext())
			{
				Element agentNode = (Element) it.next();
				AgentDescriptor agentDesc = AgentDescriptor.create(agentNode);
				desc.getAgentDescriptors().add(agentDesc);
			}
		}

		Element interpretersNode = JDOMUtility.getChild(element, XMLInterpreterNodeTag);
		List interpreterNodes = JDOMUtility.getChildren(interpretersNode, XMLInterpreterTag);

		for (int i = 0; i < interpreterNodes.size(); i++)
		{
			Element interpreterNode = (Element) interpreterNodes.get(i);
			desc.addInterpreter(JDOMUtility.getSingleChildContent(interpreterNode, XMLInterpreterTag));
		}

		return desc;
	}

	/**
	 * @param name
	 * @throws XMLParseException
	 */
	protected void throwNoAgentsDefined(String name) throws XMLParseException
	{
		throw new XMLParseException("no agents specified for capability " + name + "!"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @return Returns the agentNames.
	 */
	protected StringVector getAgentNames()
	{
		Iterator it = agentDescriptors.iterator();
		StringVector agentNames = new StringVector();
		while (it.hasNext())
		{
			AgentDescriptor desc = (AgentDescriptor) it.next();
			agentNames.add(desc.getName());
		}
		return agentNames;
	}

	/**
	 * @return Returns the agentNames.
	 */
	public void addAgentName(String agentName)
	{
		agentDescriptors.add(new AgentDescriptor(agentName));
	}

	/**
	 * @return Returns the executeWhenAppStarted.
	 */
	public String getExecuteWhenAppStarted()
	{
		return executeWhenAppStarted;
	}

	/**
	 * @param executeWhenAppStarted The executeWhenAppStarted to set.
	 */
	public void setExecuteWhenAppStarted(String executeWhenAppStarted)
	{
		this.executeWhenAppStarted = executeWhenAppStarted;
	}

	/**
	 * @param capability
	 */
	public void fill(Capability capability) throws DescriptorParseException
	{
		capability.getAgentDescriptors().addAll(getAgentDescriptors());
		capability.setDescriptorFilePathName(getDescriptor());
		capability.addInterpreters(getInterpreters());
		capability.setExecuteWhenAppStarted(getExecuteWhenAppStarted());		
	}

	/**
	 * @return Returns the agentDescriptors.
	 */
	public List getAgentDescriptors()
	{
		return agentDescriptors;
	}

	/**
	 * @param agentDescriptors The agentDescriptors to set.
	 */
	public void setAgentDescriptors(List agentDescriptors)
	{
		this.agentDescriptors = agentDescriptors;
	}

	/**
	 * 
	 */
	public void checkHasAgents() throws XMLParseException
	{
		if(getAgentNames().isEmpty())
			throwNoAgentsDefined(getName());
	}

	/**
	 * @return Returns the descriptor.
	 */
	public String getDescriptor()
	{
		return descriptor;
	}

	/**
	 * @param values
	 */
	public void visit(Map values)
	{
		if(descriptor != null)
		{
			String baseDir = (String) values.get(NodeDescriptor.DESCRIPTOR_BASE_DIR);
			descriptor = baseDir + File.separator + descriptor; //$NON-NLS-1$
		}
	}
}
