/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.node;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.xml.XMLParseException;
import org.jdom.Element;


/**
 * @author Oliver
 *
 */
public class LocalLoggerDescriptor
{	
	public static final String TAG_LOCAL_LOGGER = "localLogger"; //$NON-NLS-1$
	public static final String TAG_LOCAL_LOGGER_ATTRIBUTE_FILENAME = "fileName"; //$NON-NLS-1$
	public static final String TAG_LOCAL_LOGGER_ATTRIBUTE_PATHNAME = "pathName"; //$NON-NLS-1$
	public static final String TAG_LOCAL_LOGGER_ATTRIBUTE_ACTIVE = "active"; //$NON-NLS-1$
	public static final String TAG_LOCAL_LOGGER_ATTRIBUTE_BUFFERUNWRITTENMSGS = "bufferUnwrittenMsgs"; //$NON-NLS-1$
	public static final String TAG_MAX_SIZE = "maxSize"; //$NON-NLS-1$
	
	public static final String TAG_FILENAME_SAME_AS_NODENAME = "SameAsNodeName"; //$NON-NLS-1$
	public static final String TAG_FILENAME_SAME_AS_STARTUPDIR = "StartupDir"; //$NON-NLS-1$
	public static final String LOG_FILE_EXTENSION = ".log"; //$NON-NLS-1$
	public static final String LABEL_KB = "kb"; //$NON-NLS-1$
	public static final String LABEL_MB = "mb"; //$NON-NLS-1$
		
	protected String fileName = null;
	protected String pathName = null;
	protected String maxSize = null;
	protected boolean isActive = false;
	protected boolean isBufferUnwrittenMsgs = false;
	
	/**
	 * Creates a new <code>LocalLoggerDescriptor</code> object.
	 */
	public LocalLoggerDescriptor()
	{
		super();
	}

	/**
	 * @return Returns the fileName.
	 */
	public String getFileName(String nodeName)
	{
		if(fileName == null)
			return null;
		if(fileName.equals(TAG_FILENAME_SAME_AS_NODENAME))
			return nodeName + LOG_FILE_EXTENSION;
		if(fileName.indexOf(LOG_FILE_EXTENSION) == -1)
			return fileName + LOG_FILE_EXTENSION;
		return fileName;
	}
	
	/**
	 * @return Returns the fileName.
	 */
	public String getFileName()
	{
		return fileName;
	}

	/**
	 * @param fileName The fileName to set.
	 */
	public void setFileName(String fileName)
	{
		this.fileName = fileName;
	}

	/**
	 * @return Returns the isActive.
	 */
	public boolean isActive()
	{
		return isActive;
	}

	/**
	 * @param isActive The isActive to set.
	 */
	public void setActive(boolean isActive)
	{
		this.isActive = isActive;
	}

	/**
	 * @return Returns the pathName.
	 */
	public String getPathName()
	{
		if(pathName == null || pathName.equals(TAG_FILENAME_SAME_AS_STARTUPDIR))
		{
			return ""; //$NON-NLS-1$
		}
		return pathName;
	}

	/**
	 * @param pathName The pathName to set.
	 */
	public void setPathName(String pathName)
	{
		this.pathName = pathName;
	}
	
	public boolean isFileNameSameAsNodeName()
	{
		return TAG_FILENAME_SAME_AS_NODENAME.equals(getFileName());
	}
	
	public boolean isPathNameSameAsStartupDir()
	{
		return TAG_FILENAME_SAME_AS_STARTUPDIR.equals(getPathName());
	}
	

	/**
	 * @param element
	 */
	public void fill(Element element) throws XMLParseException
	{		
		Element loggerElement = element.getChild(TAG_LOCAL_LOGGER);
		if (loggerElement == null)
			return;
	
		fillMaxSize(loggerElement);				
		
		String fileName = loggerElement.getAttributeValue(TAG_LOCAL_LOGGER_ATTRIBUTE_FILENAME);
	
		if(fileName == null)
			throw new XMLParseException("no file name specified for local logger");	 //$NON-NLS-1$
		
		setFileName(fileName);
		
		if(!fileName.equals(TAG_FILENAME_SAME_AS_NODENAME))
		{
			if(fileName.lastIndexOf(".") == -1) //$NON-NLS-1$
				setFileName(fileName + LOG_FILE_EXTENSION); //$NON-NLS-1$
		}
		
		String pathName = loggerElement.getAttributeValue(TAG_LOCAL_LOGGER_ATTRIBUTE_PATHNAME);
		
		if(pathName == null)
			setPathName(TAG_FILENAME_SAME_AS_STARTUPDIR);
		else
			setPathName(pathName);
		
		String active = loggerElement.getAttributeValue(TAG_LOCAL_LOGGER_ATTRIBUTE_ACTIVE);
		
		if(active == null)	
		{	
			setActive(false);
		}
		else
		{
			setActive(Boolean.valueOf(active).booleanValue());
		}	
		
		String bufferUnwrittenMsgs = loggerElement.getAttributeValue(TAG_LOCAL_LOGGER_ATTRIBUTE_BUFFERUNWRITTENMSGS);
		
		if(bufferUnwrittenMsgs == null)
		{
			setBufferUnwrittenMsgs(false);
		}
		else
		{
			setBufferUnwrittenMsgs(Boolean.valueOf(bufferUnwrittenMsgs).booleanValue());
		}
		
	}
	
	/**
	 * @param loggerElement
	 */
	protected void fillMaxSize(Element loggerElement) throws XMLParseException
	{
		maxSize = loggerElement.getAttributeValue(TAG_MAX_SIZE);
		if(!StringUtility.isValid(maxSize))
			return; // value not set - no problem as value not compulsory
		
		maxSize = maxSize.toLowerCase();
		String maxSizeWithoutMeasure = maxSize;
		
		if(maxSize.endsWith(LABEL_KB))
			maxSizeWithoutMeasure = StringUtility.remove(maxSize, LABEL_KB);
		if(maxSize.endsWith(LABEL_MB))
			maxSizeWithoutMeasure = StringUtility.remove(maxSize, LABEL_MB);
			
		try
		{
			Long.valueOf(maxSizeWithoutMeasure);
		}
		catch (Exception e)
		{
			throw new XMLParseException("invalid value for maxSize: " + maxSize); //$NON-NLS-1$
		}			
	}

	public String getFilePathName(String nodeName)
	{
		if(!isActive())
			return null;
		String tempFileName = getFileName(nodeName);
		if(tempFileName == null)
			return null;
		String tempPathName = getPathName(); 
		if(tempPathName.length() == 0)
			return tempFileName;
		return getPathName() + "\\" + tempFileName;  //$NON-NLS-1$
	}

	/**
	 * @return Returns the isBufferUnwrittenMsgs.
	 */
	public boolean isBufferUnwrittenMsgs()
	{
		return isBufferUnwrittenMsgs;
	}

	/**
	 * @param isBufferUnwrittenMsgs The isBufferUnwrittenMsgs to set.
	 */
	public void setBufferUnwrittenMsgs(boolean isBufferUnwrittenMsgs)
	{
		this.isBufferUnwrittenMsgs = isBufferUnwrittenMsgs;
	}

	/**
	 * @return Returns the maxSize.
	 */
	public String getMaxSizeString()
	{
		return maxSize;
	}
	
	public long getMaxSize()
	{
		if(!StringUtility.isValid(maxSize))
			return -1;
		
		String maxSizeWithoutMeasure = maxSize;
		int factor = 1;
		
		if(maxSize.endsWith(LABEL_KB))
		{	
			maxSizeWithoutMeasure = StringUtility.remove(maxSize, LABEL_KB);
			factor = 1024;
		}
			
		if(maxSize.endsWith(LABEL_MB))
		{
			maxSizeWithoutMeasure = StringUtility.remove(maxSize, LABEL_MB);
			factor = 1024 * 1024;
		}
						
		try
		{
			return Long.valueOf(maxSizeWithoutMeasure).longValue() * factor;
		}
		catch (Exception e)
		{
			return -1;
		}		
	}

}
