/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.node;

import java.io.IOException;
import java.util.Iterator;

import javax.swing.JOptionPane;

import org.almendra.commons.util.logging.ErrorLogger;
import org.almendra.commons.util.logging.ILogPrinter;
import org.almendra.janet.cas.node.LogEntry;


public class Logger extends org.almendra.commons.util.logging.Logger
{
	protected static final String EVENT_EXCEPTION_OCCURRED = "EXCEPTION_OCCURRED"; //$NON-NLS-1$
	
	/**
	 * LogEntry has holds the default values that should be copied
	 * into any new instance on <code>LogEntry</code> when using
	 * the method <code>fillEntryTemplate</code>.
	 */
	protected LogEntry defaultValues = new LogEntry(Logger.INFO);
	
	/**
	 * Holds information read from the node descriptor xml file where the
	 * user specified whether a local log file should be written, what the
	 * log file's name is and where it is located. 
	 */
	protected LocalLoggerDescriptor loggerDescriptor = null;
	
	/**
	 * Creates a new <code>Logger</code> object.
	 */
	public Logger()
	{
		super();
	}
	
	/**
	 * Creates a new <code>Logger</code> object.
	 */
	public Logger(boolean bufferUnwrittenMsgs, long localLogFileMaxSize)
	{
		super(bufferUnwrittenMsgs, localLogFileMaxSize);
	}

	/**
	 * Creates a new <code>Logger</code> object.
	 * @param defaultValues holds default values
	 */
	public Logger(LogEntry defaultValuesEntry)
	{
		super();
		this.defaultValues = defaultValuesEntry;
	}
	
	/**
	 * Creates a new <code>Logger</code> object.
	 */
	public Logger(ILogPrinter logPrinter)
	{
		super(logPrinter);
	}

	/**
	 * Creates a new <code>Logger</code> object.
	 */
	public Logger(ILogPrinter logPrinter, boolean bufferUnwrittenMsgs)
	{
		super(logPrinter, bufferUnwrittenMsgs);
	}

	public void fillChildLogger(Logger childLogger)
	{
		childLogger.parent = this;
		childLogger.defaultValues = defaultValues;
		childLogger.logPrinter = logPrinter;
		childLogger.registeredEvents = registeredEvents;
		childLogger.subscribedEvents = subscribedEvents;
	}

	/**
	 * @return Returns the defaultValues.
	 */
	public LogEntry getDefaultValues()
	{
		return defaultValues;
	}

	/**
	 * @param defaultValues The defaultValues to set.
	 */
	protected void setDefaultValues(LogEntry defaultValuesEntry)
	{
		this.defaultValues = defaultValuesEntry;
	}

	/**
	 * Returns a child logger. A child logger inherits all the attribute values
	 * of the parent logger, but new events can be registered without affecting
	 * the parent logger. 
	 * @param logger the parent logger
	 * @return the child logger
	 */
	public static Logger getChildLogger(Logger logger)
	{
		if(logger == null)
			return null;
		LogEntry defaultEntry;
		try
		{
			defaultEntry = (LogEntry) logger.defaultValues.clone();
		}
		catch (CloneNotSupportedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			return null;
		}
		Logger childLogger = new Logger(defaultEntry);
		getChildLogger(logger, childLogger);
		return childLogger;
	}
	
	/**
	 * @return Return a new <code>LogEntry</code> object.
	 */
	public org.almendra.commons.util.logging.LogEntry getNewLogEntry()
	{
		return new LogEntry(defaultValues);
	}
	
	/**
	 * Set the node name for this logger and all it's children since
	 * the node name is the same for all loggers of the same node. 
	 * @param name
	 */
	public void setNodeName(String nodeName)
	{
		defaultValues.setNodeName(nodeName);
		Iterator it = children.iterator();
		while (it.hasNext())
		{
			Logger logger = (Logger) it.next();
			logger.setNodeName(nodeName);
		}
	}

	/**
	 * @return Returns the loggerDescriptor.
	 */
	public LocalLoggerDescriptor getLoggerDescriptor()
	{
		return loggerDescriptor;
	}

	/**
	 * @param loggerDescriptor The loggerDescriptor to set.
	 */
	public void setLoggerDescriptor(LocalLoggerDescriptor loggerDescriptor)
	{
		this.loggerDescriptor = loggerDescriptor;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.Logger#changeLocalLogFileName(java.lang.String)
	 */
	public boolean changeLocalLogFileName(String nodeName) throws IOException
	{
		if(!loggerDescriptor.isActive())
			return true;
		return super.changeLocalLogFileName(loggerDescriptor.getFilePathName(nodeName));
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.Logger#setLocalLogFileName(java.lang.String)
	 */
	public void setLocalLogFileName(String filePathName) throws IOException
	{
		String tempFilePathName = filePathName;
		if(tempFilePathName.indexOf("?") != -1) //$NON-NLS-1$
			tempFilePathName = tempFilePathName.replace('?', '$');
		super.setLocalLogFileName(tempFilePathName);
	}
	
	public void exception(Throwable ex)
	{
		LogEntry entry = (LogEntry) getNewLogEntry();
		entry.setEvent(EVENT_EXCEPTION_OCCURRED); //$NON-NLS-1$
		entry.setException(ex);
		exception(entry);
		String msg = ex.toString();
		if(defaultValues != null)
			msg = defaultValues.getNodeName() + ": " + msg; //$NON-NLS-1$
		ErrorLogger.exception(ex);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.Logger#init()
	 */
	protected void init()
	{
		super.init();		
		subscribedEvents.add(EVENT_EXCEPTION_OCCURRED);
	}

}
