/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.io.IOException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.almendra.commons.INamedObject;
import org.almendra.commons.events.Event;
import org.almendra.commons.events.EventRegistry;
import org.almendra.commons.events.IEventHandler;
import org.almendra.commons.events.IEventRegistry;
import org.almendra.commons.events.ISimpleEventRegistry;
import org.almendra.commons.events.IVetoableEventRegistry;
import org.almendra.commons.events.SimpleEventRegistry;
import org.almendra.commons.events.SimpleVetoableEventRegistry;
import org.almendra.commons.threads.ISimpleActiveObject;
import org.almendra.commons.threads.ThreadUtility;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.exception.AlreadyRegisteredException;
import org.almendra.commons.util.globalId.IUniqueGlobalIdObject;
import org.almendra.commons.util.globalId.UniqueId;
import org.almendra.commons.util.globalId.UniqueIdGenerator;
import org.almendra.commons.util.inet.InetUtility;
import org.almendra.commons.util.inet.RMIUtility;
import org.almendra.commons.util.inet.RegistryLocation;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.commons.util.logging.LoggerAccessor;
import org.almendra.commons.util.swing.BillboardPanelController;
import org.almendra.commons.util.swing.TextAreaPanelController;
import org.almendra.janet.cas.agents.IAgent;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.agents.MultiAgentProxy;
import org.almendra.janet.cas.agents.SystemAgent;
import org.almendra.janet.cas.commands.DeregisterApplicationCommand;
import org.almendra.janet.cas.commands.DeregisterNodeCommand;
import org.almendra.janet.cas.commands.RegisterApplicationCommand;
import org.almendra.janet.cas.events.EventHandlerEnvelope;
import org.almendra.janet.cas.events.EventRegistryEnvelope;
import org.almendra.janet.cas.events.HandlerList;
import org.almendra.janet.cas.events.IClusterEventRegistry;
import org.almendra.janet.cas.events.ILocalClusterEventRegistry;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.events.LocalClusterEventRegistry;
import org.almendra.janet.cas.events.LocalClusterEventRegistryRemote;
import org.almendra.janet.cas.handlers.ApplicationDeregisteredParams;
import org.almendra.janet.cas.handlers.ApplicationRegisteredParams;
import org.almendra.janet.cas.handlers.AskVetoShutdownEventParams;
import org.almendra.janet.cas.handlers.ForeignApplicationDeregisteredParams;
import org.almendra.janet.cas.handlers.ForeignApplicationRegisteredParams;
import org.almendra.janet.cas.handlers.NodeAboutToShutdownParams;
import org.almendra.janet.cas.handlers.NodeShutdownParams;
import org.almendra.janet.cas.handlers.NodeStartedEventHandler;
import org.almendra.janet.cas.handlers.NodeStartupEventParams;
import org.almendra.janet.cas.schedulers.Arbitrator;
import org.almendra.janet.cas.schedulers.CommandScheduler;
import org.almendra.janet.cas.scheduling.Acknowledge;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.spaces.ClusterSpace;
import org.almendra.janet.cas.spaces.IObjectSpace;
import org.almendra.janet.cas.spaces.IObjectSpaceProvider;
import org.almendra.janet.cas.spaces.ISerializableObjectSpace;
import org.almendra.janet.cas.spaces.ObjectSpace;


/**
 * @author Oliver
 *  
 */

public class Node extends Object implements INamedObject, ISimpleActiveObject, ILoggable, IUniqueGlobalIdObject, IObjectSpaceProvider
{
	public static final String LOG_EVENT_NODE_STARTING_UP = "NODE_STARTING_UP"; //$NON-NLS-1$
	public static final String LOG_EVENT_NODE_STARTED = "NODE_STARTED"; //$NON-NLS-1$
	public static final String LOG_EVENT_NODE_STOPPED = "NODE_STOPPED"; //$NON-NLS-1$
	public static final String LOG_EVENT_NODE_NAME_CHANGED = "NODE_NAME_CHANGED"; //$NON-NLS-1$
	public static final String LOG_EVENT_ABOUT_TO_DISCONNECT_FROM_ALL_NODES = "ABOUT_TO_DISCONNECT_FROM_ALL_NODES"; //$NON-NLS-1$
	public static final String LOG_EVENT_DISCONNECTED_FROM_ALL_NODES = "DISCONNECTED_FROM_ALL_NODES"; //$NON-NLS-1$
	public static final String LOG_EVENT_TIMEOUT_OCCURED_DISCONNECTING_FROM_ALL_NODES = "TIMEOUT_OCCURED_DISCONNECTING_FROM_ALL_NODES"; //$NON-NLS-1$
	public static final String LOG_EVENT_EXCEPTION_OCCURED_DISCONNECTING_FROM_ALL_NODES = "EXCEPTION_OCCURED_DISCONNECTING_FROM_ALL_NODES"; //$NON-NLS-1$
	public static final String LOG_EVENT_REQUESTING_PERMISSION_TO_SHUT_DOWN = "REQUESTING_PERMISSION_TO_SHUT_DOWN"; //$NON-NLS-1$
	public static final String LOG_EVENT_RECEIVED_PERMISSION_TO_SHUT_DOWN = "RECEIVED_PERMISSION_TO_SHUT_DOWN"; //$NON-NLS-1$
	public static final String LOGGER_OBJECT_NAME = "NODE"; //$NON-NLS-1$
	public static final String RMI_LABEL_PREFIX = "Janet.node."; //$NON-NLS-1$
	public static final String NO_TITLE_PREFIX_FOUND = "NoTitlePrefixFound."; //$NON-NLS-1$

	public static final int SHUTDOWN_TIMEOUT = 5000; // ms
	public static final int DISCONNECT_TIMEOUT = 10000; // ms

	/**
	 * clusterwide unique node name
	 */
	protected String name = null;

	protected UniqueId id = UniqueIdGenerator.next();

	protected Logger logger = null;

	protected SystemAgent systemAgent = null;

	protected SchedulerAnchor schedulerAnchor = null;

	protected ApplicationRegistry applicationRegistry = null;

	protected AgentDispatcher localAgentDispatcher = null;

	protected AgentDispatcherRemote agentDispatcher = null;

	protected IEventRegistry eventRegistry = new EventRegistry();
	
	protected ISimpleEventRegistry simpleEventRegistry = new SimpleEventRegistry();

	protected IObjectSpace objectSpace = new ObjectSpace();

	protected String version = "?.?"; //$NON-NLS-1$
	
	protected Map clusterSpacesByName = new Hashtable();
	
	protected Map clusterEventRegistriesByName = new Hashtable();
	
	protected Map localClusterEventRegistriesByName = new Hashtable();
	
	protected Object monitorAgentsLock = new Object();
	
	protected List monitorAgents = null;
	
	protected IVetoableEventRegistry vetoableEventRegistry = new SimpleVetoableEventRegistry();   

	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}

	protected Node()
	{
		super();
	}

	public Node(NodeDescriptor nodeDescriptor) throws DescriptorParseException, NodeStartupException
	{
		super();
		init(nodeDescriptor);
	}

	/**
	 * @param nodeDescriptor
	 */
	protected void init(NodeDescriptor nodeDescriptor) throws DescriptorParseException, NodeStartupException
	{
		setVersion(nodeDescriptor.getVersionName());
		init(nodeDescriptor.getSystemApplication(), nodeDescriptor.getLocalLoggerDescriptor());
		getSystemApplication().setCentralRegistryLocation(nodeDescriptor.getCentralRegistryLocation());
		getSystemApplication().setLocalRegistryLocation(nodeDescriptor.getLocalRegistryLocation());
		getSystemApplication().setClusterSpaceDescriptors(nodeDescriptor.getClusterSpaceDescriptors());
		getSystemApplication().setClusterEventRegistryDescriptors(nodeDescriptor.getClusterEventRegistryDescriptors());
		storeApplications(nodeDescriptor, getSystemApplication());
		getSystemApplication().getObjectSpace().put(NodeDescriptor.LABEL, nodeDescriptor);
	}

	protected void storeApplications(NodeDescriptor nodeDescriptor, SystemApplication sysApp) throws DescriptorParseException
	{
		sysApp.getObjectSpace().put(NodeAnchor.REGISTRY_KEY_STARTUP_APPLICATIONS, nodeDescriptor.getApplications());
	}

	/**
	 * @return
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Set the node name. Tell all dependent objects that need to know the node
	 * name about the it.
	 * 
	 * @param newName
	 *            the new node name
	 */
	public void setName(String newName)
	{
		String currentName = name;
		name = newName;

		if (agentDispatcher != null)
			agentDispatcher.setName(newName);

		if (logger == null)
			return;

		logger.setNodeName(newName);

		if (currentName != null && !currentName.equals(newName))
		{
			try
			{
				logger.changeLocalLogFileName(newName);
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				logger.exception(e);
			}

			if (logger.isSubscribed(LOG_EVENT_NODE_NAME_CHANGED))
			{
				LogEntry entry = (LogEntry) logger.getNewLogEntry();
				entry.setEvent(LOG_EVENT_NODE_NAME_CHANGED);
				entry.setArgument(currentName + "->" + newName); //$NON-NLS-1$
				logger.info(entry);
			}
		}
	}

	protected void init(SystemApplication sysApp, LocalLoggerDescriptor loggerDescriptor) throws NodeStartupException
	{
		setName(getTemporaryNodeName(getName()));
		init(loggerDescriptor);
		applicationRegistry = new ApplicationRegistry(this, logger);
		getLogAccessor().subscribeAll(); // TODO - for testing only, has to be
		// done explicitly by the user lateron
		schedulerAnchor = new SchedulerAnchor(this, logger);
		systemAgent = sysApp.createAgent(this, logger);
		logger.registerForEvents(registeredLogEvents());
		initAgentDispatcher();
		getLogAccessor().subscribeAll(); // TODO - for testing only, has to be
		// done explicitly by the user lateron
		localAgentDispatcher.add(systemAgent);
		registerEvents();
	}
	
	/**
	 * Answer a temporary node name if current node name is null. A node does node know it's final 
	 * name before it has finished the startup process. Return a network-wide unique temporary node name.    
	 * @param currentNodeNameOrNull
	 * @return the temporary node name
	 */	
	public String getTemporaryNodeName(String currentNodeNameOrNull)
	{
		if(currentNodeNameOrNull != null)
			return currentNodeNameOrNull;
		StringBuffer buffer = new StringBuffer();
		buffer.append(InetUtility.getHostName());
		buffer.append("-?-"); //$NON-NLS-1$
		buffer.append(UniqueIdGenerator.getMillisWithRandom());
		return buffer.toString();
	}

	protected void init(LocalLoggerDescriptor loggerDescriptor)
	{
		logger = new Logger(loggerDescriptor.isBufferUnwrittenMsgs(), loggerDescriptor.getMaxSize());

		logger.setLoggerDescriptor(loggerDescriptor);
		logger.getDefaultValues().setNodeName(getName());
		logger.getDefaultValues().setWorkstationName(InetUtility.getHostName());
		logger.getDefaultValues().setObjectName(LOGGER_OBJECT_NAME);

		try
		{
			if (loggerDescriptor.isActive())
				logger.setLocalLogFileName(loggerDescriptor.getFilePathName(getName()));
		}
		catch (IOException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}

		logger.registerForEvents(registeredLogEvents());
		getLogAccessor().subscribeAll();

		if (logger.isSubscribed(LOG_EVENT_NODE_STARTING_UP))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_NODE_STARTING_UP);
			entry.setArgument(getName());
			logger.info(entry);
		}
	}

	protected void initAgentDispatcher() throws NodeStartupException
	{
		try
		{
			localAgentDispatcher = new AgentDispatcher(this, logger, getName());
			agentDispatcher = new AgentDispatcherRemote(localAgentDispatcher);
		}
		catch (RemoteException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			StringBuffer msg = NodeStartupException.getMessageHeader("cannot create agent dispatcher"); //$NON-NLS-1$
			msg.append(e.getMessage());
			throw new NodeStartupException(msg.toString(), e.getCause());
		}
	}

	protected void registerEvents()
	{	
		IEventHandler handler = new NodeStartedEventHandler(this);
		EventHandlerEnvelope env = new EventHandlerEnvelope(handler, getSystemAgentPath());		
		HandlerList nodeStartedEventRecipients = new HandlerList(env);
		
		ApplicationAccessor app = getSystemApplication().getAccessor();
		ILocalEventRegistry eventRegistryEnv = getSystemApplication().getEventRegistry();
		eventRegistryEnv.register(NodeStartupEventParams.EVENT_NAME, nodeStartedEventRecipients, app);
		
		HandlerList foreignAppRegisteredEventRecipients = new HandlerList();
		eventRegistryEnv.register(ForeignApplicationRegisteredParams.EVENT_NAME, foreignAppRegisteredEventRecipients, app);
	}

	public void start()
	{
		getSystemApplication().start();
		schedulerAnchor.start();
		logger.info(LOG_EVENT_NODE_STARTED);
	}

	public void stop()
	{
		signalShutdown();		
		schedulerAnchor.stop();
		logger.info(LOG_EVENT_NODE_STOPPED);
		logger.close(); 
		if(getNodeDescriptor().isExitVMOnNodeShutdown())
			System.exit(0);
	}
	protected void signalShutdown()
	{
		NodeShutdownParams params = new NodeShutdownParams(getAccessor());
		Event event = new Event(NodeShutdownParams.EVENT_NAME, params);
		eventRegistry.occured(event);
		
		// TODO - implement proper mechanism to wait till all handlers have been executed. 
		ThreadUtility.sleep(2000); // wait till all handlers executed
	}

	/**
	 * @return
	 */
	public CommandScheduler getSystemScheduler()
	{
		return schedulerAnchor.supremeScheduler;
	}

	/**
	 * @return
	 */
	public IAgent getSystemAgent()
	{
		return systemAgent;
	}
	
	/**
	 * @return
	 */
	public AgentPath getSystemAgentPath()
	{
		return systemAgent.getAgentPath();
	}
	
	public IAgentProxy getSystemAgentProxy() 
	{
		return getSystemAgent().getAgentProxy();
	}

	/**
	 * @return
	 */
	public SchedulerAnchor getScheduler()
	{
		return schedulerAnchor;
	}

	/**
	 * @return
	 */
	public ApplicationRegistry getApplicationRegistry()
	{
		return applicationRegistry;
	}

	public boolean isIdle()
	{
		return getScheduler().isIdle();
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		List events = new ArrayList();
		events.add(LOG_EVENT_NODE_STARTED);
		events.add(LOG_EVENT_NODE_STOPPED);
		events.add(LOG_EVENT_NODE_NAME_CHANGED);
		events.add(LOG_EVENT_DISCONNECTED_FROM_ALL_NODES);
		events.add(LOG_EVENT_ABOUT_TO_DISCONNECT_FROM_ALL_NODES);
		events.add(LOG_EVENT_EXCEPTION_OCCURED_DISCONNECTING_FROM_ALL_NODES);
		events.add(LOG_EVENT_TIMEOUT_OCCURED_DISCONNECTING_FROM_ALL_NODES);
		events.add(LOG_EVENT_NODE_STARTING_UP);
		events.add(LOG_EVENT_REQUESTING_PERMISSION_TO_SHUT_DOWN);
		events.add(LOG_EVENT_RECEIVED_PERMISSION_TO_SHUT_DOWN);
		return events;
	}

	public LoggerAccessor getLogAccessor()
	{
		return logger.getPublicAccessor();
	}

	public NodeAccessor getAccessor()
	{
		return new NodeAccessor(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.globalId.IUniqueGlobalIdObject#getGlobalId()
	 */
	public UniqueId getGlobalId()
	{
		return id;
	}

	public Arbitrator getArbitrator()
	{
		return getScheduler().getArbitrator();
	}
	
	public Arbitrator getSystemArbitrator()
	{
		return getScheduler().getSystemArbitrator();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.ISimpleActiveObject#isActive()
	 */
	public boolean isActive()
	{
		return !getScheduler().isIdle();
	}

	/**
	 * @return Returns the local agentDispatcher.
	 */
	public AgentDispatcherRemote getAgentDispatcher()
	{
		return agentDispatcher;
	}

	public AbstractApplication getApplication(String applicationName)
	{
		return applicationRegistry.getApplication(applicationName);
	}

	public SystemApplication getSystemApplication()
	{
		return getApplicationRegistry().getSystemApplication();
	}

	/**
	 * @return
	 */
	public String getViewTitle()
	{
		String viewTitlePrefix = getSystemAnchor().getViewTitle();
		if (viewTitlePrefix == null)
			return NO_TITLE_PREFIX_FOUND;
		return viewTitlePrefix;
	}

	/**
	 * @return
	 */
	public AbstractNodeAnchor getSystemAnchor()
	{
		return (AbstractNodeAnchor) getSystemApplication().getObjectSpace().get(NodeAnchor.LABEL);
	}

	/**
	 * @param path
	 * @return
	 */
	public IAgentDispatcher getAgentDispatcher(AgentPath path) throws NoSuchAgentException
	{
		if (isLocalPath(path))
		{
			// agent resides on node itself
			int compareLevel = localAgentDispatcher.exists(new LocalAgentPath(path));
			if (compareLevel == NoSuchAgentException.NO_ERROR)
				return getAgentDispatcher();
			throw new NoSuchAgentException(path, compareLevel, path.getElement(compareLevel));
		}
		else
		{
			return getSystemAnchor().getClusterImage().getAgentDispatcher(path);
		}
	}

	/**
	 * @param path
	 * @return
	 */
	public boolean isLocalPath(AgentPath path)
	{
		if(path.isLocalTag())
			return true;
		return getName().equalsIgnoreCase(path.getNodeName());
	}

	public UniqueId getUniqueId()
	{
		AssertUtility.assertTrue(getName() != null);
		return UniqueIdGenerator.next(getName());
	}

	/**
	 * @return Returns the localAgentDispatcher.
	 */
	public AgentDispatcher getLocalAgentDispatcher()
	{
		return localAgentDispatcher;
	}

	/**
	 * @return
	 */
	public String getRMIName()
	{
		return RMI_LABEL_PREFIX + getName();
	}

	public boolean disconnectRequest()
	{		
		Event event = new Event(AskVetoShutdownEventParams.EVENT_NAME, new AskVetoShutdownEventParams(getName()));
		List results = getVetoableEventRegistry().askAnyMustVeto(event);
		boolean ignore = getVetoableEventRegistry().askUserIgnoreVetos(results, null, "Shutdown request has been vetoed!", "Accept veto and abort shutdown?"); //$NON-NLS-1$ //$NON-NLS-2$
		if(ignore)
			return false;
		
		AbstractNodeAnchor nodeAnchor = getNodeAnchor();
		if(!nodeAnchor.mayDisconnect())
			return false;
		
		NodeAboutToShutdownParams params = new NodeAboutToShutdownParams(getAccessor());
		event = new Event(NodeAboutToShutdownParams.EVENT_NAME, params);
		eventRegistry.occured(event);
		
		nodeAnchor.disconnectRequest(this);
		return true;
	}
	
	public AgentPath getAgentPath(String applicationName, String capabilityName, String agentName)
	{
		// TODO - verifiy that path is valid otherwise throw exception
		return new AgentPath(getName(), applicationName, capabilityName, agentName);
	}

	public void disconnectFromCentral()
	{
		DeregisterNodeCommand command = new DeregisterNodeCommand(getName(), getRMIName());
		IAgentProxy centralAgent = getSystemApplication().getCentralAgent();

		try
		{
			centralAgent.accept(command);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logDisconnect(e);
			logger.exception(e);			
			return;
		}
		
	}
	
	public void disconnectFromNodesWithoutCentral()
	{
		try
		{			
			MultiAgentProxy foreignSystemAgents = getSystemApplication().getAllForeignSystemAgents();
			int numAgents = foreignSystemAgents.size();
			if (numAgents == 0)
				return;

			// wait for numAgents acks from the recipient agents
			DeregisterNodeCommand command = new DeregisterNodeCommand(getName());
			Acknowledge ack = new Acknowledge();			
			foreignSystemAgents.accept(command, ack);
			ack.acquire(numAgents * DISCONNECT_TIMEOUT);

			if (ack.isTimeout())
			{
				logger.error(LOG_EVENT_TIMEOUT_OCCURED_DISCONNECTING_FROM_ALL_NODES);
				return;
			}

			if (ack.hasException())
			{
				logDisconnect(ack.getException());
				return;
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			logDisconnect(e);
		}
	}

	/**
	 * @param exception
	 */
	protected void logDisconnect(Exception ex)
	{
		LogEntry entry = (LogEntry) logger.getNewLogEntry();
		entry.setEvent(LOG_EVENT_EXCEPTION_OCCURED_DISCONNECTING_FROM_ALL_NODES);
		entry.setArgument(ex.getMessage());
		logger.error(entry);
	}

	public boolean shutdown()
	{
		return disconnectRequest();
	}

	/**
	 * Register the application in the node's application registry, create all
	 * agents, and start their schedulers. Send a notifications to all nodes in
	 * the cluster that a new application has been registered at this node.
	 * 
	 * @param app
	 * @param nodeNames
	 *            names of all nodes where to register, all foreign nodes if
	 *            null.
	 * @return @throws
	 *         AlreadyRegisteredException
	 * @throws IllegalArgumentException
	 */
	public Application registerApplication(Application app, List nodeNames)
		throws AlreadyRegisteredException, IllegalArgumentException, RemoteException
	{
		if (app == null)
			throw new IllegalArgumentException("application is null!"); //$NON-NLS-1$

		app.setNode(this);
		app.checkValidity();

		// first register the application locally
		applicationRegistry.registerApplication(app);
		getLocalNodeImage().registerApplicationAgents(app.getAllAgentPaths());

		ApplicationRegisteredParams params = new ApplicationRegisteredParams(app);
		Event event = new Event(ApplicationRegisteredParams.EVENT_NAME, params);
		eventRegistry.occured(event);

		return propagateRegisteredApplication(app, nodeNames);
	}

	/**
	 * @return
	 */
	protected AbstractNodeImage getLocalNodeImage()
	{
		return getNodeAnchor().getClusterImage().getLocalNodeImage();
	}
	
	protected Application propagateRegisteredApplication(Application app, List nodeNames) throws RemoteException
	{
		// Tell all nodes in <nodeNames> that the application <app> exists at this node.
		if (app.isPropagated())
			return app;

		MultiAgentProxy agentsProxy = getSystemApplication().getForeignSystemAgents(nodeNames);
		if(agentsProxy == null)
			return null;

				
		try
		{
			RegisterApplicationCommand cmd = new RegisterApplicationCommand(getName(), app.toForeignDescriptor(), getNodeLocation());
			agentsProxy.accept(cmd);
		}
		catch (InvalidCommandException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}

		app.setPropagated(true);

		if (!app.start())
			return null;

		return app;
	}

	/**
	 * @return
	 */
	public NodeLocation getNodeLocation()
	{
		NodeLocation location = new NodeLocation(getName(), getSystemApplication().getLocalRegistryLocation());
		return location;
	}

	protected void propagateDeregisteredApplication(String applicationName, List nodeNames)
	{
		// TODO - not yet tested.
		MultiAgentProxy agentsProxy = null;

		try
		{
			if (nodeNames == null)
				agentsProxy = getSystemApplication().getAllForeignSystemAgents();
			else
				agentsProxy = getSystemApplication().getAllSystemAgentsOnNodes(nodeNames);
		}
		catch (NoSuchAgentException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return;
		}
		catch (IllegalAgentAccessException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return;
		}

		DeregisterApplicationCommand cmd = new DeregisterApplicationCommand(getName(), applicationName);
		try
		{
			agentsProxy.accept(cmd);
		}
		catch (RemoteException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
			// TODO - have to decide what to do if connection to agent is
			// broken. Remove application from foreign node image?
		}
		catch (InvalidCommandException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}

	}

	/**
	 * @return Returns the version.
	 */
	public String getVersion()
	{
		return version;
	}

	/**
	 * @param version
	 *            The version to set.
	 */
	public void setVersion(String version)
	{
		this.version = version;
	}

	/**
	 * @return Returns the eventRegistry.
	 */
	public ILocalEventRegistry getNodeEventRegistry()
	{
		return new EventRegistryEnvelope(this, eventRegistry);
	}

	/**
	 * @param name2
	 */
	public void deregisterApplication(String applicationName)
	{
		getApplicationRegistry().deregisterApplication(applicationName);
		getLocalNodeImage().deregisterApplicationAgents(applicationName);
		propagateDeregisteredApplication(applicationName, null);

		ForeignApplicationDeregisteredParams params = new ForeignApplicationDeregisteredParams(getName(), applicationName);
		Event event = new Event(ForeignApplicationDeregisteredParams.EVENT_NAME, params);
		eventRegistry.occured(event);

		// TODO - all agents belonging to application have to finish processing
		// all commands they have and then terminate
	}

	/**
	 * @param applicationName
	 * @return
	 */
	public boolean isForeignApplicationRegistered(String applicationName, String nodeName)
	{
		return getNodeAnchor().isForeignApplicationRegistered(applicationName, nodeName);
	}

	public AbstractNodeAnchor getNodeAnchor()
	{
		return (AbstractNodeAnchor) getSystemApplication().getObjectSpace().get(NodeAnchor.LABEL);
	}

	/**
	 * @param capabilityName
	 * @param applicationName
	 * @param nodeName
	 * @return
	 */
	public boolean isForeignCapabilityRegistered(CapabilityPath path)
	{
		return getNodeAnchor().isForeignCapabilityRegistered(path);
	}

	/**
	 * @return Returns the objectSpace.
	 */
	public IObjectSpace getNodeObjectSpace()
	{
		return objectSpace;
	}

	/**
	 * @param clusterSpacesByName
	 */
	public ISerializableObjectSpace getClusterObjectSpace(String name) throws NoSuchElementException
	{
		ISerializableObjectSpace clusterOS = (ISerializableObjectSpace) clusterSpacesByName.get(name);

		if(clusterOS != null)
			return clusterOS;
			
		clusterOS = lookupClusterSpace(name);
		if(clusterOS != null)
			return clusterOS;
		
		throw new NoSuchElementException("cluster objectSpace " + StringUtility.print(name) + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	public Collection getClusterEventRegistries()
	{
		return clusterEventRegistriesByName.values();
	}
	
	/**
	 * @param clusterSpacesByName
	 */
	public ILocalClusterEventRegistry getClusterEventRegistry(String name) throws NoSuchElementException, RemoteException
	{
		LocalClusterEventRegistry localClusterRegistry = (LocalClusterEventRegistry) localClusterEventRegistriesByName.get(name);
		if(localClusterRegistry != null)
			return localClusterRegistry;
		
		IClusterEventRegistry clusterRegistry = (IClusterEventRegistry) clusterEventRegistriesByName.get(name);
		if(clusterRegistry == null)
			throw new NoSuchElementException("no cluster event registry found with name " + StringUtility.print(name) + "!"); //$NON-NLS-1$ //$NON-NLS-2$
		
		localClusterRegistry = new LocalClusterEventRegistry(this, clusterRegistry, getLogger());
		
		try
		{
			LocalClusterEventRegistryRemote localRegistryRemote = new LocalClusterEventRegistryRemote(localClusterRegistry);			
			String rmiName = LocalClusterEventRegistry.RMI_PREFIX + getName(); //$NON-NLS-1$
			RegistryLocation registryLocation = getSystemApplication().getLocalRegistryLocation();
			Registry registry = RMIUtility.getRegistry(registryLocation);
			registry.bind(rmiName, localRegistryRemote);
			clusterRegistry.register(rmiName, registryLocation);
		}
		catch (RemoteException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
			throw e;
		}
		catch (AlreadyBoundException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
			AssertUtility.assertTrue(false); // would be logical error
		}
		
		localClusterEventRegistriesByName.put(name, localClusterRegistry);
		
		return localClusterRegistry;
	}
	
	/**
	 * @param name
	 * @return
	 */
	protected ISerializableObjectSpace lookupClusterSpace(String name)
	{
		ISerializableObjectSpace os = null;
		Iterator it = getSystemApplication().getClusterSpaceDescriptors().iterator();
		while (it.hasNext())
		{
			ClusterSpaceDescriptor desc = (ClusterSpaceDescriptor) it.next();
			if(desc.getName().equals(name))
			{
				try
				{
					Registry registry = RMIUtility.getRegistry(desc.getRegistryLocation());
					os = (ISerializableObjectSpace) registry.lookup(ClusterSpace.RMI_PREFIX + desc.getName());
					clusterSpacesByName.put(desc.getName(), os);
				}
				catch (RemoteException e)
				{
					System.out.println(e.getMessage());
					e.printStackTrace();
					return null;
				}
				catch (NotBoundException e)
				{
					System.out.println(e.getMessage());
					e.printStackTrace();
					return null;
				}
			}
		}
		return os;
	}

	/**
	 * @param clusterSpacesByName
	 */
	public ISerializableObjectSpace getClusterObjectSpace() throws NoSuchElementException
	{
		Iterator it = clusterSpacesByName.values().iterator();
		while (it.hasNext())
		{
			return (ISerializableObjectSpace) it.next();
		}
		throw new NoSuchElementException("cluster objectSpace not found!"); //$NON-NLS-1$		
	}
	
	/**
	 * @param clusterSpacesByName
	 */
	public void addClusterSpaces(Map clusterSpacesByName)
	{
		this.clusterSpacesByName.putAll(clusterSpacesByName);
	}

	public void addClusterEventRegistries(Map clusterEventRegistriesByName)
	{
		this.clusterEventRegistriesByName.putAll(clusterEventRegistriesByName);
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpaceProvider#getSpace()
	 */
	public IObjectSpace getObjectSpace()
	{
		return getNodeObjectSpace();
	}
	
	public boolean isLastNodeOnWorkstation() 
	throws RemoteException
	{
		Registry registry = RMIUtility.getRegistry(getSystemApplication().getLocalRegistryLocation());
		String[] list = registry.list();
		int count = 0;
		for (int i = 0; i < list.length; i++)
		{
			if(list[i].indexOf(Node.RMI_LABEL_PREFIX) != -1)
				count++;
		}
		AssertUtility.assertTrue(count >= 1);
		return count == 1;
	}
	
	public NodeDescriptor getNodeDescriptor()
	{
		return (NodeDescriptor) getSystemApplication().getObjectSpace().get(NodeDescriptor.LABEL);
	}
	
	public AbstractClusterImage getClusterImage()
	{
		return getNodeAnchor().getClusterImage();
	}

	/**
	 * 
	 */
	public void registerSystemCapabilities() throws RemoteException
	{
		getLocalNodeImage().registerApplicationAgents(getSystemApplication().getAllAgentPaths());
		getSystemApplication().propagateRegistered(null);
	}

	/**
	 * @return
	 */
	public List getMonitorAgents()
	{
		synchronized(monitorAgentsLock)
		{
			return monitorAgents;
		}
	}
	
	public void addMonitorAgent(IAgentProxy monitorAgent)
	{
		synchronized(monitorAgentsLock)
		{
			if(monitorAgents == null)
				monitorAgents = new LinkedList();
			monitorAgents.add(monitorAgent);
		}
	}
	
	public boolean removeMonitorAgents(IAgentProxy monitorAgent)
	{
		synchronized(monitorAgentsLock)
		{
			boolean found = monitorAgents.remove(monitorAgent);
			if(monitorAgents.isEmpty())
				monitorAgents = null;
			return found;
		}
	}

	/**
	 * @return
	 */
	public IMonitorProducerAnchor getMonitorProducerAnchor()
	{
		return (IMonitorProducerAnchor) getSystemApplication().getObjectSpace().get(NodeAnchor.MONITOR_PRODUCER_ANCHOR_LABEL);
	}

	/**
	 * @param cmd
	 */
	public void sendToMonitorAgents(ICommand cmd)
	{
		synchronized(monitorAgentsLock)
		{
			Iterator it = monitorAgents.iterator();
			while (it.hasNext())
			{
				try
				{
					IAgentProxy monitorAgent = (IAgentProxy) it.next();
					if(monitorAgents.contains(monitorAgent))
						monitorAgent.accept(cmd);
				}
				catch (Exception e)
				{
					System.out.println(e.getMessage());
					e.printStackTrace();
					getLogger().exception(e);
				}				
			}
		}
	}


	/**
	 * @return Returns the simpleEventRegistry.
	 */
	public ISimpleEventRegistry getSimpleEventRegistry()
	{
		return simpleEventRegistry;
	}

	/**
	 * @return
	 */
	public String getNodeLabel()
	{
		return getNodeAnchor().getLabel();
	}
	
	public BillboardPanelController getBillboard()
	{
		return (BillboardPanelController) getNodeObjectSpace().get(BillboardPanelController.LABEL);
	}
	
	public TextAreaPanelController getFreeTextArea()
	{
		return (TextAreaPanelController) getNodeObjectSpace().get(TextAreaPanelController.LABEL);
	}
	
	/**
	 * @return Returns the vetoableEventRegistry.
	 */
	public IVetoableEventRegistry getVetoableEventRegistry()
	{
		return vetoableEventRegistry;
	}
}
