/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
package org.almendra.janet.cas.node;

import java.util.ArrayList;
import java.util.List;

import org.almendra.commons.threads.ISimpleActiveObject;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.schedulers.Arbitrator;
import org.almendra.janet.cas.schedulers.SupremeScheduler;


/**
 * 
 * @author Oliver Plohmann
 */
public class SchedulerAnchor implements ISimpleActiveObject, ILoggable
{
	public static final int SUPREME_PRIORITY = Thread.MAX_PRIORITY - 1 ;
	public static final int ARBITRATOR_PRIORITY = SUPREME_PRIORITY - 2;
	public static final int SYSTEM_PRIORITY = Thread.NORM_PRIORITY + 2;
	public static final int APPLICATION_PRIORITY = Thread.NORM_PRIORITY;
	public static final int REDUCED_APPLICATION_PRIORITY = Thread.MIN_PRIORITY;
	
	public static final String SCHEDULER_NAME = "SCHEDULER"; //$NON-NLS-1$
	public static final String LOGGER_OBJECT_NAME = "SCHEDULER"; //$NON-NLS-1$
	public static final String LOG_EVENT_SCHEDULER_STARTED = "SCHEDULER_STARTED";	//$NON-NLS-1$
	public static final String LOG_EVENT_SCHEDULER_STOPPED = "SCHEDULER_STOPPED"; 	//$NON-NLS-1$

	protected Node node = null;
	protected Logger logger = null;
	protected Arbitrator arbitrator = null;
	protected Arbitrator systemArbitrator = null;
	protected SupremeScheduler supremeScheduler = null;
	
	protected SupremeScheduler responseScheduler = null;
	
	protected SchedulerAnchor()
	{
		super();
		init();
	}
	
	public SchedulerAnchor(Node node, Logger logger)
	{
		super();
		this.node = node;
		Logger childLogger = Logger.getChildLogger(logger); 
		childLogger.getDefaultValues().setObjectName(SCHEDULER_NAME);
		this.logger = childLogger;
		arbitrator = new Arbitrator(this, childLogger,  SchedulerAnchor.ARBITRATOR_PRIORITY);
		systemArbitrator = new Arbitrator(this, childLogger,  SchedulerAnchor.ARBITRATOR_PRIORITY + 1);
		supremeScheduler = new SupremeScheduler(this, childLogger);
		responseScheduler = new SupremeScheduler(this, childLogger);
		init();
	}

	/**
	 * @return
	 */
	public Node getNode()
	{
		return node;
	}

	/**
	 * @param node
	 */
	public void setNode(Node node)
	{
		this.node = node;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.IScheduler#start()
	 */
	public void start()
	{
		supremeScheduler.start();
		responseScheduler.start();
		systemArbitrator.start(node.getNodeDescriptor().isEnableSystemArbitrator());
		arbitrator.start(node.getNodeDescriptor().isEnableApplicationArbitrator());
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.IScheduler#stop()
	 */
	public void stop()
	{
		supremeScheduler.stop();
		responseScheduler.stop();
		arbitrator.stop();
		systemArbitrator.stop();
	}

	/**
	 * @return
	 */
	public Arbitrator getArbitrator()
	{
		return arbitrator;
	}

	/**
	 * @return
	 */
	public synchronized boolean isIdle()
	{
		if(!supremeScheduler.isIdle())
			return false;
		if(!responseScheduler.isIdle())
			return false;
		if(!arbitrator.isIdle())
			return false;
		return arbitrator.isIdle();
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		List events = new ArrayList();
		events.add(LOG_EVENT_SCHEDULER_STARTED);
		events.add(LOG_EVENT_SCHEDULER_STOPPED);
		return events;
	}
	
	protected void init()
	{
		logger.registerForEvents(registeredLogEvents());
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.ISimpleActiveObject#isActive()
	 */
	public boolean isActive()
	{
		return !isIdle();
	}

	/**
	 * @return Returns the responseScheduler.
	 */
	protected SupremeScheduler getResponseScheduler()
	{
		return responseScheduler;
	}

	/**
	 * @return
	 */
	public Arbitrator getSystemArbitrator()
	{
		return systemArbitrator;
	}

}
