/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node.interpreters;

import java.rmi.RemoteException;

import org.almendra.commons.util.collection.StringVector;
import org.almendra.commons.util.exception.InsufficientPrivilegeException;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.commands.RegisterNodeCommand;
import org.almendra.janet.cas.commands.RegisteredNodeCommand;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.InvalidCommandException;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeLocation;
import org.almendra.janet.cas.scheduling.CommandAccessor;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.IInterpreter;


/**
 * After the node at startup has registered with the central it receives a list of nodes 
 * from the central the node needs to connect to. In general, the other nodes are all other 
 * nodes in the cluster.
 * 
 * @author Oliver
 *  
 */
public class RegisteredNodeInterpreter extends NodeInterpreter implements IInterpreter
{
	protected AbstractRegisterNodeTrait registerNodeTrait = createRegisterNodeTrait();
	protected RegisterApplicationTrait registerApplicationTrait = new RegisterApplicationTrait();
	
	public RegisteredNodeInterpreter()
	{
		super();
	}

	protected AbstractRegisterNodeTrait createRegisterNodeTrait()
	{
		return new RegisterNodeTrait();
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.commands.RegisteredNodeInterpreter#registerForeignNode(java.lang.String)
	 */
	protected void registerForeignNode(NodeLocation nodeLocation, CommandAccessor cmdAccessor)
	{
		registerNodeTrait.registerForeignNode(nodeLocation, cmdAccessor);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.IInterpreter#execute(org.almendra.commons.cas.node.ICommand,
	 *      org.almendra.commons.cas.node.IObjectSpace)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if (cmdAccessor.getCommand() instanceof RegisteredNodeCommand)
		{
			RegisteredNodeCommand command = (RegisteredNodeCommand) cmdAccessor.getCommand();
			NodeLocation[] nodeLocations = command.getNodeLocations();
			
			for (int i = 0; i < nodeLocations.length; i++)
			{
				try
				{					
					registerForeignNode(nodeLocations[i], cmdAccessor);
					foreignNodeRegisterMyself(nodeLocations[i].getNodeName(), cmdAccessor);
					foreignNodeRegisterMyApplications(nodeLocations[i].getNodeName(), cmdAccessor);
				}
				catch (Exception e)
				{
					System.out.println(e.getMessage());
					e.printStackTrace();
					cmdAccessor.log(e);
					// TODO - handle error propperly, e.g. foreign node reference must be removed again, etc.
				}
			}			
		}
	}

	/**
	 * @param string
	 * @param cmdAccessor
	 */
	protected void foreignNodeRegisterMyApplications(String foreignNodeName, CommandAccessor cmdAccessor) 
	throws RemoteException, NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException, InvalidCommandException
	{
		NodeLocation nodeLocation = cmdAccessor.getNodeIfPrivilege().getNodeLocation();
		registerApplicationTrait.foreignNodeRegisterMyApplications(foreignNodeName, nodeLocation, cmdAccessor);
	}

	/**
	 * @param string
	 * @param cmdAccessor
	 */
	protected void foreignNodeRegisterMyself(String foreignNodeName, CommandAccessor cmdAccessor) 
	throws RemoteException, InvalidCommandException
	{
		Node node = cmdAccessor.getNodeIfPrivilege();
		IAgentProxy agentProxy = null;
		
		try
		{
			agentProxy = cmdAccessor.getSystemAgent(foreignNodeName);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.log(e);
			return;
		}
		  
		ICommand command = new RegisterNodeCommand(node.getNodeLocation());
		agentProxy.accept(command);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public StringVector commandNames()
	{
		return new StringVector(RegisteredNodeCommand.QualifiedName);
	}

}
