/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.schedulers;

import java.util.ArrayList;
import java.util.List;

import org.almendra.commons.threads.ActiveSchedulingObject;
import org.almendra.commons.util.collection.IQueue;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;


/**
 * @author Oliver
 *
 */
public abstract class AbstractScheduler extends ActiveSchedulingObject
{

	public static final String LOG_EVENT_RESUMING_SCHEDULER = "RESUMING_SCHEDULER"; //$NON-NLS-1$
	public static final String LOG_EVENT_SUSPENDING_SCHEDULER = "SUSPENDING_SCHEDULER"; //$NON-NLS-1$
	public static final String LOG_EVENT_STARTED_SCHEDULER = "STARTED_SCHEDULER"; //$NON-NLS-1$
	public static final String LOG_EVENT_STOPPED_SCHEDULER = "STOPPED_SCHEDULER"; //$NON-NLS-1$
	
	protected Logger logger = null;
	
	protected AbstractScheduler()
	{
		super();
	}
	
	/**
	 * Creates a new <code>AbstractScheduler</code> object.
	 */
	public AbstractScheduler(IQueue queue, Logger logger)
	{
		super(queue);
		setLogger(logger);
	}

	/**
	 * Creates a new <code>AbstractScheduler</code> object.
	 */
	public AbstractScheduler(IQueue queue, boolean verbose, Logger logger)
	{
		super(queue, verbose);
		setLogger(logger);
	}

	/**
	 * Creates a new <code>AbstractScheduler</code> object.
	 */
	public AbstractScheduler(IQueue queue, boolean verbose, int priority, Logger logger)
	{
		super(queue, verbose, priority);
		setLogger(logger);
	}

	/**
	 * Creates a new <code>AbstractScheduler</code> object.
	 */
	public AbstractScheduler(Logger logger)
	{
		super();
		setLogger(logger);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveObject#isDaemon()
	 */
	protected boolean isDaemon()
	{
		return false;
	}	

	/**
	 * @return
	 */
	public Logger getLogger()
	{
		return logger;
	}

	/**
	 * @param logger
	 */
	protected void setLogger(Logger logger)
	{
		Logger childLogger = Logger.getChildLogger(logger);
		childLogger.getDefaultValues().setSchedulerName(getName());
		childLogger.registerForEvents(registeredLogEvents());
		this.logger = childLogger;
	}	

	public List registeredLogEvents()
	{
		List events = new ArrayList();
		events.add(LOG_EVENT_STARTED_SCHEDULER);
		events.add(LOG_EVENT_STOPPED_SCHEDULER);
		events.add(LOG_EVENT_RESUMING_SCHEDULER);
		events.add(LOG_EVENT_SUSPENDING_SCHEDULER);
		return events;
	}

	/**
	 * @param command
	 * @param interpreter
	 */
	protected void log(String eventName)
	{
		if (logger.isSubscribed(eventName))
		{
			logger.info(getDefaultLogEntry(eventName));
		}
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.schedulers.AbstractScheduler#getDefaultLogEntry(java.lang.String)
	 */
	protected LogEntry getDefaultLogEntry(String eventName)
	{
		LogEntry entry = (LogEntry) logger.getNewLogEntry();
		entry.setEvent(eventName);
		return entry;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveObject#messageStarting()
	 */
	protected void messageStarting()
	{
		log(LOG_EVENT_STARTED_SCHEDULER);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveObject#messageStopping()
	 */
	protected void messageStopping()
	{
		log(LOG_EVENT_STOPPED_SCHEDULER);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#messageResuming()
	 */
	protected void messageResuming()
	{
		logger.info(getDefaultLogEntry(LOG_EVENT_RESUMING_SCHEDULER));
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#messageSuspending()
	 */
	protected void messageSuspending()
	{
		logger.info(getDefaultLogEntry(LOG_EVENT_SUSPENDING_SCHEDULER));
	}

}
