/*
 * Created on 18.09.2003
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.schedulers;

/**
 * The Arbitrator makes sure that an application schedulerAnchor cannot cause
 * starvation of other application schedulers. Therefore, each application
 * schedulerAnchor receives a certain time slice of the CPU time only. There is no
 * time slicing if there is only one application schedulerAnchor present.
 * AbstractApplication schedulers that are not started, were stopped, or have
 * no commands to execute are not time sliced, either. All application
 * schedulers that are not time sliced are stored in <nonTimeSlicedSchedulers>.
 */

import java.util.LinkedList;
import java.util.List;

import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.threads.LockUtility;
import org.almendra.commons.threads.ThreadUtility;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.SchedulerAnchor;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;


/**
 * @author Oliver Plohmann
 */
public class Arbitrator extends AbstractScheduler implements ILoggable
{
	public static final String ARBITRATOR_NAME = "ARBITRATOR"; //$NON-NLS-1$
	public static final String LOG_EVENT_CHANGING_ACTIVE_SCHEDULER = "CHANGING_ACTIVE_SCHEDULER"; //$NON-NLS-1$
	public static final String LOGGER_OBJECT_NAME = "ARBITRATOR"; //$NON-NLS-1$

	protected ArbitratedScheduler currentScheduler = null;
	protected volatile int timeSliceInMillis = 500;
	protected Node node = null;
	protected SchedulerAnchor scheduler = null;
	protected List sleepingSchedulers = new LinkedList();
	protected List waitingSchedulers = new LinkedList();
	protected ReentrantLock producerConsumerLock = new ReentrantLock();
	protected volatile boolean isEnabled = true;

	protected Arbitrator()
	{
		super();
		init();
	}

	public Arbitrator(SchedulerAnchor scheduler, Logger logger, int priority)
	{
		super(new CommandSchedulerQueue(), true, priority, logger);
		this.name = ARBITRATOR_NAME;
		this.node = scheduler.getNode();
		this.scheduler = scheduler;
		init();
	}

	protected IExecutable getNext()
	{

		ApplicationScheduler appScheduler = (ApplicationScheduler) super.getNext();

		LockUtility.acquire(producerConsumerLock);

		if (appScheduler.isIdle())
		{
			setSleeping(appScheduler);
			producerConsumerLock.release();
			return getNext();
		}

		if (!hasNext())
		{
			setWaiting(appScheduler);
			producerConsumerLock.release();
			return getNext();
		}

		add(appScheduler);
		producerConsumerLock.release();
		
		return appScheduler;

	}

	protected void setSleeping(ArbitratedScheduler arbitratedScheduler)
	{
		arbitratedScheduler.setSleeping();
		sleepingSchedulers.add(arbitratedScheduler);
	}

	protected void setWaiting(ApplicationScheduler appScheduler)
	{
		appScheduler.setWaiting();
		waitingSchedulers.add(appScheduler);
	}

	/**
	 * @param i
	 */
	public void setTimeSliceInMillis(int i)
	{
		timeSliceInMillis = i;
	}

	protected void commandAddedTo(ArbitratedScheduler arbitratedScheduler)
	{
		if (arbitratedScheduler.isRunning())
			return;

		if(!isEnabled)
		{
			if(!arbitratedScheduler.isRunning())
				start(arbitratedScheduler);
			return;
		}
		
		LockUtility.acquire(producerConsumerLock);

		sleepingSchedulers.remove(arbitratedScheduler);

		if (available() > 1)
		{
			setScheduled(arbitratedScheduler);
			producerConsumerLock.release();
			return;
		}

		if (waitingSchedulers.size() > 0)
		{
			setScheduled(arbitratedScheduler);
			setWaitingSchedulerScheduled(waitingSchedulers);
			producerConsumerLock.release();
			return;
		}

		if (available() == 0)
		{
			arbitratedScheduler.setWaiting();
			waitingSchedulers.add(arbitratedScheduler);
			start(arbitratedScheduler);
			producerConsumerLock.release();
			return;
		}

		producerConsumerLock.release();
	}

	protected void setScheduled(ArbitratedScheduler arbitratedScheduler)
	{
		arbitratedScheduler.setScheduled();
		add(arbitratedScheduler);
	}

	protected void setWaitingSchedulerScheduled(List arbitratedSchedulers)
	{
		while (!arbitratedSchedulers.isEmpty())
		{
			setScheduled((ApplicationScheduler) arbitratedSchedulers.remove(0));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.scheduling.ActiveSchedulingObject#execute(java.lang.Object)
	 */
	protected void execute(IExecutable executable)
	{
		if (executable instanceof ArbitratedScheduler)
		{
			if (currentScheduler != null)
				currentScheduler.setPriority(SchedulerAnchor.REDUCED_APPLICATION_PRIORITY);
			ArbitratedScheduler scheduler = (ArbitratedScheduler) executable;
			if (scheduler != null)
			{
				start(scheduler);
				sleep(scheduler);
			}
			currentScheduler = scheduler;
		}
	}

	protected void start(ArbitratedScheduler arbitratedScheduler)
	{
		int priority = arbitratedScheduler.getPriority();
		priority = Math.min(priority, arbitratedScheduler.getMaxPriority());		
		arbitratedScheduler.setPriority(priority);
		logChangedActiveScheduler(currentScheduler, arbitratedScheduler);
		arbitratedScheduler.start();
	}

	protected void sleep(ArbitratedScheduler arbitratedScheduler)
	{
		int share = arbitratedScheduler.getNumberOfSchedulersInSameApplication();
		AssertUtility.assertTrue(share > 0);
		ThreadUtility.sleep(timeSliceInMillis / share);
	}

	/**
	 * @param schedulerAnchor
	 */
	protected void logChangedActiveScheduler(CommandScheduler previouslyActiveScheduler, CommandScheduler currentlyActiveScheduler)
	{
		if (logger.isSubscribed(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER);
			String argument = "prevActive:" + previouslyActiveScheduler + " | " + "currentActive:" + currentlyActiveScheduler; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			entry.setArgument(argument);
			logger.info(entry);
		}
	}

	public void stop()
	{	
		super.stop();
		while (queue.hasNext())
		{
			ArbitratedScheduler scheduler = (ArbitratedScheduler) queue.getNext();
			scheduler.stop();
		}		
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.scheduling.ActiveObject#isDaemon()
	 */
	public boolean isDaemon()
	{
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		List events = super.registeredLogEvents();
		events.add(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER);
		return events;
	}

	protected void init()
	{
		if (logger != null)
		{
			logger.registerForEvents(registeredLogEvents());
		}
	}

	/**
	 * @param appScheduler
	 */
	public void created(ArbitratedScheduler arbitratedScheduler)
	{
		AssertUtility.assertTrue(arbitratedScheduler.available() == 0);
		arbitratedScheduler.setNew();
		sleepingSchedulers.add(arbitratedScheduler);
	}

	/**
	 * @param appScheduler
	 */
	public void removeScheduler(ApplicationScheduler appScheduler)
	{
		if (appScheduler.isActive())
		{
			remove(appScheduler);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#isIdle()
	 */
	public boolean isIdle()
	{
		return (waitingSchedulers.isEmpty() && super.isIdle());
	}

	/**
	 * @param appScheduler
	 */
	public void isIdle(ArbitratedScheduler arbitratedScheduler)
	{
		LockUtility.acquire(producerConsumerLock);
		if (arbitratedScheduler.isWaiting())
		{
			waitingSchedulers.remove(arbitratedScheduler);
			setSleeping(arbitratedScheduler);
		}
		producerConsumerLock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.schedulers.AbstractScheduler#setLogger(org.almendra.commons.janet.cas.logger.Logger)
	 */
	protected void setLogger(Logger logger)
	{
		super.setLogger(logger);
		this.logger.getDefaultValues().setSchedulerName(LOGGER_OBJECT_NAME);
	}

	/**
	 * @param b
	 */
	public void start(boolean isEnabled)
	{
		this.isEnabled = isEnabled;
		if(isEnabled)
			start();
	}

}
