/*
 * Created on 18.09.2003
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.schedulers;

import java.util.List;

import org.almendra.commons.INamedObject;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.threads.ISimpleActiveObject;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.scheduling.CommandEnvelope;
import org.almendra.janet.cas.scheduling.ICommand;
import org.almendra.janet.cas.scheduling.IInterpreter;



public abstract class CommandScheduler extends AbstractScheduler implements Runnable, ISimpleActiveObject, INamedObject, ILoggable
{
	public static final String LOG_EVENT_PRE_EXEC_CMD = "PRE_EXEC_CMD"; //$NON-NLS-1$
	public static final String LOG_EVENT_POST_EXEC_CMD = "POST_EXEC_CMD"; //$NON-NLS-1$
	public static final String LOG_EVENT_NO_SUITABLE_INTERPRETER_FOUND = "NO_SUITABLE_INTERPRETER_FOUND"; //$NON-NLS-1$
	protected Node node = null;
	
	public CommandScheduler()
	{
		super();
		init();
	}

	public CommandScheduler(Logger logger)
	{
		super(logger);
		init();
	}

	public CommandScheduler(int priority)
	{
		this.priority = priority;
		init();
	}

	public CommandScheduler(Logger logger, int priority)
	{
		this.priority = priority;
		setLogger(logger);
		init();
	}

	public CommandScheduler(int priority, boolean verbose)
	{
		setVerbose(verbose);
		this.priority = priority;
		init();
	}

	public CommandScheduler(Logger logger, int priority, boolean verbose)
	{
		setVerbose(verbose);
		this.priority = priority;
		setLogger(logger);
		init();
	}

	protected void execute(IExecutable executable)
	{
		try
		{
			if (executable instanceof CommandEnvelope)
			{
				CommandEnvelope commandEnvelope = (CommandEnvelope) executable;
				IInterpreter interpreter = getInterpreter(commandEnvelope);
				if (interpreter == null)
				{
					logInterpreterNotFound(commandEnvelope);
				}
				else
				{
					execute(commandEnvelope, interpreter);
				}
			}			
		}
		catch (Exception ex)
		{
			// TODO - make entry in logger
			System.out.println("exception occured executing " + executable + " !"); //$NON-NLS-1$ //$NON-NLS-2$
			System.out.println(ex.getMessage());
			logger.exception(ex);
			ex.printStackTrace();
		}
	}

	protected void logInterpreterNotFound(CommandEnvelope commandEnv)
	{
		LogEntry entry = (LogEntry) logger.getNewLogEntry();
		entry.setEvent(LOG_EVENT_NO_SUITABLE_INTERPRETER_FOUND);
		entry.setArgument(commandEnv.toPrintString());
		logger.error(entry);
		System.out.println("no interpreter found for command " + commandEnv.getCommand() + " on node " + node.getName()); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @param command
	 * @param interpreter
	 */
	protected void logExecuting(CommandEnvelope env, String eventName)
	{
		if (logger.isSubscribed(eventName))
		{
			LogEntry entry = (LogEntry) logger.getNewLogEntry();
			entry.setEvent(eventName);
			entry.setArgument(env.toPrintString());
			entry.setAgentName(env.getCurrentAgent().getName());
			entry.setApplicationName(env.getCurrentAgent().getApplication().getName());
			entry.setCapabilityName(env.getCurrentAgent().getCapability().getName());
			logger.info(entry);
		}
	}

	/**
	 * @return
	 */
	protected abstract String getLogName();

	protected void execute(CommandEnvelope env, IInterpreter interpreter)
	{		
		monitorStartedCommand(env);
		logExecuting(env, LOG_EVENT_PRE_EXEC_CMD);
		env.execute(interpreter);
		logExecuting(env, LOG_EVENT_POST_EXEC_CMD);
		monitorFinishedCommand(env);
	}

	/**
	 * @param isEventHandlerCommand
	 */
	protected void monitorRemovedWaitingCommand(CommandEnvelope env)
	{
		List monitorAgents = node.getMonitorAgents();
		if(monitorAgents == null)
			return;
		
		if(!env.isEventHandlerCommand())
		{
			ICommand cmd = node.getMonitorProducerAnchor().getRemovedWaitingCommandNotification(node.getName(), env);
			node.sendToMonitorAgents(cmd);
		}
	}
	
	/**
	 * @param isEventHandlerCommand
	 */
	protected void monitorFinishedCommand(CommandEnvelope env)
	{
		List monitorAgents = node.getMonitorAgents();
		if(monitorAgents == null)
			return;
		
		if(!env.isEventHandlerCommand())
		{
			ICommand cmd = node.getMonitorProducerAnchor().getFinishedCommandNotification(node.getName(), env);
			node.sendToMonitorAgents(cmd);
		}
	}

	/**
	 * @param isEventHandlerCommand
	 */
	protected void monitorStartedCommand(CommandEnvelope env)
	{
		List monitorAgents = node.getMonitorAgents();
		if(monitorAgents == null)
			return;
		
		if(!env.isEventHandlerCommand())
		{
			ICommand cmd = node.getMonitorProducerAnchor().getStartedCommandNotification(node.getName(), env);
			node.sendToMonitorAgents(cmd);
		}
	}

	public void accept(CommandEnvelope env)
	{
		add(env);
		env.setCurrentScheduler(this);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.scheduling.ActiveObject#isDaemon()
	 */
	public boolean isDaemon()
	{
		return false;
	}

	protected void init()
	{
		queue = new CommandQueue();
		setVerbose(true);
	}

	/**
	 * @param command
	 * @return
	 */
	public abstract IInterpreter getInterpreter(CommandEnvelope commandEnv);

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List registeredLogEvents()
	{
		List events = super.registeredLogEvents();
		events.add(LOG_EVENT_PRE_EXEC_CMD);
		events.add(LOG_EVENT_POST_EXEC_CMD);
		events.add(LOG_EVENT_NO_SUITABLE_INTERPRETER_FOUND);
		return events;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#messageResuming()
	 */
	protected void messageResuming()
	{
		log(LOG_EVENT_RESUMING_SCHEDULER);
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#messageSuspending()
	 */
	protected void messageSuspending()
	{
		log(LOG_EVENT_SUSPENDING_SCHEDULER);
	}

	
	protected void setLogger(Logger logger)
	{
		super.setLogger(logger);
		logger.getDefaultValues().setSchedulerName(getName());
		logger.getDefaultValues().setAgentName(getCurrentAgentName());
		logger.getDefaultValues().setApplicationName(getApplicationName());
		logger.getDefaultValues().setCapabilityName(getCapabilityName());
	}
	
	/**
	 * @return
	 */
	protected abstract String getApplicationName();

	/**
	 * @return
	 */
	protected abstract String getCapabilityName();

	/**
	 * @return
	 */
	protected abstract String getCurrentAgentName();
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.IExecutable#execute(java.lang.Object)
	 */
	public void execute(Object object)
	{
		if(object instanceof IExecutable)
		{
			IExecutable executable = (IExecutable) object;
			execute(executable);
		}		
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#executableRemoved(org.almendra.commons.commons.threads.IExecutable)
	 */
	protected void executableRemoved(IExecutable executable)
	{
	}
}