/*
 * Created on 18.09.2003
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.schedulers;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.collection.IQueue;
import org.almendra.janet.cas.scheduling.*;

import EDU.oswego.cs.dl.util.concurrent.ReentrantWriterPreferenceReadWriteLock;

import junit.framework.Assert;


public class PriorityQueue implements IQueue
{
	protected LinkedList[] queues = null;
	protected int size = 0;
	protected ReentrantWriterPreferenceReadWriteLock lock = new ReentrantWriterPreferenceReadWriteLock();

	public PriorityQueue()
	{
		super();
		init(0);
	}

	public PriorityQueue(int numQueues)
	{
		init(numQueues);
	}

	public boolean add(Object object)
	{
		if (object instanceof IPrioritizedObject)
		{
			IPrioritizedObject prioritizedObject = (IPrioritizedObject) object; 
			return add(prioritizedObject);
		}
		return false;
	}

	protected boolean add(IPrioritizedObject prioritizedObject)
	{
		int priority = prioritizedObject.getPriority();
		if (priority < 0 || priority > queues.length)
		{
			System.err.println("invalid parameter in " + getClass().getName() + ".add(...)\n" + //$NON-NLS-1$ //$NON-NLS-2$
			"- value of priority out of range: " + priority); //$NON-NLS-1$		
			return false;
		}	
		
		try
		{
			lock.writeLock().acquire();
			LinkedList queue = queues[priority];
			queue.addLast(prioritizedObject);
			size++;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
		return false;
	}

	public Object getNext()
	{
		try
		{
			lock.writeLock().acquire();
			for (int i = 0; i < queues.length; i++)
			{
				LinkedList queue = queues[i];
				if (!queue.isEmpty())
				{
					size--;
					return queue.removeFirst();
				}
			}
			return null;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
		return null;
	}

	protected void init(int numQueues)
	{
		queues = new LinkedList[numQueues];
		for (int i = 0; i < queues.length; i++)
		{
			queues[i] = new LinkedList();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.scheduling.SharedQueue#isEmpty()
	 */
	public boolean isEmpty()
	{
		try
		{
			lock.readLock().acquire();
			return size == 0;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.scheduling.SharedQueue#size()
	 */
	public int size()
	{
		try
		{
			lock.readLock().acquire();
			return size;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.cas.node.scheduling.SharedQueue#removePrim(java.lang.Object)
	 */
	/**
	 * Remove object from aPriorityQueue. This method is inefficient and should
	 * be avoided. Use PriorityQueue.getNext() instead.
	 */
	public Object remove(Object object)
	{
		try
		{
			lock.writeLock().acquire();
			if (object instanceof IPrioritizedObject)
			{
				IPrioritizedObject prioritizedObject = (IPrioritizedObject) object;
				LinkedList queue = queues[prioritizedObject.getPriority()];
				boolean found = queue.remove(prioritizedObject);
				if(!found)
					return null;
				return object;
			}
			for (int i = 0; i < queues.length; i++)
			{
				LinkedList queue = queues[i];
				if (queue.contains(object))
					queue.remove(object);
				return object;
			}
			AssertUtility.assertTrue("object " + object + " not in list", false); //$NON-NLS-1$ //$NON-NLS-2$
			return null;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.writeLock().release();
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.collection.IQueue#hasNext()
	 */
	public boolean hasNext()
	{
		try
		{
			lock.readLock().acquire();
			return size > 0;
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return false;	
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.collection.IQueue#iterator()
	 */
	public Iterator iterator()
	{
		try
		{
			lock.readLock().acquire();
			LinkedList list = new LinkedList();
			for (int i = 0; i < queues.length; i++)
			{
				List commands = queues[i];
				list.addAll(commands);
			}
			return list.iterator();
		}
		catch (InterruptedException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			lock.readLock().release();
		}
		return null;
		
	}

}
