/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.oliverplohmann.de/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.scheduling;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.NoSuchElementException;

import org.almendra.commons.Result;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.exception.InsufficientPrivilegeException;
import org.almendra.commons.util.exception.InvalidMessageException;
import org.almendra.commons.util.globalId.UniqueId;
import org.almendra.commons.util.globalId.UniqueIdGenerator;
import org.almendra.janet.cas.agents.AgentProxy;
import org.almendra.janet.cas.agents.IAgent;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.commands.ExecuteEventHandlerCommand;
import org.almendra.janet.cas.events.ILocalClusterEventRegistry;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.IAgentDispatcher;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LocalCapabilityPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeAccessor;
import org.almendra.janet.cas.node.NodeName;
import org.almendra.janet.cas.schedulers.CommandScheduler;
import org.almendra.janet.cas.spaces.IObjectSpace;


public class CommandEnvelope implements Serializable, Cloneable, IPrioritizedObject, IExecutable, ICommandHolder
{
	public static final int COMMAND_TYPE_BALANCING = 1;
	public static final int COMMAND_TYPE_SHARING = 2;
	
	protected ICommand command = null;
	protected transient Logger logger = null;
	protected CommandHistory history = new CommandHistory();
	protected UniqueId id = UniqueIdGenerator.next();
	protected AgentPath destinationAgent = null;
	protected transient IAgent currentAgent = null;	
	protected transient Node node = null;
	protected transient IInterpreter predefinedInterpreter = null; 
	protected boolean isForeign = false;
	protected transient boolean isIgnore = false;
	protected volatile transient boolean suspended = false;
	protected Serializable suspendedLock = ""; //$NON-NLS-1$ 
	protected transient boolean executionStarted = false;
	protected transient boolean executionFinished = false;
	protected transient CommandScheduler currentScheduler = null;
	protected transient String qualifiedCapabilityName = null;
	protected int kind = -1;

	/**
	 * Creates a new <code>CommandEnvelope</code> object.
	 *  
	 */
	protected CommandEnvelope()
	{
		super();
	}
	
	/**
	 * Creates a new <code>CommandEnvelope</code> object.
	 *  
	 */
	protected CommandEnvelope(CommandEnvelope env)
	{
		super();
		command = env.command;
		currentAgent = env.currentAgent;
		destinationAgent = env.destinationAgent;
		history = env.history;
		logger = env.logger;
		node = env.node;
		id = UniqueIdGenerator.next();
	}
	
	public CommandEnvelope(AgentPath agentPath, ICommand command, Logger logger, String nodeName)
	{
		super();
		destinationAgent = new AgentPath(agentPath);
		this.command = command;
		this.logger = logger;
		getNewHistoryItem(nodeName);
		setHistoryItem(agentPath, nodeName, command);
	}
	
	public void setNewDestination(AgentPath agentPath, String nodeName)
	{
		destinationAgent = new AgentPath(agentPath);
		getNewHistoryItem(nodeName);
		setHistoryItem(agentPath, nodeName, command);
	}
	
	/**
	 * @param item
	 */
	protected void addHistoryItem(CommandHistoryItem item, String nodeName)
	{
		item.setCreationNodeName(nodeName);
		history.addItem(item);
	}
	/**
	 * @return
	 */
	public ICommand getCommand()
	{
		return command;
	}
	
	/**
	 * Return the agent that received this <code>CommandEnvelope</code>. This method must
	 * not be public as the user could get unrestricted access to the node object. Such
	 * access should only be granted to system commands that are executed by the system 
	 * agent. Such access has to be controlled by throwing an <code>InsufficientPrivelegeException</code>
	 * if the owning agent is not the system agent. 
	 * @return Returns the currentAgent.
	 */
	public IAgent getCurrentAgent()
	{
		return currentAgent;
	}
	
	public IAgentProxy getCurrentAgentProxy() 
	{
		return getCurrentAgent().getAgentProxy();
	}
	
	/**
	 * @return
	 */
	public IObjectSpace getApplicationObjectSpace() 
	{
		return currentAgent.getObjectSpace();
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.globalId.IUniqueGlobalIdObject#getGlobalId()
	 */
	public UniqueId getGlobalId()
	{
		return id;
	}
	
	/**
	 * @return Returns the history.
	 */
	public CommandHistory getHistory()
	{
		return history;
	}

	/**
	 * @return
	 */
	public CommandHistoryItem getResponseHistoryItem(String nodeName)
	{
		CommandHistoryItem item = CommandHistoryItem.requestItem();
		addHistoryItem(item, nodeName);
		return item;
	}
	
	/**
	 * @return
	 */
	public CommandHistoryItem getNewHistoryItem(String nodeName)
	{
		CommandHistoryItem item = CommandHistoryItem.requestItem();
		addHistoryItem(item, nodeName);
		return item;
	}
	
	/**
	 *  
	 */
	public void setArrivalTime()
	{
		history.setArrivalTime();
	}
	
	/**
	 *  
	 */
	public void setCreationTime()
	{
		history.setCreationTime();
	}
	
	/**
	 * @param currentAgent
	 *            The currentAgent to set.
	 */
	public void setCurrentAgent(IAgent currentAgent)
	{
		this.currentAgent = currentAgent;
	}
	
	/**
	 *  
	 */
	public void setEndExecutionTime()
	{
		history.setEndExecutionTime();
	}
	
	/**
	 *  
	 */
	public void setSendTime()
	{
		history.setSendTime();
	}
	
	/**
	 *  
	 */
	public void setStartExecutionTime()
	{
		history.setStartExecutionTime();
	}
	
	public String toLogString()
	{
		return command.toLogString();
	}
	
	public String toPrintString()
	{
		return getCommand().toLogString() + getSuffix() + "{id:" + getGlobalId().getValue() + "}"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @return
	 */
	protected String getSuffix()
	{
		return ""; //$NON-NLS-1$
	}

	/**
	 * @param string
	 * @return
	 */
	public IAgentDispatcher getAgentDispatcher(String qualifiedAgentPath)
	throws NoSuchAgentException, IllegalAgentAccessException
	{
		AgentPath agentPath = new AgentPath(qualifiedAgentPath);
		return getCurrentAgent().getApplication().getAgentDispatcher(agentPath);
	}
	
	/**
	 * @return
	 */
	public CommandAccessor getAccessor()
	{
		return new CommandAccessor(this);
	}

	/**
	 * @return
	 */
	public String getOriginDestinationAgentName()
	{
		CommandHistoryItem item = getLastHistoryItem();
		if(item == null)
			return "?"; //$NON-NLS-1$
		if(item.getLocalAgentPath() != null)
			return item.getLocalAgentPath().getQualifiedPathName();
		if(item.getAgentPath() != null)
			return item.getAgentPath().getQualifiedPathName();
		return "?"; //$NON-NLS-1$
	}

	/**
	 * @return
	 */
	protected CommandHistoryItem getLastHistoryItem()
	{
		if(getHistory() == null)
			return null;
		return getHistory().getLast();
	}
	
	public String getNodeName()
	{
		return node.getName();
	}
	
	public Node getNode() throws InsufficientPrivilegeException
	{
		return currentAgent.getNode();
	}
	
	public NodeAccessor getNodeAccessor()
	{
		return new NodeAccessor(node);
	}	

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.node.AbstractCommandEnvelope#getApplicationName()
	 */
	public ILocalEventRegistry getApplicationEventRegistry()
	{
		return getCurrentAgent().getApplication().getEventRegistry();
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.node.AbstractCommandEnvelope#getApplicationName()
	 */
	public String getApplicationName()
	{
		return destinationAgent.getApplicationName();
	}

	/**
	 * @return Returns the name.
	 */
	public String getCapabilityName()
	{
		return destinationAgent.getCapabilityName();
	}

	/**
	 * @return Returns the destinationAgentName.
	 */
	public String getDestinationAgentName()
	{
		return destinationAgent.getAgentName();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.node.CommandEnvelope#getLocalizedDestinationAgentName()
	 */
	public String getLocalizedDestinationAgentName()
	{
		return destinationAgent.getLocalAgentName();
	}

	/**
	 * @return Returns the destinationAgentName.
	 */
	public String getQualifiedDestinationAgentName()
	{
		return destinationAgent.getQualifiedPathName();
	}

	/**
	 * @param interpreter
	 */
	public void execute(IInterpreter interpreter)
	{		
		setStartExecutionTime();
		executionStarted = true;
		interpreter.execute(getAccessor());
		executionFinished = true;
		setEndExecutionTime();
	}

	public void execute(Object object)
	{
		if(object instanceof IInterpreter)
		{
			IInterpreter interpreter = (IInterpreter) object;
			execute(interpreter);
		}		
	}
	
	/**
	 * @param replyObject
	 */
	public void setResponse(Serializable responseObject)
	throws InvalidMessageException
	{
		throw new InvalidMessageException("message only appropriate when sent to a CommandReplyEnvelope object!"); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.schedulers.IPrioritizedObject#getPriority()
	 */
	public int getPriority()
	{
		int priority = command.getPriority();
		if(getCurrentAgent().isApplicationAgent())
			return Math.max(priority, ICommand.ApplicationMaxPriority);
		return priority;
	}

	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}

	/**
	 * @param logger The logger to set.
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
		if(logger == null)
			return;
		LogEntry entry = logger.getDefaultValues();
		entry.setAgentName(getCurrentAgent().getName());
		entry.setApplicationName(getCurrentAgent().getApplication().getName());
		entry.setCapabilityName(getCurrentAgent().getCapability().getName());
		entry.setCommandEnvelope(this);
	}

	/**
	 * @param ex
	 */
	public void log(Exception ex)
	{
		if(logger != null)
			logger.exception(ex);
	}

	/**
	 * 
	 */
	public void aboutToBeSent()
	{
		logger = null;
		qualifiedCapabilityName = null;
		node = null;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return super.toString() + "(command:" + getCommand() + ")"; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @param node The node to set.
	 */
	public void setNode(Node node)
	{
		this.node = node;
	}

	/**
	 * @return
	 */
	public ILocalEventRegistry getNodeEventRegistry()
	{
		return node.getNodeEventRegistry();
	}

	/**
	 * @return
	 */
	public IObjectSpace getNodeObjectSpace()
	{
		return node.getNodeObjectSpace();
	}

	public boolean isSynchronousResponse()
	{
		return false;
	}
	/**
	 * @return
	 */
	public AgentPath getAgentPath()
	{
		return currentAgent.getAgentPath();
	}
	
	public LocalAgentPath getLocalAgentPath()
	{
		return new LocalAgentPath(currentAgent.getAgentPath());
	}

	public AgentProxy getCrossAgent(AgentPath path)
	throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		if(!currentAgent.getApplication().isSystemApplication())
			throw new InsufficientPrivilegeException("privilege only granted to system application"); //$NON-NLS-1$
		return currentAgent.getApplication().getCrossAgent(path);
	}
	
	/**
	 * @return
	 */
	public IAgentProxy getAgentProxy()
	{
		return currentAgent.getAgentProxy();
	}

	/**
	 * @param capabilityName
	 * @return
	 * @throws IllegalArgumentException
	 * @throws NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 */
	public IAgentProxy getAllForeignAgents(String capabilityName)
		throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return currentAgent.getAllForeignAgents(capabilityName);
	}

	/**
	 * @param nodeName
	 * @param capabilityName
	 * @return
	 * @throws IllegalArgumentException
	 * @throws NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 */
	public IAgentProxy getForeignAgent(String nodeName, String capabilityName)
		throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return currentAgent.getForeignAgent(nodeName, capabilityName);
	}
	
	public String getOriginNodeName()
	{
		if(history.getHistoryItems().isEmpty())
			return null;
		CommandHistoryItem item = (CommandHistoryItem) history.getHistoryItems().get(0);
		return item.getCreationNodeName();
	}
	
	public String getLastOriginNodeName()
	{
		return getLastHistoryItem().getAgentPath().getNodeName();
	}
	
	public AgentPath getOriginAgentPath()
	{
		if(history.getHistoryItems().isEmpty())
			return null;
		CommandHistoryItem item = (CommandHistoryItem) history.getHistoryItems().get(0);
		return item.getAgentPath();
	}

	/**
	 * @return Returns the predefinedInterpreter.
	 */
	public IInterpreter getPredefinedInterpreter()
	{
		return predefinedInterpreter;
	}

	/**
	 * @param predefinedInterpreter The predefinedInterpreter to set.
	 */
	public void setPredefinedInterpreter(IInterpreter predefinedInterpreter)
	{
		this.predefinedInterpreter = predefinedInterpreter;
	}

	/**
	 * @param sourceDestinationAgentPath
	 * @param originNodeName
	 * @param command2
	 */
	public void setHistoryItem(AgentPath sourceDestinationAgentPath, String originNodeName, ICommand command)
	{
		CommandHistoryItem item = getHistory().getLast();
		item.setCommandName(command.getQualifiedName());
		item.setAgentPath(new AgentPath(originNodeName, sourceDestinationAgentPath));
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException
	{
		CommandEnvelope env = (CommandEnvelope) super.clone();
		env.command = (ICommand) env.getCommand().clone(); 
		env.history = (CommandHistory) env.clone();
		env.id = UniqueIdGenerator.next();
		return env;
	}

	/**
	 * @param name
	 * @return
	 * @throws NoSuchElementException
	 * @throws RemoteException
	 */
	public ILocalClusterEventRegistry getClusterEventRegistry(String name) throws NoSuchElementException, RemoteException {
		return node.getClusterEventRegistry(name);
	}

	/**
	 * @return
	 */
	public boolean isEventHandlerCommand()
	{
		return getCommand() instanceof ExecuteEventHandlerCommand;
	}
	/**
	 * @return Returns the destinationAgent.
	 */
	public AgentPath getDestinationAgent()
	{
		return destinationAgent;
	}

	/**
	 * @return
	 */
	public boolean isForeign()
	{
		return isForeign;
	}
	
	/**
	 * @return
	 */
	public boolean isLocal()
	{
		return !isForeign;
	}

	/**
	 * @param node2
	 */
	public void setLocalOrForeign(Node nodeArg)
	{
		isForeign = !getLastOriginNodeName().equals(nodeArg.getName());
	}

	/**
	 * @return
	 */
	public String getQualifiedCapabilityName()
	{
		if(qualifiedCapabilityName == null)
		{
			AgentPath agentPath = getDestinationAgent(); 
			qualifiedCapabilityName = agentPath.getApplicationName() + "." + agentPath.getCapabilityName(); //$NON-NLS-1$
		}
		return qualifiedCapabilityName;
	}

	/**
	 * @return Returns the executionFinished.
	 */
	public boolean isExecutionFinished()
	{
		return executionFinished;
	}
	/**
	 * @param executionFinished The executionFinished to set.
	 */
	public void setExecutionFinished(boolean executionFinished)
	{
		this.executionFinished = executionFinished;
	}
	/**
	 * @return Returns the executionStarted.
	 */
	public boolean isExecutionStarted()
	{
		return executionStarted;
	}
	/**
	 * @param executionStarted The executionStarted to set.
	 */
	public void setExecutionStarted(boolean executionStarted)
	{
		this.executionStarted = executionStarted;
	}
	/**
	 * @return Returns the currentScheduler.
	 */
	public CommandScheduler getCurrentScheduler()
	{
		return currentScheduler;
	}
	/**
	 * @param currentScheduler The currentScheduler to set.
	 */
	public void setCurrentScheduler(CommandScheduler currentScheduler)
	{
		this.currentScheduler = currentScheduler;
	}
	/**
	 * @return Returns the suspended.
	 */
	public boolean isSuspended()
	{
		synchronized(suspendedLock)
		{
			return suspended;
		}
	}
	/**
	 * @param suspended The suspended to set.
	 */
	public void setSuspended(boolean suspended)
	{
		synchronized(suspendedLock)
		{
			this.suspended = suspended;
		}		
	}

	public void arrivedAt(Node node)
	{
		setLocalOrForeign(node);
		setArrivalTime();
	}
	
	/**
	 * @param command The command to set.
	 */
	public void setCommand(ICommand command)
	{
		this.command = command;
	}
	/**
	 * @param destinationAgent The destinationAgent to set.
	 */
	public void setDestinationAgent(AgentPath destinationAgent)
	{
		this.destinationAgent = destinationAgent;
	}

	/**
	 * @return
	 */
	public NodeName getNodeNameObject()
	{
		return new NodeName(node.getName());
	}

	/**
	 * @return
	 */
	public LocalCapabilityPath getLocalCapabilityPath()
	{
		AgentPath path = getAgentPath();
		return new LocalCapabilityPath(path.getApplicationName(), path.getCapabilityName());
	}

	/**
	 * @return
	 */
	public boolean removeFromQueue()
	{
		if(currentScheduler == null)
			return false;
		return currentScheduler.remove(this) != null;
	}

	/**
	 * 
	 */
	public void reinsertIntoQueue ()
	{
		// currentScheduler must not be null, because CommandEnvelope is re-inserted into queue.
		AssertUtility.assertTrue(currentScheduler != null);
		currentScheduler.accept(this);
	}
	
	public LocalCapabilityPath getDestinationAgentCapabilityPath()
	{
		return new LocalCapabilityPath(destinationAgent.getApplicationName(), destinationAgent.getCapabilityName());
	}
	
	/**
	 * @return Returns the kind.
	 */
	public int getKind()
	{
		return kind;
	}
	/**
	 * @param kind The kind to set.
	 */
	public void setKind(int kind)
	{
		this.kind = kind;
	}
	
	public boolean testSetSuspended()
	{
		synchronized(suspendedLock)
		{
			if(suspended)
				return false;
			suspended = true;
			return true;
		}
	}
	
	public Result invokeOnCommand(String methodName) 
	throws SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException
	{
		if(command != null)
		{
			Method method = command.getClass().getDeclaredMethod(methodName, null);
			if(method == null)
				return new Result(false);
			return (Result) method.invoke(command, null);
		}
		return new Result(false);
	}

	/**
	 * @return
	 */
	public String getCapabilityPath()
	{
		return getLocalCapabilityPath().getQualifiedPathName();
	}
	
	public boolean isWorkloadCommand()
	{
		return kind == COMMAND_TYPE_BALANCING || kind == COMMAND_TYPE_SHARING;
	}

	/**
	 * @return
	 */
	public boolean isIgnore()
	{
		return isIgnore;
	}

	/**
	 * @param isIgnore The isIgnore to set.
	 */
	public void setIgnore(boolean isIgnore)
	{
		this.isIgnore = isIgnore;
	}
}
