/*
 * 
 * Copyright (C) 2003, 2004 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.oliverplohmann.de/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE JDOM AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.cas.spaces;

import java.io.Serializable;

import org.almendra.commons.threads.LockUtility;


/**
 * The <code>AbstractObjectSpace</code> class implements an objectSpace for all nodes
 * in a cluster where all agents in the cluster can insert, retrieve, or remove objects.
 * 
 * @author Oliver Plohmann
 * @version 1.00, 15.02.2004
 * @see ISerializableObjectSpace
 * @see ClusterSpaceRemote
 */

public class ClusterSpace extends AbstractObjectSpace implements ISerializableObjectSpace
{
	public final static String RMI_PREFIX = "Janet.clusterSpace."; //$NON-NLS-1$
	
	/**
	 * Creates a new <code>ClusterSpace</code> object.
	 */
	public ClusterSpace()
	{
		super();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#removeAttacher(java.lang.String)
	 */
	public ISerializableAttacher removeAttacher(String key)
	{
		LockUtility.acquire(lock);
		ISerializableAttacher attacher = (ISerializableAttacher) attachers.remove(key);
		lock.release();
		return attacher;
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#removeDetacher(java.lang.String)
	 */
	public ISerializableDetacher removeDetacher(String key)
	{
		LockUtility.acquire(lock);
		ISerializableDetacher detacher = (ISerializableDetacher) detachers.remove(key);
		lock.release();
		return detacher;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#attach(java.lang.String, java.io.Serializable, org.almendra.commons.janet.cas.spaces.ISerializableAttacher)
	 */
	public void attach(String key, Serializable attachment, ISerializableAttacher attacher)
	{
		LockUtility.acquire(lock);
		Serializable object = (Serializable) objects.get(key);
		attacher.attach(attachment, object);
		lock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#put(java.lang.String, java.io.Serializable)
	 */
	public void put(String key, Serializable object)
	{
		LockUtility.acquire(lock);
		objects.put(key, object);
		lock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#installDetacher(java.lang.String, org.almendra.commons.janet.cas.spaces.ISerializableDetacher)
	 */
	public void installDetacher(String key, ISerializableDetacher detacher)
	{
		LockUtility.acquire(lock);
		detachers.put(key, detacher);
		lock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#installAttacher(java.lang.String, org.almendra.commons.janet.cas.spaces.ISerializableAttacher)
	 */
	public void installAttacher(String key, ISerializableAttacher attacher)
	{
		LockUtility.acquire(lock);
		attachers.put(key, attacher);
		lock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#detach(java.lang.String, org.almendra.commons.janet.cas.spaces.ISerializableDetacher)
	 */
	public Serializable detach(String key, ISerializableDetacher detacher)
	{
		LockUtility.acquire(lock);
		Serializable object = (Serializable) objects.get(key);
		Serializable attachment = detacher.detach(object);
		lock.release();
		return attachment;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#attach(java.lang.String, java.io.Serializable)
	 */
	public void attach(String key, Serializable attachment) throws NoInstalledAttacher
	{
		LockUtility.acquire(lock);
		IAttacher attacher = (IAttacher) attachers.get(key);
		if (attacher == null)
			throwNoInstalledAttacher(key);
		Object object = objects.get(key);
		attacher.attach(attachment, object);
		lock.release();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#remove(java.lang.String)
	 */
	public Serializable remove(String key)
	{
		LockUtility.acquire(lock);
		Serializable object = (Serializable) objects.remove(key);
		lock.release();
		return object;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#detach(java.lang.String)
	 */
	public Serializable detach(String key) throws NoInstalledDetacher
	{
		LockUtility.acquire(lock);
		Serializable object = (Serializable) objects.get(key);
		ISerializableDetacher detacher = (ISerializableDetacher) detachers.get(key);
		if (detacher == null)
			throwNoInstalledDetacher(key);
		Serializable attachment = detacher.detach(object);
		lock.release();
		return attachment;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.ISerializableObjectSpace#get(java.lang.String)
	 */
	public Serializable get(String key)
	{
		LockUtility.acquire(lock);
		Serializable object = (Serializable) objects.get(key);
		lock.release();
		return object;
	}

}
